//@author  Ajit Jagtap
//@version Thu Jul 05 2018 06:12:07 GMT+0530 (IST)

// Package cachemdl will help cache object into memory. It Uses LRU algo
package cachemdl

import (
	"time"

	"github.com/patrickmn/go-cache"
)

// FastCacheHelper -
type FastCacheHelper struct {
	FastCache   *cache.Cache
	Expiration  time.Duration
	CleanupTime time.Duration

	MaxEntries int
}

type fastCacheOption func(*FastCacheHelper)

func FCWithMaxEntries(i int) fastCacheOption {
	return func(cfg *FastCacheHelper) {
		cfg.MaxEntries = i
	}
}

func FCWithExpiration(exp time.Duration) fastCacheOption {
	return func(cfg *FastCacheHelper) {
		cfg.Expiration = exp
	}
}

func FCWithCleanupInterval(ivl time.Duration) fastCacheOption {
	return func(cfg *FastCacheHelper) {
		cfg.CleanupTime = ivl
	}
}

// Setup initializes fastcache cache for application. Must be called only once.
func (fastCacheHelper *FastCacheHelper) Setup(maxEntries int, expiration time.Duration, cleanupTime time.Duration) {

	fastCacheHelper.MaxEntries = maxEntries
	fastCacheHelper.Expiration = expiration
	fastCacheHelper.FastCache = cache.New(fastCacheHelper.Expiration, fastCacheHelper.CleanupTime)

}

// SetupFastCache initializes fastcache cache for application and returns its instance.
func SetupFastCache(opts ...fastCacheOption) *FastCacheHelper {
	fc := new(FastCacheHelper)

	for i := range opts {
		opts[i](fc)
	}

	fc.FastCache = cache.New(fc.Expiration, fc.CleanupTime)
	return fc
}

// Get -
func (fastCacheHelper *FastCacheHelper) Get(key string) (interface{}, bool) {
	return fastCacheHelper.FastCache.Get(key)
}

// GetItems -
func (fastCacheHelper *FastCacheHelper) GetItems() map[string]cache.Item {
	return fastCacheHelper.FastCache.Items()
}

// SetNoExpiration -
func (fastCacheHelper *FastCacheHelper) SetNoExpiration(key string, object interface{}) {
	fastCacheHelper.FastCache.Set(key, object, cache.NoExpiration)
}

// Set -
func (fastCacheHelper *FastCacheHelper) Set(key string, object interface{}) {
	fastCacheHelper.FastCache.Set(key, object, cache.DefaultExpiration)
}

// SetWithExpiration -
func (fastCacheHelper *FastCacheHelper) SetWithExpiration(key string, object interface{}, duration time.Duration) {
	fastCacheHelper.FastCache.Set(key, object, duration)
}

// Purge -
func (fastCacheHelper *FastCacheHelper) Purge() {
	fastCacheHelper.FastCache.Flush()
}

// Delete -
func (fastCacheHelper *FastCacheHelper) Delete(key string) {
	fastCacheHelper.FastCache.Delete(key)
}

// GetItemsCount : Number of items in the cache
func (fastCacheHelper *FastCacheHelper) GetItemsCount() int {
	return fastCacheHelper.FastCache.ItemCount()
}

func (fh *FastCacheHelper) Type() int {
	return TypeFastCache
}

// GetAll returns all keys with values present in memory. **This is not intended for production use. May hamper performance**
func (fastCacheHelper *FastCacheHelper) GetAll() map[string]interface{} {
	items := fastCacheHelper.FastCache.Items()

	result := make(map[string]interface{}, len(items))
	for k, v := range items {
		result[k] = v.Object
	}

	return result
}
