package bucket

import (
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"

	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/filetype"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/locker"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/utiliymdl/guidmdl"
)

type PackBucket struct {
	Bucket
	InFileIndexSchemaMap map[string]filetype.InFileIndex `json:"inFileIndexMap"`
	// TODO: filepointer cache
	packFiles map[string]filetype.PackFile
}

func NewPackBucket(bucketNameQuery string, isDynamicName bool, bucketPath string, inFileIndexSchemaMap map[string]filetype.InFileIndex) (*PackBucket, error) {
	if bucketNameQuery == "" {
		return nil, errormdl.Wrap("please provide value of bucketNameQuery")
	}

	bucket := Bucket{
		BucketID:        guidmdl.GetGUID(),
		BucketNameQuery: bucketNameQuery,
		IsDynamicName:   isDynamicName,
		BucketPath:      bucketPath,
	}
	packBucket := PackBucket{}
	packBucket.Bucket = bucket
	if inFileIndexSchemaMap != nil {
		packBucket.InFileIndexSchemaMap = inFileIndexSchemaMap
	} else {
		packBucket.InFileIndexSchemaMap = make(map[string]filetype.InFileIndex)
	}
	return &packBucket, nil
}

// TODO: add fdb index data call
func (pb *PackBucket) Insert(filePath string, data *gjson.Result) error {
	requestedFileType := data.Get("fileType").String()
	if len(requestedFileType) == 0 {
		return errormdl.Wrap("please specify fileType")
	}
	_, ok := pb.InFileIndexSchemaMap[requestedFileType]
	if !ok {
		return errormdl.Wrap("filetype not found: " + requestedFileType)
	}
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()
	if err != nil {
		return err
	}
	err = packFile.Write(data)
	return err
}

func (pb *PackBucket) findOne(filePath string, queries []string, data *gjson.Result) (string, error) {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()
	if err != nil {
		return "", err
	}
	result, err := packFile.Read(queries, data)
	return result, err
}

func (pb *PackBucket) Find(filePaths []string, queries []string, data *gjson.Result) (string, error) {
	requestedFileType := data.Get("fileType").String()
	if len(requestedFileType) == 0 {
		return "", errormdl.Wrap("please specify fileType")
	}
	_, ok := pb.InFileIndexSchemaMap[requestedFileType]
	if !ok {
		return "", errormdl.Wrap("filetype not found: " + requestedFileType)
	}
	queries = append(queries, `#[fileType==`+requestedFileType+`]`)
	resultArray := "[]"
	for i := range filePaths {
		result, err := pb.findOne(filePaths[i], queries, data)
		if err != nil {
			return resultArray, err
		}
		for _, val := range gjson.Parse(result).Array() {
			resultArray, _ = sjson.Set(resultArray, "-1", val.Value())
		}
	}
	return resultArray, nil
}
func (pb *PackBucket) updateOne(filePath string, queries []string, data *gjson.Result) (gjson.Result, error) {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()
	if err != nil {
		return gjson.Result{}, err
	}
	result, err := packFile.Update(queries, data)
	return result, err
}

func (pb *PackBucket) Update(filePaths []string, queries []string, data *gjson.Result) (*gjson.Result, []error) {
	requestedFileType := data.Get("fileType").String()
	if len(requestedFileType) == 0 {
		loggermdl.LogError("please specify fileType")
		return nil, []error{errormdl.Wrap("please specify fileType")}
	}
	_, ok := pb.InFileIndexSchemaMap[requestedFileType]
	if !ok {
		return nil, []error{errormdl.Wrap("filetype not found: " + requestedFileType)}
	}
	queries = append(queries, `#[fileType=="`+requestedFileType+`"]`)

	finalResultArray := []gjson.Result{}
	errList := []error{}

	for i := range filePaths {
		resultArray, err := pb.updateOne(filePaths[i], queries, data)
		if err != nil {
			errList = append(errList, err)
			continue
		}
		finalResultArray = append(finalResultArray, resultArray.Array()...)
	}

	resultListStr := "[]"
	for _, resultObj := range finalResultArray {
		resultListStr, _ = sjson.Set(resultListStr, "-1", resultObj.Value())
	}
	result := gjson.Parse(resultListStr)
	return &result, errList
}

func (pb *PackBucket) deleteOne(filePath string, queries []string, data *gjson.Result) (int, error) {

	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()
	recordsDeletedCnt, err := packFile.Remove(queries)
	return recordsDeletedCnt, err
}

func (pb *PackBucket) Delete(filePaths []string, queries []string, data *gjson.Result) (int, []error) {
	recordsDeletedCnt := 0
	fileType := data.Get("fileType").String()
	if len(fileType) == 0 {
		loggermdl.LogError("fileType value not provided")
		return recordsDeletedCnt, []error{errormdl.Wrap("please specify fileType")}
	}

	_, ok := pb.InFileIndexSchemaMap[fileType]
	if !ok {
		loggermdl.LogError("infileIndex for specified fileType not found")
		return recordsDeletedCnt, []error{errormdl.Wrap("infileIndex for specified fileType not found")}
	}

	queries = append(queries, `#[fileType=="`+fileType+`"]`)
	noDataFoundCnt := 0
	errList := []error{}
	for i := range filePaths {
		deletedRecordsCnt, err := pb.deleteOne(filePaths[i], queries, data)
		if err != nil {
			if err.Error() == "not found" {
				noDataFoundCnt++
				continue
			}
			errList = append(errList, err)
			continue
		}
		recordsDeletedCnt += deletedRecordsCnt
	}

	if noDataFoundCnt == len(filePaths) {
		errList = []error{errormdl.Wrap("data not found")}
	}
	return recordsDeletedCnt, errList
}

func (pb *PackBucket) WriteMedia(filePath string, mediaData []byte, rs *gjson.Result) (string, error) {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()
	if err != nil {
		return "", err
	}
	recordID, err := packFile.WriteMedia(mediaData, rs)
	return recordID, err
}

func (pb *PackBucket) ReadMedia(filePath string, recordID string) ([]byte, *gjson.Result, error) {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()

	if err != nil {
		return nil, nil, err
	}
	dataByte, result, err := packFile.ReadMedia(recordID)
	return dataByte, result, err
}

func (pb *PackBucket) UpdateMedia(filePath string, recordID string, mediaData []byte, rs *gjson.Result) error {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()
	if err != nil {
		return err
	}

	err = packFile.UpdateMedia(recordID, mediaData, rs)
	return err
}

func (pb *PackBucket) UpsertMedia(filePath string, recordID string, mediaData []byte, rs *gjson.Result) (string, error) {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		packFile.Close()
		locker.Unlock()
	}()
	if err != nil {
		return recordID, err
	}
	result, err := packFile.UpsertMedia(recordID, mediaData, rs)
	return result, err
}

func (pb *PackBucket) DeleteMedia(filePath string, recordID string) error {
	// TODO: implement media delete
	return nil
}

func (pb *PackBucket) ReorgFile(filePath string) error {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	packFile, err := filetype.NewPackFile(filePath, pb.InFileIndexSchemaMap, pb.Bucket.SecurityProvider)
	defer func() {
		if packFile != nil {
			packFile.Close()
		}
		locker.Unlock()
	}()
	if err != nil {
		loggermdl.LogError("ReorgFile error:", err, " file path:", filePath)
		return err
	}
	err = packFile.Reorg()
	return err
}

func (pb *PackBucket) Reorg(filePaths []string) (errList []error) {
	for i := range filePaths {
		err := pb.ReorgFile(filePaths[i])
		if err != nil {
			errList = append(errList, err)
			continue
		}
	}
	return
}
