package bucket

import (
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/locker"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/securityprovider"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/filetype"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/utiliymdl/guidmdl"
	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"
)

type SimpleBucket struct {
	Bucket
	// TODO: implement lazy
	EnableLazy       bool
	securityProvider securityprovider.SecurityProvider
}

func NewSimpleBucket(bucketNameQuery string, isDynamicName bool, isLazyEnable bool, bucketPath string) (*SimpleBucket, error) {
	if bucketNameQuery == "" {
		return nil, errormdl.Wrap("please provide value of bucketNameQuery")
	}

	b := Bucket{
		BucketID:        guidmdl.GetGUID(),
		BucketNameQuery: bucketNameQuery,
		IsDynamicName:   isDynamicName,
		BucketPath:      bucketPath,
	}

	bucket := SimpleBucket{
		EnableLazy: isLazyEnable,
	}
	bucket.Bucket = b
	return &bucket, nil
}

func (sb *SimpleBucket) Insert(filePath string, data *gjson.Result) error {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	simpleFile, err := filetype.NewSimpleFile(filePath, sb.Bucket.SecurityProvider)
	defer func() {
		simpleFile.Close()
		locker.Unlock()
	}()
	if err != nil {
		return err
	}
	err = simpleFile.Write(data)
	return err
}

func (sb *SimpleBucket) findOne(filePath string, data *gjson.Result) ([]byte, error) {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	simpleFile, err := filetype.NewSimpleFile(filePath, sb.Bucket.SecurityProvider)
	defer func() {
		simpleFile.Close()
		locker.Unlock()
	}()

	if err != nil {
		return nil, err
	}

	dataByte, err := simpleFile.Read(data)
	return dataByte, err
}

func (sb *SimpleBucket) Find(filePaths []string, queries []string, data *gjson.Result) (string, error) {
	resultArray := "[]"

	for i := range filePaths {

		result, err := sb.findOne(filePaths[i], data)
		if err != nil {
			return resultArray, err
		}
		resultArray, _ = sjson.Set(resultArray, "-1", gjson.ParseBytes(result).Value())
	}
	return resultArray, nil
}

func (sb *SimpleBucket) updateOne(filePath string, queries []string, data *gjson.Result) (gjson.Result, error) {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	simpleFile, err := filetype.NewSimpleFile(filePath, sb.Bucket.SecurityProvider)
	defer func() {
		simpleFile.Close()
		locker.Unlock()
	}()

	if err != nil {
		return gjson.Result{}, err
	}
	result, err := simpleFile.Update(data)
	return result, err
}

func (sb *SimpleBucket) Update(filePaths []string, queries []string, data *gjson.Result) (*gjson.Result, []error) {
	errList := []error{}
	resultListStr := "[]"

	for i := range filePaths {
		updatedData, err := sb.updateOne(filePaths[i], queries, data)
		if err != nil {
			errList = append(errList, err)
			continue
		}
		resultListStr, _ = sjson.Set(resultListStr, "-1", updatedData.Value())
	}
	result := gjson.Parse(resultListStr)
	return &result, errList
}

func (sb *SimpleBucket) deleteOne(filePath string, queries []string, data *gjson.Result) error {
	locker := locker.NewLocker(filePath)
	locker.Lock()
	simpleFile, err := filetype.NewSimpleFile(filePath, sb.Bucket.SecurityProvider)
	defer func() {
		simpleFile.Close()
		locker.Unlock()
	}()

	if err != nil {
		return err
	}
	err = simpleFile.Remove()
	return err
}

func (sb *SimpleBucket) Delete(filePaths, queries []string, data *gjson.Result) (recordsDeletedCnt int, errList []error) {
	noDataFoundCnt := 0
	for i := range filePaths {
		err := sb.deleteOne(filePaths[i], queries, data)
		if err != nil {
			if err.Error() == "not found" {
				noDataFoundCnt++
				continue
			}
			errList = append(errList, err)
			continue
		}
		recordsDeletedCnt++
	}

	if noDataFoundCnt == len(filePaths) {
		errList = []error{errormdl.Wrap("no data found")}
	}
	return
}
