package corefdb

import (
	"errors"
	"fmt"
	"io"
	"net/http"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/cachemdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/bucket"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/filetype"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/index"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/locker"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/securityprovider"
	"github.com/tidwall/buntdb"
	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"
)

const (
	// INDEXFOLDER -INDEXFOLDER
	INDEXFOLDER = "index"
	// LazyCallBackFnAppendBucket - LazyCallBackFnAppendBucket
	LazyCallBackFnAppendBucket = "LazyWriterAppendBucketCallBackFn"
	// LazyCallBackFnSaveIndex - LazyCallBackFnSaveIndex
	LazyCallBackFnSaveIndex = "LazyWriterCallBackFnAppendBucketSaveIndex"
	lineBreak               = "\r\n"
	IndexKeyValSeperator    = "="
	FileType                = "fileType"
	MigrationTypeUpdate     = "MigrationTypeUpdate"
	MigrationTypeReplace    = "MigrationTypeReplace"
	MigrationTypeKeyword    = "migrationType"
	MigrationConfigFilename = "migrationConfig"
	PathSeperator           = "/"
	LookupTypeReference     = "Reference"
	keyLookupType           = "lookupType"
	keyURL                  = "url"
	keyCachedPath           = "cachedPath"
	keyRowID                = "rowID"
	// DhawalD - New Constants Added - 31-July-2021
	QueryOperatorOR  = "OR"
	QueryOperatorAND = "AND"
)

// ErrNoDataFound - This error describes that the required data might be deleted and not found. Kindly ignore this error in caller.
var ErrNoDataFound = errors.New("data not found")

var databases cachemdl.FastCacheHelper
var defaultDB string
var IsDebugModeOn bool

func init() {
	databases.Setup(1, 1000, 1000)
	if os.Getenv("FDBMODE") == "debug" {
		fmt.Println("fdb debug mode is on")
		IsDebugModeOn = true
	}
}

// FDB - FDB
type FDB struct {
	DBName            string
	DBPath            string `json:"dbPath"`
	EnableSecurity    bool   `json:"enableSec"` // if enabled, fdb files will be encrypted
	EnableCompression bool   `json:"enableCmp"` // if enabled, fdb files will be compressed and then encrypted
	indexes           map[string]*index.Index
	indexMux          sync.Mutex
	buckets           map[string]bucket.Store
	bLocker           sync.Mutex
	securityProvider  securityprovider.SecurityProvider
}

// CreateFDBInstance - creates fdb instance
func CreateFDBInstance(dbPath, dbName string, isDefault bool) (*FDB, error) {
	fdb := &FDB{
		DBPath:   dbPath,
		indexes:  make(map[string]*index.Index),
		indexMux: sync.Mutex{},
		buckets:  make(map[string]bucket.Store),
		bLocker:  sync.Mutex{},
		DBName:   dbName,
	}

	if isDefault {
		defaultDB = dbName
	}
	databases.SetNoExpiration(dbName, fdb)
	return fdb, nil
}

// GetFDBInstance - returns fdb instance
func GetFDBInstance(dbName string) (*FDB, error) {
	if dbName == "" {
		dbName = defaultDB
	}

	rawDB, ok := databases.Get(dbName)
	if !ok {
		loggermdl.LogError("Database instance not found")
		return nil, errormdl.Wrap("Database instance not found")
	}
	fdb, ok := rawDB.(*FDB)
	if !ok {
		loggermdl.LogError("Can not cast object into *FDB")
		return nil, errormdl.Wrap("Can not cast object into *FDB")
	}
	return fdb, nil
}

func (fdb *FDB) SetSecurityProvider(securityProvider securityprovider.SecurityProvider) error {
	for key := range fdb.buckets {
		val := fdb.buckets[key]
		if bucketObj, ok := val.(bucket.Securable); ok {
			bucketObj.Secure(securityProvider)
			bucketStore, _ := bucketObj.(bucket.Store)
			fdb.buckets[key] = bucketStore
		}
	}
	fdb.securityProvider = securityProvider
	return nil
}

// EnableFDBSecurity enables security. Files will be encrypted.
func (fdb *FDB) EnableFDBSecurity(sec bool) {
	if !sec {
		return
	}
	fdb.EnableSecurity = sec
}

// RegisterNewIndex - RegisterNewIndex returns new index
func (fdb *FDB) RegisterNewIndex(indexID, indexNameQuery string, isDynamicName bool, indexFields []index.IndexField) (*index.Index, error) {
	fdb.indexMux.Lock()
	defer fdb.indexMux.Unlock()

	indexFilePath, err := filepath.Abs(filepath.Join(fdb.DBPath, INDEXFOLDER, indexID))
	if err != nil {
		return nil, err
	}
	index, err := index.NewIndex(indexID, indexNameQuery, isDynamicName, indexFilePath)
	if err != nil {
		return nil, err
	}
	index.SetFields(indexFields...)
	if _, ok := fdb.indexes[indexID]; ok {
		return nil, errormdl.Wrap("Index ID already found")
	}
	err = index.CreateIndex()
	if err != nil {
		return nil, err
	}
	fdb.indexes[indexID] = index
	return index, nil
}

func (fdb *FDB) AddBucket(bucketID string, bucketObj bucket.Store) error {

	if bucketObj == nil {
		return errormdl.Wrap("bucket is nil")
	}
	if _, ok := fdb.buckets[bucketID]; ok {
		return errormdl.Wrap("bucket already present: " + bucketID)
	}
	if fdb.securityProvider != nil {
		if securable, ok := bucketObj.(bucket.Securable); ok {
			securable.Secure(fdb.securityProvider)
			bucketObj, _ = securable.(bucket.Store)
		}

	}
	fdb.buckets[bucketID] = bucketObj

	return nil
}

// GetFDBIndex - returns index
func (f *FDB) GetFDBIndex(indexID string) (*index.Index, bool) {
	index, ok := f.indexes[indexID]
	return index, ok
}

// GetFDBBucketStore - returns bucket
func (f *FDB) GetFDBBucketStore(bucketID string) (bucket.Store, bool) {
	store, ok := f.buckets[bucketID]
	return store, ok
}

func SaveDataInFDB(dbName string, indexID string, data *gjson.Result) error {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return errormdl.Wrap("fdb instance not found " + dbName)
	}
	// get index from fdb index map
	index, ok := fdb.indexes[indexID]
	if !ok {
		loggermdl.LogError("index not found: ", indexID)
		return errormdl.Wrap("index not found: " + indexID)
	}
	//  get bucket id from index
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return errormdl.Wrap("no buckets available")
	}
	bucketID := index.BucketSequence[bktCnt-1]
	//  get bucket from fdb map
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("bucket not found: ", bucketID)
		return errormdl.Wrap("bucket not found: " + bucketID)
	}
	path, err := fdb.ResolvePath(index, data)
	if err != nil {
		loggermdl.LogError("could not resolve filepath: ", err)
		return errormdl.Wrap("could not resolve filepath: " + err.Error())
	}
	prevVal, err := index.GetEntryByPath(path)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	if data.Get(keyLookupType).String() == LookupTypeReference {
		referenceURL := strings.TrimSpace(data.Get(keyURL).String())
		if referenceURL == "" {
			loggermdl.LogError("invalid data, missing required field - url")
			return errormdl.Wrap("invalid data, missing field - url")
		}
		cachedPath := strings.TrimSpace(data.Get(keyCachedPath).String())
		if cachedPath == "" {
			loggermdl.LogError("invalid data, missing required field - cachedPath")
			return errormdl.Wrap("invalid data, missing field - cachedPath")
		}
		prevVal, _ = sjson.Set(prevVal, keyLookupType, LookupTypeReference)
		prevVal, _ = sjson.Set(prevVal, keyURL, referenceURL)
		prevVal, _ = sjson.Set(prevVal, keyCachedPath, cachedPath)
	}

	filePath, err := filepath.Abs(filepath.Join(fdb.DBPath, path))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	//  call save on bucket
	err = bucketObj.Insert(filePath, data)
	if err != nil {
		loggermdl.LogError("fail to insert data: ", err)
		return errormdl.Wrap("fail to insert data:: " + err.Error())
	}
	if IsDebugModeOn {
		var bucketStoreCpy bucket.Store
		if bucketIns, ok := bucketObj.(*bucket.PackBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}
		if bucketIns, ok := bucketObj.(*bucket.SimpleBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}
		if bucketIns, ok := bucketObj.(*bucket.AppendBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}

		if securable, ok := bucketStoreCpy.(bucket.Securable); ok {
			securable.Secure(nil)
		}

		err = bucketStoreCpy.Insert(filePath+"_debug", data)
		if err != nil {
			loggermdl.LogError("debug error: fail to insert data: ", err)
			// return errormdl.Wrap("fail to insert data:: " + err.Error())
		}
	}

	// save index record in index store
	// basepath, err := filepath.Abs(fdb.DBPath)
	// if err != nil {
	// 	loggermdl.LogError(err)
	// 	return err
	// }
	// basepath = basepath + string(filepath.Separator)
	// path := strings.TrimPrefix(filePath, basepath)
	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	prevVal, _ = sjson.Set(prevVal, keyRowID, rowID)
	updatedJSON, err := updateIndexJSON(index, prevVal, data)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	updatedJSONObj := gjson.Parse(updatedJSON)
	err = index.AddEntry(path, &updatedJSONObj)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	return nil
}

func GetFilePaths(dbName, indexID string, queries []string) ([]string, error) {
	filePaths := make([]string, 0)
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return filePaths, errormdl.Wrap("fdb instance not found " + dbName)
	}
	//  get index Id from index map
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return filePaths, errormdl.Wrap("INDEX not found: " + indexID)
	}

	indexKeyValueMap, err := index.GetEntriesByQueries(queries)
	if err != nil {
		loggermdl.LogError(err)
		return filePaths, err
	}
	if len(indexKeyValueMap) == 0 {
		return filePaths, nil
	}
	for filePath := range indexKeyValueMap {
		filePath, err = filepath.Abs(filepath.Join(fdb.DBPath, filePath))
		if err != nil {
			loggermdl.LogError(err)
			return filePaths, err
		}
		filePaths = append(filePaths, filePath)
	}
	return filePaths, nil
}
func ReadDataFromFDB(dbName, indexID string, data *gjson.Result, queries []string, infileIndexQueries []string) (*gjson.Result, error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return nil, errormdl.Wrap("fdb instance not found " + dbName)
	}
	//  get index Id from index map
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return nil, errormdl.Wrap("INDEX not found: " + indexID)
	}
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return nil, errormdl.Wrap("no buckets available")
	}
	bucketID := index.BucketSequence[bktCnt-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return nil, errormdl.Wrap("Bucket not found: " + bucketID)
	}
	indexKeyValueMap, err := index.GetEntriesByQueries(queries)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	resultToReturn := gjson.Parse("[]")
	if len(indexKeyValueMap) == 0 {
		return &resultToReturn, nil
	}
	filePaths := make([]string, 0)
	// updatedIndexEntries := make(map[string]string)
	for filePath, value := range indexKeyValueMap {
		indexVal := gjson.Parse(value)
		if indexVal.Get(keyLookupType).String() == LookupTypeReference {
			relativeCachedPath := indexVal.Get(keyCachedPath).String()
			cachedPath := filepath.Join(fdb.DBPath, relativeCachedPath)
			err := getReferenceFilePath(cachedPath, indexVal)
			if err != nil {
				loggermdl.LogError(err)
				return nil, err
			}
			value, _ = sjson.Set(value, keyCachedPath, cachedPath)
			// updatedIndexEntries[filePath] = value
			filePath = relativeCachedPath
		}
		filePath, err = filepath.Abs(filepath.Join(fdb.DBPath, filePath))
		if err != nil {
			loggermdl.LogError(err)
			return nil, err
		}

		filePaths = append(filePaths, filePath)
	}
	resultArray, err := bucket.Find(filePaths, infileIndexQueries, data)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	resultToReturn = gjson.Parse(resultArray)
	// err = index.AddEntries(updatedIndexEntries)
	return &resultToReturn, nil
}

func downloadFile(url, destinationPath string) error {
	res, err := http.Get(url)
	if err != nil {
		return err
	}
	defer res.Body.Close()
	if res.StatusCode != http.StatusOK {
		return fmt.Errorf("failed with status %d", res.StatusCode)
	}
	destFolder, _ := filepath.Split(destinationPath)
	err = filemdl.CreateDirectoryRecursive(destFolder)
	if err != nil {
		return err
	}
	out, createError := os.Create(destinationPath)
	defer out.Close()
	if createError != nil {
		return createError
	}
	_, copyError := io.Copy(out, res.Body)
	return copyError
}

func getReferenceFilePath(cachedPath string, indexRow gjson.Result) error {
	url := indexRow.Get(keyURL).String()
	// cachedPath := indexRow.Get(keyCachedPath).String()
	_, err := filemdl.FileInfo(cachedPath)
	if err != nil {
		// _, fileName := filepath.Split(url)
		// destinationPath := filepath.Join(dbBasePath, fileName)
		err := downloadFile(url, cachedPath)
		if err != nil {
			return err
		}
		// loggermdl.LogError("downloading file")
	}
	return nil
}

func UpdateDataInFDB(dbName, indexID string, data *gjson.Result, queries []string, infileIndexQueries []string) (*gjson.Result, []error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return nil, []error{errormdl.Wrap("fdb instance not found " + dbName)}
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return nil, []error{errormdl.Wrap("INDEX not found: " + indexID)}
	}
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return nil, []error{errormdl.Wrap("no buckets available")}
	}
	bucketID := index.BucketSequence[bktCnt-1]
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return nil, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
	}
	indexKeyValueMap, err := index.GetEntriesByQueries(queries)
	if err != nil {
		loggermdl.LogError(err)
		return nil, []error{err}
	}
	resultToReturn := gjson.Parse("[]")
	if len(indexKeyValueMap) == 0 {
		loggermdl.LogError("files not found")
		return &resultToReturn, []error{ErrNoDataFound}
	}
	filePaths := make([]string, 0)
	for filePath := range indexKeyValueMap {
		filePath, err = filepath.Abs(filepath.Join(fdb.DBPath, filePath))
		if err != nil {
			loggermdl.LogError(err)
			return nil, []error{err}
		}
		filePaths = append(filePaths, filePath)
	}
	referenceURL := ""
	cachedPath := ""
	if data.Get(keyLookupType).String() == LookupTypeReference {
		referenceURL = strings.TrimSpace(data.Get(keyURL).String())
		if referenceURL == "" {
			loggermdl.LogError("invalid data, missing required field - url")
			return nil, []error{errormdl.Wrap("invalid data, missing field - url")}
		}
		cachedPath = strings.TrimSpace(data.Get(keyCachedPath).String())
		if cachedPath == "" {
			loggermdl.LogError("invalid data, missing required field - cachedPath")
			return nil, []error{errormdl.Wrap("invalid data, missing field - cachedPath")}
		}
	}
	resultArray, errList := bucketObj.Update(filePaths, infileIndexQueries, data)
	if len(errList) > 1 {
		loggermdl.LogError(errList)
		return nil, errList
	}
	if IsDebugModeOn {
		debugFilePath := make([]string, len(filePaths))
		for i, fpath := range filePaths {
			debugFilePath[i] = fpath + "_debug"
		}
		var bucketStoreCpy bucket.Store
		if bucketIns, ok := bucketObj.(*bucket.PackBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}
		if bucketIns, ok := bucketObj.(*bucket.SimpleBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}
		if bucketIns, ok := bucketObj.(*bucket.AppendBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}

		if securable, ok := bucketStoreCpy.(bucket.Securable); ok {
			securable.Secure(nil)
		}

		_, errList := bucketStoreCpy.Update(debugFilePath, infileIndexQueries, data)
		if len(errList) > 1 {
			loggermdl.LogError("debug error: ", errList)
		}
	}

	for filePath, json := range indexKeyValueMap {
		rowID, err := GenRowID(filePath)
		if err != nil {
			errList = append(errList, err)
			continue
		}
		json, _ = sjson.Set(json, keyRowID, rowID)
		if data.Get(keyLookupType).String() == LookupTypeReference {
			json, _ = sjson.Set(json, keyLookupType, LookupTypeReference)
			json, _ = sjson.Set(json, keyURL, referenceURL)
			json, _ = sjson.Set(json, keyCachedPath, cachedPath)
		}
		updatedJSON, err := updateIndexJSON(index, json, data)
		if err != nil {
			errList = append(errList, err)
			continue
		}

		updatedJSONObj := gjson.Parse(updatedJSON)
		err = index.AddEntry(filePath, &updatedJSONObj)
		if err != nil {
			errList = append(errList, err)
			continue
		}
	}
	return resultArray, nil
}

func DeleteDataFromFDB(dbName, indexID string, rs *gjson.Result, queries []string, infileIndexQueries []string) (recordsDeletedCnt int, errList []error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return recordsDeletedCnt, []error{errormdl.Wrap("fdb instance not found " + dbName)}
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return recordsDeletedCnt, []error{errormdl.Wrap("INDEX not found: " + indexID)}
	}
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return 0, []error{errormdl.Wrap("no buckets available")}
	}
	bucketID := index.BucketSequence[bktCnt-1]
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return recordsDeletedCnt, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
	}
	indexKeyValueMap, err := index.GetEntriesByQueries(queries)
	if err != nil {
		loggermdl.LogError(err)
		return recordsDeletedCnt, []error{err}
	}
	if len(indexKeyValueMap) == 0 {
		loggermdl.LogError("files not found")
		return recordsDeletedCnt, []error{ErrNoDataFound}
	}
	filePaths := make([]string, 0)
	for filePath := range indexKeyValueMap {
		filePath, err = filepath.Abs(filepath.Join(fdb.DBPath, filePath))
		if err != nil {
			loggermdl.LogError(err)
			return recordsDeletedCnt, []error{err}
		}
		filePaths = append(filePaths, filePath)
	}
	cnt, errList := bucketObj.Delete(filePaths, infileIndexQueries, rs)
	if len(errList) > 0 {
		loggermdl.LogError(errList)
		return recordsDeletedCnt, errList
	}

	if IsDebugModeOn {
		debugFilePaths := make([]string, len(filePaths))
		for i, fpath := range filePaths {
			debugFilePaths[i] = fpath + "_debug"
		}
		var bucketStoreCpy bucket.Store
		if bucketIns, ok := bucketObj.(*bucket.PackBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}
		if bucketIns, ok := bucketObj.(*bucket.SimpleBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}
		if bucketIns, ok := bucketObj.(*bucket.AppendBucket); ok {
			cpy := *bucketIns
			bucketStoreCpy = &cpy
		}

		if securable, ok := bucketStoreCpy.(bucket.Securable); ok {
			securable.Secure(nil)
		}
		_, errList := bucketStoreCpy.Delete(debugFilePaths, infileIndexQueries, rs)
		if len(errList) > 1 {
			loggermdl.LogError("debug", errList)
		}

	}

	if _, ok := bucketObj.(*bucket.SimpleBucket); ok {
		for path := range indexKeyValueMap {
			err := index.Delete(path)
			if err != nil {
				errList = append(errList, err)
			}
		}
	}
	return cnt, errList
}

func SaveMediaInFDB(dbName, indexID string, mediaData []byte, data *gjson.Result) (recordPath string, err error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return recordPath, errormdl.Wrap("fdb instance not found " + dbName)
	}
	// get index from fdb index map
	index, ok := fdb.indexes[indexID]
	if !ok {
		loggermdl.LogError("index not found: ", indexID)
		return recordPath, errormdl.Wrap("index not found: " + indexID)
	}
	//  get bucket id from index
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return recordPath, errormdl.Wrap("no buckets available")
	}
	bucketID := index.BucketSequence[bktCnt-1]
	//  get bucket from fdb map
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("bucket not found: ", bucketID)
		return recordPath, errormdl.Wrap("bucket not found: " + bucketID)
	}
	path, err := fdb.ResolvePath(index, data)
	if err != nil {
		loggermdl.LogError("could not resolve filepath: ", err)
		return recordPath, errormdl.Wrap("could not resolve filepath: " + err.Error())
	}
	prevVal, err := index.GetEntryByPath(path)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	filePath, err := filepath.Abs(filepath.Join(fdb.DBPath, path))
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}

	mediaStore, ok := bucketObj.(bucket.MediaStore)
	if !ok {
		loggermdl.LogError("cannot write media data on this bucket: ", bucketID)
		return recordPath, errormdl.Wrap("cannot write media data on this bucket: " + bucketID)
	}
	//  call save on bucket
	recordID, err := mediaStore.WriteMedia(filePath, mediaData, data)
	if err != nil {
		loggermdl.LogError("fail to insert data: ", err)
		return recordPath, errormdl.Wrap("fail to insert data:: " + err.Error())
	}
	if IsDebugModeOn {
		_, err := mediaStore.WriteMedia(filePath+"_debug", mediaData, data)
		if err != nil {
			loggermdl.LogError("debug error : fail to insert data: ", err)
		}
	}

	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	prevVal, _ = sjson.Set(prevVal, keyRowID, rowID)
	updatedJSON, err := updateIndexJSON(index, prevVal, data)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	updatedJSONObj := gjson.Parse(updatedJSON)
	err = index.AddEntry(path, &updatedJSONObj)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	recordPath = fdb.DBName + PathSeperator + indexID + PathSeperator + rowID + PathSeperator + recordID
	return recordPath, nil
}

func GetMediaFromFDB(dbName, indexID string, rowID, recordID string) (dataByte []byte, fileMeta gjson.Result, err error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return dataByte, fileMeta, errormdl.Wrap("fdb instance not found " + dbName)
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return dataByte, fileMeta, errormdl.Wrap("INDEX not found: " + indexID)
	}
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return dataByte, fileMeta, errormdl.Wrap("no buckets available")
	}
	bucketID := index.BucketSequence[bktCnt-1]
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return dataByte, fileMeta, errormdl.Wrap("Bucket not found: " + bucketID)
	}
	mediaReader, ok := bucketObj.(bucket.MediaStore)
	if !ok {
		loggermdl.LogError("cannot write media data on this bucket: ", bucketID)
		return dataByte, fileMeta, errormdl.Wrap("cannot write media data on this bucket: " + bucketID)
	}
	queries := []string{`#[` + keyRowID + `=` + rowID + `]`}
	indexKeyValueMap, err := index.GetEntriesByQueries(queries)
	if err != nil {
		loggermdl.LogError(err)
		return dataByte, fileMeta, err
	}
	if len(indexKeyValueMap) == 0 {
		loggermdl.LogError("files not found")
		return dataByte, fileMeta, nil
	}
	filePath := ""
	// updatedIndexEntries := make(map[string]string)
	for path, value := range indexKeyValueMap {
		indexVal := gjson.Parse(value)
		if indexVal.Get(keyLookupType).String() == LookupTypeReference {
			relativeCachedPath := indexVal.Get(keyCachedPath).String()
			cachedPath := filepath.Join(fdb.DBPath, relativeCachedPath)
			err := getReferenceFilePath(cachedPath, indexVal)
			if err != nil {
				loggermdl.LogError(err)
				return dataByte, fileMeta, err
			}
			value, _ = sjson.Set(value, keyCachedPath, cachedPath)
			// updatedIndexEntries[filePath] = value
			path = relativeCachedPath
		}

		filePath, err = filepath.Abs(filepath.Join(fdb.DBPath, path))
		if err != nil {
			loggermdl.LogError(err)
			return dataByte, fileMeta, err
		}

		// find only one file
		break
	}
	dataByte, metaData, err := mediaReader.ReadMedia(filePath, recordID)
	if err != nil {
		loggermdl.LogError(err)
		return dataByte, fileMeta, err
	}
	fileMeta = metaData.Get("requiredData")
	return dataByte, gjson.Parse(fileMeta.String()), nil
}

func UpdateMediaInFDB(dbName, indexID, recordID string, mediaData []byte, rs *gjson.Result) (string, error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return "", errormdl.Wrap("fdb instance not found " + dbName)
	}
	recordID = strings.TrimSpace(recordID)
	if recordID == "" {
		return "", errormdl.Wrap("please provide recordID")
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: " + indexID)
		return "", errormdl.Wrap("index not found: " + indexID)
	}
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return "", errormdl.Wrap("no buckets available")
	}
	bucketID := index.BucketSequence[bktCnt-1]
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return "", errormdl.Wrap("Bucket not found: " + bucketID)
	}
	path, err := fdb.ResolvePath(index, rs)
	if err != nil {
		loggermdl.LogError("could not find filepath: ", err)
		return "", errormdl.Wrap("could not resolve filepath: " + err.Error())
	}
	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return "", err
	}
	queries := []string{`#[` + keyRowID + `=` + rowID + `]`}

	indexKeyValueMap, err := index.GetEntriesByQueries(queries)
	if err != nil {
		loggermdl.LogError(err)
		return "", err
	}
	if len(indexKeyValueMap) == 0 {
		loggermdl.LogError("files not found")
		return "", errormdl.Wrap("no data found to update")
	}
	filePath := ""
	for path := range indexKeyValueMap {
		filePath, err = filepath.Abs(filepath.Join(fdb.DBPath, path))
		if err != nil {
			loggermdl.LogError(err)
			return "", err
		}
	}
	mediaStore, ok := bucketObj.(bucket.MediaStore)
	if !ok {
		loggermdl.LogError("cannot write media data on this bucket: ", bucketID)
		return "", errormdl.Wrap("cannot write media data on this bucket: " + bucketID)
	}

	err = mediaStore.UpdateMedia(filePath, recordID, mediaData, rs)
	if err != nil {
		loggermdl.LogError(err)
		return "", err
	}

	if IsDebugModeOn {
		err = mediaStore.UpdateMedia(filePath+"_debug", recordID, mediaData, rs)
		if err != nil {
			loggermdl.LogError("debug error:", err)
		}
	}
	for path, val := range indexKeyValueMap {
		json, _ := sjson.Set(val, keyRowID, rowID)
		updatedJSON, err := updateIndexJSON(index, json, rs)
		if err != nil {
			return "", err
		}

		updatedJSONObj := gjson.Parse(updatedJSON)
		err = index.AddEntry(path, &updatedJSONObj)
		if err != nil {
			return "", err
		}
	}
	recordPath := fdb.DBName + PathSeperator + indexID + PathSeperator + rowID + PathSeperator + recordID
	return recordPath, nil
}

func UpsertMediaInFDB(dbName, indexID, recordID string, mediaData []byte, rs *gjson.Result) (string, error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", dbName)
		return "", errormdl.Wrap("fdb instance not found " + dbName)
	}
	recordID = strings.TrimSpace(recordID)
	if recordID == "" {
		return "", errormdl.Wrap("please provide recordID")
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: " + indexID)
		return "", errormdl.Wrap("index not found: " + indexID)
	}
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return "", errormdl.Wrap("no buckets available")
	}
	bucketID := index.BucketSequence[bktCnt-1]
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return "", errormdl.Wrap("Bucket not found: " + bucketID)
	}
	path, err := fdb.ResolvePath(index, rs)
	if err != nil {
		loggermdl.LogError("could not find filepath: ", err)
		return "", errormdl.Wrap("could not resolve filepath: " + err.Error())
	}
	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return "", err
	}
	queries := []string{`#[` + keyRowID + `=` + rowID + `]`}

	indexKeyValueMap, err := index.GetEntriesByQueries(queries)
	if err != nil {
		loggermdl.LogError(err)
		return "", err
	}
	if len(indexKeyValueMap) == 0 {
		loggermdl.LogError("files not found")
		return "", errormdl.Wrap("no data found to update")
	}
	filePath := ""
	for path := range indexKeyValueMap {
		filePath, err = filepath.Abs(filepath.Join(fdb.DBPath, path))
		if err != nil {
			loggermdl.LogError(err)
			return "", err
		}
	}
	mediaStore, ok := bucketObj.(bucket.MediaStore)
	if !ok {
		loggermdl.LogError("cannot write media data on this bucket: ", bucketID)
		return "", errormdl.Wrap("cannot write media data on this bucket: " + bucketID)
	}
	recordID, err = mediaStore.UpsertMedia(filePath, recordID, mediaData, rs)
	if err != nil {
		loggermdl.LogError(err)
		return "", err
	}
	if IsDebugModeOn {
		_, err = mediaStore.UpsertMedia(filePath+"_debug", recordID, mediaData, rs)
		if err != nil {
			loggermdl.LogError("debug error:", err)
		}
	}
	for path, val := range indexKeyValueMap {
		json, _ := sjson.Set(val, keyRowID, rowID)
		updatedJSON, err := updateIndexJSON(index, json, rs)
		if err != nil {
			return "", err
		}

		updatedJSONObj := gjson.Parse(updatedJSON)
		err = index.AddEntry(path, &updatedJSONObj)
		if err != nil {
			return "", err
		}
	}
	recordPath := fdb.DBName + PathSeperator + indexID + PathSeperator + rowID + PathSeperator + recordID
	return recordPath, nil
}

func ReorganizeFiles(dbName string) (errList []error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("Error occured while fetching DB instance", err)
		return []error{errormdl.Wrap("Error occured while fetching DB instance")}
	}
	for _, index := range fdb.indexes {
		bktCnt := len(index.BucketSequence)
		if bktCnt == 0 {
			loggermdl.LogError("no buckets available")
			return []error{errormdl.Wrap("no buckets available")}
		}
		bucketID := index.BucketSequence[bktCnt-1]
		bucketObj, ok := fdb.buckets[bucketID]
		if !ok {
			loggermdl.LogError("Bucket not found: " + bucketID)
			return []error{errormdl.Wrap("Bucket not found: " + bucketID)}
		}

		if packBucketObj, ok := bucketObj.(*bucket.PackBucket); ok {

			indexKeyValMap, err := index.GetAllEntries()
			if err != nil {
				loggermdl.LogError("index data not found", err)
				return []error{errormdl.Wrap("index data not found")}
			}
			if len(indexKeyValMap) == 0 {
				// loggermdl.LogError("no data found to reorganize")
				// return []error{}
				continue
			}
			filePaths := make([]string, len(indexKeyValMap))
			i := 0
			for filePath := range indexKeyValMap {
				sourceFile, err := filepath.Abs(filepath.Join(fdb.DBPath, filePath))
				if err != nil {
					errList = append(errList, errormdl.Wrap("Error occured during reOrg of file data"))
					continue
				}
				filePaths[i] = sourceFile
				i++
			}
			reorgErrs := packBucketObj.Reorg(filePaths[:i])
			if len(reorgErrs) > 0 {
				errList = append(errList, reorgErrs...)
			}
		}
	}
	return errList
}

func ReorganizeFDBBucketData(dbName, indexId string, queries []string) (errList []error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("Error occured while fetching DB instance", err)
		return []error{errormdl.Wrap("Error occured while fetching DB instance")}
	}
	index, found := fdb.GetFDBIndex(indexId)
	if !found {
		loggermdl.LogError("index not found")
		return []error{errormdl.Wrap("index not found")}
	}
	bktCnt := len(index.BucketSequence)
	if bktCnt == 0 {
		loggermdl.LogError("no buckets available")
		return []error{errormdl.Wrap("no buckets available")}
	}
	bucketID := index.BucketSequence[bktCnt-1]
	bucketObj, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return []error{errormdl.Wrap("Bucket not found: " + bucketID)}
	}

	if packBucketObj, ok := bucketObj.(*bucket.PackBucket); ok {
		var indexKeyValMap map[string]string
		var err error
		if len(queries) > 0 {
			indexKeyValMap, err = index.GetEntriesByQueries(queries)
		} else {
			indexKeyValMap, err = index.GetAllEntries()
		}
		if err != nil {
			loggermdl.LogError("index data not found", err)
			return []error{errormdl.Wrap("index data not found")}
		}
		if indexKeyValMap == nil || len(indexKeyValMap) == 0 {
			loggermdl.LogError("no data found to reorganize")
			return []error{}
		}
		filePaths := make([]string, len(indexKeyValMap))
		i := 0
		for filePath := range indexKeyValMap {
			sourceFile, err := filepath.Abs(filepath.Join(fdb.DBPath, filePath))
			if err != nil {
				errList = append(errList, errormdl.Wrap("Error occured during reOrg of file data"))
				continue
			}
			filePaths[i] = sourceFile
			i++
		}
		reorgErrs := packBucketObj.Reorg(filePaths[:i])
		if len(reorgErrs) > 0 {
			errList = append(errList, reorgErrs...)
		}
		if IsDebugModeOn {
			var debugFilePaths []string
			for _, fPath := range filePaths[:i] {
				debugFilePaths = append(debugFilePaths, fPath+"_debug")
			}
			reorgErrs := packBucketObj.Reorg(debugFilePaths)
			if len(reorgErrs) > 0 {
				loggermdl.LogError("debug error", reorgErrs)
			}
		}
	}

	return errList
}

func (fdb *FDB) ResolvePath(index *index.Index, rs *gjson.Result) (string, error) {
	path := ""
	for _, bucketID := range index.BucketSequence {
		bucketObj, ok := fdb.buckets[bucketID]
		if !ok {
			loggermdl.LogError("bucket not found: " + bucketID)
			return "", errormdl.Wrap("bucket not found: " + bucketID)
		}

		pathResolver, ok := bucketObj.(bucket.PathProvider)
		if !ok {
			return "", errormdl.Wrap("cant not find bucket path")
		}
		bucketPath, err := pathResolver.GetPath(rs)
		if err != nil {
			return "", err
		}
		path = filepath.Join(path, bucketPath)
	}
	indexName := index.IndexNameQuery
	if index.IsDynamicName {
		indexName = rs.Get(index.IndexNameQuery).String()
	}
	if indexName == "" {
		return "", errormdl.Wrap("required attribute not provided:" + index.IndexNameQuery)
	}
	path = filepath.Join(path, indexName)
	return path, nil
}

// updateIndexJSON - update JSON with index field data
func updateIndexJSON(index *index.Index, existingData string, rs *gjson.Result) (string, error) {
	json := existingData
	var err error
	for _, indexField := range index.IndexFields {
		if rs.Get(indexField.Query).Value() == nil {
			continue
		}
		json, err = sjson.Set(json, indexField.FieldName, rs.Get(indexField.Query).Value())
		// loggermdl.LogError("Error - ", err)
	}
	return json, err
}

// GenRowID generates hash for the given filename. The length of hash is 16
func GenRowID(name string) (string, error) {
	name = strings.ReplaceAll(filepath.Clean(name), string(filepath.Separator), "")
	rowID, err := securitymdl.GetHash(name)
	if err != nil {
		return "", err
	}

	if len(rowID) > 16 {
		return rowID[:16], nil
	}

	return rowID, nil
}

func addMigrationReplaceConfig(targetBasePath string, securityProvider securityprovider.SecurityProvider) error {
	configfilePath, err := filepath.Abs(filepath.Join(targetBasePath, MigrationConfigFilename))
	if err != nil {
		return err
	}
	migartionConfigStr, _ := sjson.Set("", MigrationTypeKeyword, MigrationTypeReplace)
	rs := gjson.Parse(migartionConfigStr)
	lockerObj := locker.NewLocker(configfilePath)
	lockerObj.Lock()
	simpleFile, err := filetype.NewSimpleFile(configfilePath, securityProvider)
	defer func() {
		simpleFile.Close()
		lockerObj.Unlock()
	}()
	if err != nil {
		return errormdl.Wrap("fail to add migration config: " + err.Error())
	}
	err = simpleFile.Write(&rs)
	return err
}

func addMigrationUpdateConfig(targetBasePath string, securityProvider securityprovider.SecurityProvider) error {
	configfilePath, err := filepath.Abs(filepath.Join(targetBasePath, MigrationConfigFilename))
	if err != nil {
		return err
	}

	migartionConfigStr, _ := sjson.Set("", MigrationTypeKeyword, MigrationTypeUpdate)
	rs := gjson.Parse(migartionConfigStr)
	lockerObj := locker.NewLocker(configfilePath)
	lockerObj.Lock()
	simpleFile, err := filetype.NewSimpleFile(configfilePath, securityProvider)
	defer func() {
		simpleFile.Close()
		lockerObj.Unlock()
	}()
	if err != nil {
		return errormdl.Wrap("fail to add migration config: " + err.Error())
	}
	return simpleFile.Write(&rs)
}

func getMigrationConfig(sourcePath string, rs *gjson.Result, securityProvider securityprovider.SecurityProvider) (*gjson.Result, error) {
	configfilePath, err := filepath.Abs(filepath.Join(sourcePath, MigrationConfigFilename))
	if err != nil {
		return nil, err
	}
	if !filemdl.FileAvailabilityCheck(configfilePath) {
		return nil, errormdl.Wrap("file not found")
	}
	lockerObj := locker.NewLocker(configfilePath)
	lockerObj.Lock()
	simpleFile, err := filetype.NewSimpleFile(configfilePath, securityProvider)
	defer func() {
		simpleFile.Close()
		lockerObj.Unlock()
	}()
	if err != nil {
		return nil, err
	}
	dataByte, err := simpleFile.Read(rs)
	if err != nil {
		return nil, err
	}
	migrationConfig := gjson.ParseBytes(dataByte)
	return &migrationConfig, nil
}

type ZipImporter struct {
	FdbName    string
	IndexID    string
	SourcePath string
	Data       *gjson.Result
}

// ZipExporter is a DataExporter
// allow to export fdb data as zip
// DhawalD - added QueryOperator - 31-July-2021
type ZipExporter struct {
	FdbName       string
	IndexID       string
	Queries       []string
	DestPath      string
	MigrationType string
	QueryOperator string
}

// DataExport exports fdb data as zip
func (z ZipExporter) DataExport() (err error) {
	fdb, err := GetFDBInstance(z.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", z.FdbName)
		return errormdl.Wrap("fdb instance not found: " + z.FdbName)
	}
	index, ok := fdb.GetFDBIndex(z.IndexID)
	if !ok {
		return errormdl.Wrap("INDEX not found: " + z.IndexID)
	}
	sourcePath := ""
	timeStamp := time.Now().Nanosecond()
	targetBasePath := filepath.Join(filemdl.TempDir, strconv.Itoa(timeStamp))
	filteredKeyValMap := map[string]string{}

	// DhawalD - operator check - 31-July-2021
	if z.QueryOperator == QueryOperatorOR {
		filteredKeyValMap, err = index.GetEntriesByOrQueries(z.Queries)
	} else {
		filteredKeyValMap, err = index.GetEntriesByQueries(z.Queries)
	}
	if err != nil {
		return err
	}
	if len(filteredKeyValMap) == 0 {
		return errormdl.Wrap("no data found to export")
	}
	defer func() {
		// removes created zip
		filemdl.DeleteDirectory(targetBasePath)
	}()
	// copy data files
	for path, indexValStr := range filteredKeyValMap {
		indexVal := gjson.Parse(indexValStr)
		if indexVal.Get(keyLookupType).String() == LookupTypeReference {
			relativeCachedPath := indexVal.Get(keyCachedPath).String()
			cachedPath := filepath.Join(fdb.DBPath, relativeCachedPath)
			err := getReferenceFilePath(cachedPath, indexVal)
			if err != nil {
				return err
			}
			targetPath := filepath.Join(targetBasePath, relativeCachedPath)
			_, err = filemdl.CopyFile(cachedPath, targetPath, true)
			if err != nil {
				return err
			}
		}
		sourcePath = filepath.Join(fdb.DBPath, path)
		targetPath := filepath.Join(targetBasePath, path)

		_, err := filemdl.CopyFile(sourcePath, targetPath, true)
		if err != nil {
			return err
		}
	}

	// copy index file
	targetPath := filepath.Join(targetBasePath, INDEXFOLDER, z.IndexID)
	err = AddIndexEntriesInFile(targetPath, filteredKeyValMap, index.SecurityProvider)
	if err != nil {
		return err
	}

	switch z.MigrationType {
	case MigrationTypeUpdate:
		err = addMigrationUpdateConfig(targetBasePath, index.SecurityProvider)
	case MigrationTypeReplace:
		err = addMigrationReplaceConfig(targetBasePath, index.SecurityProvider)
	default:
		return errormdl.Wrap("fail to export data: export operation not allowed on migration type - " + z.MigrationType)
	}

	if err != nil {
		loggermdl.LogError("fail to export data: ", err)
		return errormdl.Wrap("fail to export data: " + err.Error())
	}
	// make zip of copied data to destination folder
	// zip will have name of indexId
	destinationPath := filepath.Join(z.DestPath, z.IndexID)
	return filemdl.Zip(targetBasePath, destinationPath)
}

// DataImport imports data from zip
func (z ZipImporter) DataImport() (err error) {
	fdb, err := GetFDBInstance(z.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", z.FdbName)
		return errormdl.Wrap("fdb instance not found: " + z.FdbName)
	}
	index, ok := fdb.GetFDBIndex(z.IndexID)
	if !ok {
		loggermdl.LogError("index not found: ", z.IndexID)
		return errormdl.Wrap("index not found: " + z.IndexID)
	}
	archivePath := z.SourcePath
	if !filemdl.FileAvailabilityCheck(archivePath) {
		loggermdl.LogError("archive file not found at specified location: ", archivePath)
		return errormdl.Wrap("archive file not found at location: " + archivePath)
	}
	timeStamp := time.Now().Nanosecond()
	pathToExtractZip := filepath.Join(filemdl.TempDir, strconv.Itoa(timeStamp))

	err = filemdl.Unzip(archivePath, pathToExtractZip)
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	defer func() {
		// removes extracted files
		filemdl.DeleteDirectory(pathToExtractZip)
	}()
	childDirs, err := filemdl.ListDirectory(pathToExtractZip)
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	if len(childDirs) == 0 {
		loggermdl.LogError("no data found to import")
		return errormdl.Wrap("no data found to import")
	}
	if !childDirs[0].IsDir() {
		loggermdl.LogError("invalid archive file")
		return errormdl.Wrap("invalid archive file")
	}
	sourcePath := filepath.Join(pathToExtractZip, childDirs[0].Name())
	fdbBasePath := fdb.DBPath

	// loggermdl.LogDebug(sourcePath)
	migrationConfig, err := getMigrationConfig(sourcePath, z.Data, index.SecurityProvider)
	if err != nil {
		loggermdl.LogError("fail to get migration config", err)
		return errormdl.Wrap("invalid archived file")
	}
	migrationType := migrationConfig.Get(MigrationTypeKeyword).String()
	if migrationType != MigrationTypeReplace && migrationType != MigrationTypeUpdate {
		return errormdl.Wrap("import operation not allowed on migration type - " + migrationType)
	}
	err = filepath.Walk(sourcePath, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			loggermdl.LogError("err", err)
			return err
		}
		if info.IsDir() {
			return nil
		}
		//  ignore config file from copying
		if strings.Contains(path, MigrationConfigFilename) {
			return nil
		}

		foundAtIndex := strings.LastIndex(path, sourcePath)
		if foundAtIndex == -1 {
			return errormdl.Wrap("invalid archived file")
		}
		// loggermdl.LogDebug(path)

		// if migartion type is MigrationTypeUpdate then copy index entries from index files else replace index files
		if migrationType == MigrationTypeUpdate && strings.Contains(path, INDEXFOLDER) {
			// load index entries from
			err := ImportIndexEntries(path, fdb, z.IndexID)
			if err != nil {
				loggermdl.LogError("fail to load indexes from data", err)
				return errormdl.Wrap("fail to load indexes")
			}
			err = index.WriteIndexEntriesInFile()
			if err != nil {
				loggermdl.LogError("fail to add indexes: ", err)
				return errormdl.Wrap("fail to add indexes")
			}
			return nil
		}
		destPath := filepath.Join(fdbBasePath, path[foundAtIndex+len(sourcePath):])
		if !filemdl.FileAvailabilityCheck(destPath) {
			dir, _ := filepath.Split(destPath)
			err = filemdl.CreateDirectoryRecursive(dir)
			if err != nil {
				return err
			}
		}
		return filemdl.AtomicReplaceFile(path, destPath)
	})
	if err != nil {
		loggermdl.LogError("fail to import data: ", err)
		return errormdl.Wrap("fail to import data: " + err.Error())
	}

	err = index.LoadIndexEntriesFromFile()
	if err != nil {
		loggermdl.LogError("fail to add indexes", err)
		return errormdl.Wrap("fail to add indexes")
	}

	return nil
}

// ImportIndexEntries -
func ImportIndexEntries(indexFilePath string, fdb *FDB, indexID string) error {
	index, found := fdb.GetFDBIndex(indexID)
	if !found {
		return errormdl.Wrap("index not found")
	}
	if !filemdl.FileAvailabilityCheck(indexFilePath) {
		return nil
	}
	fileData, err := filemdl.FastReadFile(indexFilePath)
	if err != nil {
		loggermdl.LogError("failed to load FDB index from: ", indexFilePath)
		return err
	}
	fileData, err = index.SecurityProvider.Decrypt(fileData, indexFilePath, nil)
	if err != nil {
		loggermdl.LogError("failed to decrypt FDB index data: ", err)
		return errormdl.Wrap("failed to decrypt FDB index data: " + err.Error())
	}
	data := string(fileData)
	indexRecords := strings.Split(data, lineBreak)
	indexDataMap := make(map[string]string)
	for _, indexRecord := range indexRecords {
		indexValues := strings.Split(indexRecord, IndexKeyValSeperator)
		if len(indexValues) == 2 {
			indexDataMap[indexValues[0]] = indexValues[1]
		}
	}
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}

	// update index file by reading all data and updating index file
	return index.AddEntries(indexDataMap)
}

// FileImporter is a DataImporter
// allow to import fdb data from exported folder
type FileImporter struct {
	FdbName    string
	IndexID    string
	SourcePath string
	Data       *gjson.Result
}

// FileExporter is a DataExporter
// allow to export fdb data in a folder
// DhawalD - added QueryOperator - 31-July-2021
type FileExporter struct {
	FdbName       string
	IndexID       string
	Queries       []string
	DestPath      string
	MigrationType string
	QueryOperator string
}

// DataExport exports fdb data in a folder
func (f FileExporter) DataExport() (err error) {
	fdb, err := GetFDBInstance(f.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", f.FdbName)
		return errormdl.Wrap("fdb instance not found: " + f.FdbName)
	}
	index, ok := fdb.GetFDBIndex(f.IndexID)
	if !ok {
		return errormdl.Wrap("INDEX not found: " + f.IndexID)
	}
	sourcePath := ""
	targetBasePath := filepath.Join(f.DestPath, f.IndexID)
	filteredKeyValMap := map[string]string{}
	// DhawalD - operator check - 31-July-2021
	if f.QueryOperator == QueryOperatorOR {
		filteredKeyValMap, err = index.GetEntriesByOrQueries(f.Queries)
	} else {
		filteredKeyValMap, err = index.GetEntriesByQueries(f.Queries)
	}
	if err != nil {
		return err
	}

	if len(filteredKeyValMap) == 0 {
		return errormdl.Wrap("no data found to export")
	}

	for path := range filteredKeyValMap {
		sourcePath = filepath.Join(fdb.DBPath, path)
		targetPath := filepath.Join(targetBasePath, path)
		_, err := filemdl.CopyFile(sourcePath, targetPath, true)
		if err != nil {
			return err
		}
	}

	// copy index file
	targetPath := filepath.Join(targetBasePath, INDEXFOLDER, f.IndexID)
	err = AddIndexEntriesInFile(targetPath, filteredKeyValMap, index.SecurityProvider)
	if err != nil {
		return err
	}

	switch f.MigrationType {
	case MigrationTypeUpdate:
		err = addMigrationUpdateConfig(targetBasePath, index.SecurityProvider)
	case MigrationTypeReplace:
		err = addMigrationReplaceConfig(targetBasePath, index.SecurityProvider)
	default:
		return errormdl.Wrap("export operation not allowed on migration type - " + f.MigrationType)
	}

	return err
}

// DataImport imports data from exported folder
func (f FileImporter) DataImport() (err error) {
	fdb, err := GetFDBInstance(f.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", f.FdbName)
		return errormdl.Wrap("fdb instance not found: " + f.FdbName)
	}
	index, ok := fdb.GetFDBIndex(f.IndexID)
	if !ok {
		loggermdl.LogError("index not found: ", f.IndexID)
		return errormdl.Wrap("index not found: " + f.IndexID)
	}
	if !filemdl.FileAvailabilityCheck(f.SourcePath) {
		loggermdl.LogError("archive file not found at specified location: ", f.SourcePath)
		return errormdl.Wrap("archive file not found at location: " + f.SourcePath)
	}

	timeStamp := time.Now().Nanosecond()
	tempDir := filepath.Join(filemdl.TempDir, strconv.Itoa(timeStamp))
	err = filemdl.CopyDir(f.SourcePath, tempDir)
	defer func() {
		filemdl.DeleteDirectory(tempDir)
	}()
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("fail to copy data")
	}

	childDirs, err := filemdl.ListDirectory(tempDir)
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	if len(childDirs) == 0 {
		loggermdl.LogError("no data found to import")
		return errormdl.Wrap("no data found to import")
	}
	fdbBasePath := fdb.DBPath

	// loggermdl.LogDebug(f.SourcePath)
	migrationConfig, err := getMigrationConfig(tempDir, f.Data, index.SecurityProvider)
	if err != nil {
		loggermdl.LogError("fail to get migration config", err)
		return errormdl.Wrap("invalid archived file")
	}
	migrationType := migrationConfig.Get(MigrationTypeKeyword).String()
	if migrationType != MigrationTypeReplace && migrationType != MigrationTypeUpdate {
		return errormdl.Wrap("import operation not allowed on migration type - " + migrationType)
	}
	err = filepath.Walk(tempDir, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			loggermdl.LogError("err", err)
			return err
		}
		// loggermdl.LogError(path)
		if info.IsDir() {
			return nil
		}
		//  ignore config file from copying
		if strings.Contains(path, MigrationConfigFilename) {
			return nil
		}

		foundAtIndex := strings.LastIndex(path, tempDir)
		if foundAtIndex == -1 {
			return errormdl.Wrap("invalid archived file")
		}
		// if file is index file then copy index entries from index files
		if strings.Contains(path, INDEXFOLDER) {
			// load index entries from file
			err := ImportIndexEntries(path, fdb, f.IndexID)
			if err != nil {
				loggermdl.LogError("fail to import indexes", err)
				return errormdl.Wrap("fail to import indexes")
			}
			err = index.WriteIndexEntriesInFile()
			if err != nil {
				loggermdl.LogError("fail to import indexes: ", err)
				return errormdl.Wrap("fail to import indexes")
			}
			return nil
		}

		destPath := filepath.Join(fdbBasePath, path[foundAtIndex+len(tempDir):])
		if err != nil {
			return err
		}
		if !filemdl.FileAvailabilityCheck(destPath) {
			dir, _ := filepath.Split(destPath)
			err = filemdl.CreateDirectoryRecursive(dir)
			if err != nil {
				return err
			}
		}
		return filemdl.AtomicReplaceFile(path, destPath)
	})

	if err != nil {
		loggermdl.LogError("fail to import data: ", err)
		return errormdl.Wrap("fail to import data: " + err.Error())
	}

	return nil
}

// AddIndexEntriesInFile -AddIndexEntriesInFile
func AddIndexEntriesInFile(indexFilePath string, entries map[string]string, securityProvider securityprovider.SecurityProvider) error {
	// dbPath := filepath.Join(fdbPath, INDEXFOLDER)
	// loggermdl.LogDebug("in log fdb index")
	dataToStore := ""
	for key, val := range entries {
		dataToStore = dataToStore + key + IndexKeyValSeperator + val + lineBreak
	}
	dataByteToWriteRes := []byte{}
	var hashError error
	if len(dataToStore) > 0 {
		_, fileName := filepath.Split(indexFilePath)
		dataByteToWriteRes, hashError = securityProvider.Encrypt([]byte(dataToStore), fileName, nil)
		if errormdl.CheckErr1(hashError) != nil {
			return errormdl.CheckErr1(hashError)
		}
	}
	return filemdl.WriteFile(indexFilePath, dataByteToWriteRes, true, false)
}
