package corefdb

import (
	"log"
	"testing"
	"time"

	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/bucket"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/filetype"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/index"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/securityprovider"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
)

var (
	dbPath              = "/home/legerrac/fdb_data"
	dbName              = "myfdb"
	CheckLazyIndexWrite = false
	lazyWriteInterval   = 3
)

func init() {

	fdb, err := CreateFDBInstance(dbPath, dbName, true)
	if err != nil {
		log.Fatal(err)
	}

	// enable security on fdb
	{

		fdbSecurityProvider := securityprovider.New(securityprovider.SecurityConfig{
			EncKey:         "myenckey",
			UserDefinedKey: "mkcl",
			FieldQuery:     "instituteId",
		})

		err = fdb.SetSecurityProvider(fdbSecurityProvider)
		if err != nil {
			log.Fatal(err)
		}
	}

	//  creating simple bucket
	{
		simpleBucket, err := bucket.NewSimpleBucket("Simple", false, false, "")
		if err != nil {
			log.Fatal(err)
		}
		fields := []index.IndexField{
			{
				FieldName: "name",
				Query:     "name",
			},
		}
		i, err := fdb.RegisterNewIndex("stdId", "studentId", true, fields)
		if err != nil {
			log.Fatal(err)
		}
		err = simpleBucket.AddIndex(i)
		if err != nil {
			log.Fatal(err)
		}
		err = fdb.AddBucket(simpleBucket.BucketID, simpleBucket)
		if err != nil {
			log.Fatal(err)
		}
	}
	//  creating pack bucket
	{

		inFileIndexSchemaMap := map[string]filetype.InFileIndex{
			"Exam": {
				FileType: "Exam",
				IndexFields: []filetype.InFileIndexField{
					{
						FieldName: "examId",
						Query:     "examId",
					},
				},
			},
			"Profile": {
				FileType: "Profile",
				IndexFields: []filetype.InFileIndexField{
					{
						FieldName: "class",
						Query:     "class",
					},
				},
			},
		}

		packBucket, err := bucket.NewPackBucket("PackBucket", false, "", inFileIndexSchemaMap)
		if err != nil {
			log.Fatal(err)
		}
		packIndexfields := []index.IndexField{
			{
				FieldName: "name",
				Query:     "name",
			},
		}
		packbucketIndex, err := fdb.RegisterNewIndex("studentPack", "stdId", true, packIndexfields)
		if err != nil {
			log.Fatal(err)
		}
		err = packBucket.AddIndex(packbucketIndex)
		if err != nil {
			log.Fatal(err)
		}
		err = fdb.AddBucket(packBucket.BucketID, packBucket)
		if err != nil {
			log.Fatal(err)
		}
	}

	{
		inFileIndexSchemaMap := map[string]filetype.InFileIndex{
			"PresentationDetails": {
				FileType: "PresentationDetails",
				IndexFields: []filetype.InFileIndexField{
					{
						FieldName: "presentationId",
						Query:     "presentationId",
					},
				},
			},
		}

		packBucket, err := bucket.NewPackBucket("PackBucket", false, "", inFileIndexSchemaMap)
		if err != nil {
			log.Fatal(err)
		}
		packIndexfields := []index.IndexField{
			{
				FieldName: "presentationId",
				Query:     "presentationId",
			},
		}
		packbucketIndex, err := fdb.RegisterNewIndex("presentation", "presentationId", true, packIndexfields)
		if err != nil {
			log.Fatal(err)
		}
		err = packBucket.AddIndex(packbucketIndex)
		if err != nil {
			log.Fatal(err)
		}
		err = fdb.AddBucket(packBucket.BucketID, packBucket)
		if err != nil {
			log.Fatal(err)
		}
	}
	//  creating append bucket
	{
		appendBucket, err := bucket.NewAppendBucket("Append", false, false, "")
		if err != nil {
			log.Fatal(err)
		}
		fields := []index.IndexField{
			{
				FieldName: "name",
				Query:     "name",
			},
		}
		i, err := fdb.RegisterNewIndex("stdResponse", "studentId", true, fields)
		if err != nil {
			log.Fatal(err)
		}
		err = appendBucket.AddIndex(i)
		if err != nil {
			log.Fatal(err)
		}
		err = fdb.AddBucket(appendBucket.BucketID, appendBucket)
		if err != nil {
			log.Fatal(err)
		}
	}
	// load index record from index files
	// {
	// 	indexIds := []string{"stdId", "studentPack", "stdResponse"}
	// 	for _, indexID := range indexIds {
	// 		indexFilePath := filepath.Join(fdb.DBPath, INDEXFOLDER, indexID)
	// 		err = LoadFDBIndexFromFile(indexFilePath, fdb, indexID)
	// 		if err != nil {
	// 			log.Fatal(err)
	// 		}
	// 	}

	// }
}
func TestSaveDataInSimpleBucket(t *testing.T) {
	tests := []struct {
		data           string
		ShouldGetError bool
	}{
		{
			data:           `{"instituteId":"geca" ,"name": "ajay","fileType": "EXAM","studentId": 1234,"examId":"exam001","examName": "test102","totalQuestions":50,"marks": 38}`,
			ShouldGetError: false,
		},
		{
			data:           `{"instituteId":"geca" ,"name": "ajay","fileType": "EXAM","studentId": 1235,"examId":"exam001","examName": "test102","totalQuestions":50,"marks": 38}`,
			ShouldGetError: false,
		},
		{
			data:           `{"instituteId":"geca" ,"name": "ajay","fileType": "EXAM","studentId": 1236,"examId":"exam001","examName": "test102","totalQuestions":50,"marks": 38}`,
			ShouldGetError: false,
		},
	}

	indexID := "stdId"

	for i, test := range tests {
		studentObj := gjson.Parse(test.data)
		t.Logf("\t Test: %d", i)
		now := time.Now()
		err := SaveDataInFDB(dbName, indexID, &studentObj)
		if err != nil {
			log.Fatal(err)
		}
		timeElapsed := time.Since(now)
		// loggermdl.LogError("timeElapsed", timeElapsed)
		t.Logf("\t %s\t should be able to save data %s", "succeed", timeElapsed)
	}
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestGetDataFromSimpleBucket(t *testing.T) {
	indexID := "stdId"

	data, _ := sjson.Set("", "fileType", "Exam")
	data, _ = sjson.Set(data, "instituteId", "geca")
	studentObj := gjson.Parse(data)

	queries := []string{`#[name==ajay]`}
	inFileIndexQueries := []string{}

	result, err := ReadDataFromFDB(dbName, indexID, &studentObj, queries, inFileIndexQueries)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogDebug("result", result.String())
}
func TestUpdateDataInNormalBucket(t *testing.T) {
	indexID := "stdId"
	data, _ := sjson.Set("", "abc", 10)
	data, _ = sjson.Set(data, "marks", 30)
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "fileType", "Exam")

	queries := []string{`#[name=="ajay"]`}
	infileIndexQueries := []string{}
	studentObj := gjson.Parse(data)

	updatedData, errList := UpdateDataInFDB(dbName, indexID, &studentObj, queries, infileIndexQueries)

	if len(errList) > 0 {
		loggermdl.LogError(errList)
	}
	loggermdl.LogDebug("updatedData", updatedData)
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestDeleteDataFromNormalBucket(t *testing.T) {
	indexID := "stdId"

	queries := []string{`#[name==ajay]`}
	infileIndexQueries := []string{}

	// data, _ := sjson.Set("", "fileType", "Exam")
	// data, _ := sjson.Set(data, "studentId", 1234)

	studentObj := gjson.Result{}

	recordsDeletedCnt, errList := DeleteDataFromFDB(dbName, indexID, &studentObj, queries, infileIndexQueries)
	if len(errList) > 0 {
		loggermdl.LogError("errList", errList)
	}
	loggermdl.LogDebug("recordsDeletedCnt", recordsDeletedCnt)
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

// DhawalD - TestCase for TestSaveDataInPackBucket - 31-July-2021
func TestSaveDataInPackBucket(t *testing.T) {
	indexID := "studentPack"
	// get fdb obj

	data, _ := sjson.Set("", "name", "ajay")
	data, _ = sjson.Set(data, "stdId", 1238)
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "examName", "unit2")
	data, _ = sjson.Set(data, "totalQuestion", 50)
	data, _ = sjson.Set(data, "marks", 26)
	data, _ = sjson.Set(data, "examId", "MATH001")
	data, _ = sjson.Set(data, "fileType", "Exam")

	studentObj := gjson.Parse(data)

	err := SaveDataInFDB(dbName, indexID, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	data, _ = sjson.Set("", "name", "dhawal")
	data, _ = sjson.Set(data, "stdId", 1239)
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "examName", "unit2")
	data, _ = sjson.Set(data, "totalQuestion", 50)
	data, _ = sjson.Set(data, "marks", 28)
	data, _ = sjson.Set(data, "examId", "MATH001")
	data, _ = sjson.Set(data, "fileType", "Exam")

	studentObj = gjson.Parse(data)

	err = SaveDataInFDB(dbName, indexID, &studentObj)
	if err != nil {
		log.Fatal(err)
	}

	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}
func TestGetDataFromPackBucket(t *testing.T) {
	indexID := "studentPack"
	data, _ := sjson.Set("", "name", "ajay")
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "fileType", "Exam")

	studentObj := gjson.Parse(data)

	queries := []string{`#[name=ajay]`}
	infileIndexQueries := []string{`#[examId=MATH001]`}

	result, err := ReadDataFromFDB(dbName, indexID, &studentObj, queries, infileIndexQueries)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogError("result", result)
}

func TestUpdateDataInPackBucket(t *testing.T) {
	indexID := "studentPack"

	data, _ := sjson.Set("", "abc", "123")
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "fileType", "Exam")

	studentObj := gjson.Parse(data)
	queries := []string{`#[name=ajay]`}
	infileIndexQueries := []string{`#[examId=MATH002]`}
	result, errList := UpdateDataInFDB(dbName, indexID, &studentObj, queries, infileIndexQueries)
	if len(errList) > 0 {
		log.Fatal(errList)
	}
	loggermdl.LogError("result", result)
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestDeleteDataFromPackBucket(t *testing.T) {
	indexID := "studentPack"

	data, _ := sjson.Set("", "abc", "123")
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "fileType", "Exam")

	studentObj := gjson.Parse(data)

	queries := []string{`#[name=vijay]`}
	infileIndexQueries := []string{`#[examId=MATH002]`}
	cnt, errList := DeleteDataFromFDB(dbName, indexID, &studentObj, queries, infileIndexQueries)
	if len(errList) > 0 {
		log.Fatal(errList)
	}
	loggermdl.LogError("delete cnt", cnt)
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestSaveMediaInPackBucket(t *testing.T) {
	indexID := "studentPack"

	data, _ := sjson.Set("", "name", "vijay")
	data, _ = sjson.Set(data, "stdId", 1239)

	studentObj := gjson.Parse(data)

	filePath := "C:\\Users\\vivekn\\Pictures\\gopher.png"

	dataByte, err := filemdl.ReadFile(filePath)
	if err != nil {
		log.Fatal(err)
	}
	recordID, err := SaveMediaInFDB(dbName, indexID, dataByte, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogError("recordId", recordID)
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}

		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestReadMediaFromPackBucket(t *testing.T) {
	indexID := "studentPack"

	rowID := "-568385317811827"
	recordID := "1ZtdPpUYLuKmcHJpTn2LXQ4XABM"
	_, fileMeta, err := GetMediaFromFDB(dbName, indexID, rowID, recordID)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogError("fileMeta", fileMeta)
}

func TestUpdateMediaInPackBucket(t *testing.T) {
	indexID := "studentPack"

	data, _ := sjson.Set("", "name", "vijay")
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "stdId", 1239)

	studentObj := gjson.Parse(data)

	recordID := "1ZsY055dnvgL6qutjy5sbnsubS8"
	filePath := "C:\\Users\\vivekn\\Pictures\\gopher.png"

	dataByte, err := filemdl.ReadFile(filePath)
	if err != nil {
		log.Fatal(err)
	}
	recordPath, err := UpdateMediaInFDB(dbName, indexID, recordID, dataByte, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogError("recordPath", recordPath)
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestUpsertMediaInPackBucket(t *testing.T) {
	indexID := "studentPack"

	data, _ := sjson.Set("", "name", "vijay")
	data, _ = sjson.Set(data, "stdId", 1239)

	studentObj := gjson.Parse(data)

	recordID := "dfsdg123243"
	filePath := "C:\\Users\\vivekn\\Pictures\\ghg.jpg"

	dataByte, err := filemdl.ReadFile(filePath)
	if err != nil {
		log.Fatal(err)
	}
	recordPath, err := UpsertMediaInFDB(dbName, indexID, recordID, dataByte, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogError("recordPath", recordPath)
	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestSaveDataInAppendBucket(t *testing.T) {
	tests := []struct {
		data           string
		ShouldGetError bool
	}{
		{
			data:           `{"name": "ajay","instituteId": "geca", "fileType": "EXAM","studentId": 1234,"examId":"exam001","examName": "test102","totalQuestions":50,"marks": 38}`,
			ShouldGetError: false,
		},
		{
			data:           `{"name": "sanjay","instituteId": "geca","fileType": "EXAM","studentId": 1234,"examId":"exam002","examName": "test102","totalQuestions":50,"marks": 38}`,
			ShouldGetError: false,
		},
	}

	indexID := "stdResponse"

	for i, test := range tests {
		studentObj := gjson.Parse(test.data)
		t.Logf("\t Test: %d", i)
		err := SaveDataInFDB(dbName, indexID, &studentObj)
		if err != nil {
			log.Fatal(err)
		}
		t.Logf("\t %s\t should be able to save data", "succeed")
	}

	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestExportDataAsZip(t *testing.T) {
	indexID := "studentPack"

	zipExporter := ZipExporter{
		DestPath:      "D:\\exe\\backup",
		IndexID:       indexID,
		MigrationType: MigrationTypeUpdate,
		Queries:       []string{},
		FdbName:       dbName,
	}

	err := zipExporter.DataExport()
	if err != nil {
		log.Fatal(err)
	}
}

// DhawalD - Test code for TestExportDataAsZipForOrQuery - 31-July-2021
func TestExportDataAsZipForOrQuery(t *testing.T) {
	indexID := "studentPack"
	// TestSaveDataInPackBucket(t)
	zipExporter := ZipExporter{
		DestPath:      "/home/legerrac/Documents/ExportFDB",
		IndexID:       indexID,
		MigrationType: MigrationTypeUpdate,
		Queries:       []string{`#[name==dhawal]`, `#[name==ajay]`},
		FdbName:       dbName,
	}
	zipExporter.QueryOperator = QueryOperatorOR
	err := zipExporter.DataExport()
	if err != nil {
		log.Fatal(err)
	}
}

func TestImportDataFromZip(t *testing.T) {
	indexID := "studentPack"
	fdbObj, err := GetFDBInstance(dbName)
	if err != nil {
		log.Fatal(err)
	}

	zipImporter := ZipImporter{
		Data:       nil,
		SourcePath: "D:\\exe\\backup\\" + indexID,
		FdbName:    dbName,
		IndexID:    indexID,
	}
	err = zipImporter.DataImport()
	// err := ImportZip("myfdb", "home/vivekn/fdb_data/dest/stdId")
	if err != nil {
		loggermdl.LogError(err)
		log.Fatal(err)
	}

	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}

}

func TestExportDataAsFiles(t *testing.T) {
	indexID := "stdId"
	exporter := FileExporter{
		DestPath:      "D:\\exe\\backup",
		IndexID:       indexID,
		MigrationType: MigrationTypeUpdate,
		Queries:       []string{},
		FdbName:       dbName,
	}

	err := exporter.DataExport()
	if err != nil {
		log.Fatal(err)
	}
}

// DhawalD - TestCase for QueryOperatorOR - 31-July-2021
func TestExportDataAsFilesForOrQuery(t *testing.T) {
	indexID := "studentPack"
	// TestSaveDataInPackBucket(t)
	exporter := FileExporter{
		DestPath:      "/home/legerrac/Documents/ExportFDB",
		IndexID:       indexID,
		MigrationType: MigrationTypeReplace,
		Queries:       []string{`#[name==dhawal]`, `#[name==ajay]`},
		FdbName:       dbName,
	}
	exporter.QueryOperator = QueryOperatorOR
	err := exporter.DataExport()
	if err != nil {
		log.Fatal(err)
	}
}
func TestImportDataFromFile(t *testing.T) {
	indexID := "stdId"
	fdbObj, err := GetFDBInstance(dbName)
	if err != nil {
		log.Fatal(err)
	}

	fileImporter := FileImporter{
		Data:       nil,
		SourcePath: "D:\\exe\\backup\\" + indexID,
		FdbName:    dbName,
		IndexID:    indexID,
	}
	err = fileImporter.DataImport()
	// err := ImportZip("myfdb", "home/vivekn/fdb_data/dest/stdId")
	if err != nil {
		loggermdl.LogError(err)
		log.Fatal(err)
	}

	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestReorg(t *testing.T) {
	indexID := "studentPack"
	fdbObj, err := GetFDBInstance(dbName)
	if err != nil {
		log.Fatal(err)
	}
	data, _ := sjson.Set("", "name", "ajay")
	data, _ = sjson.Set(data, "stdId", 1237)
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "examName", "unit2")
	data, _ = sjson.Set(data, "totalQuestion", 50)
	data, _ = sjson.Set(data, "marks", 26)
	data, _ = sjson.Set(data, "examId", "MATH001")
	data, _ = sjson.Set(data, "fileType", "Exam")
	studentObj := gjson.Parse(data)
	err = SaveDataInFDB(dbName, indexID, &studentObj)
	if err != nil {
		log.Fatal(err)
	}
	data, _ = sjson.Set(data, "examId", "MATH003")
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "stdId", 1238)
	std2 := gjson.Parse(data)
	err = SaveDataInFDB(dbName, indexID, &std2)
	if err != nil {
		log.Fatal(err)
	}
	data, _ = sjson.Set(data, "examId", "MATH001")
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "stdId", 1237)
	std3 := gjson.Parse(data)
	err = SaveDataInFDB(dbName, indexID, &std3)
	if err != nil {
		log.Fatal(err)
	}
	data, _ = sjson.Set(data, "examId", "MATH002")
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "stdId", 1238)
	std4 := gjson.Parse(data)
	err = SaveDataInFDB(dbName, indexID, &std4)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogDebug("saved")
	queries := []string{`#[name=="ajay"]`}
	infileIndexQueries := []string{`#[examId=="MATH001"]`}
	recordsDeleted, errList := DeleteDataFromFDB(dbName, indexID, &studentObj, queries, infileIndexQueries)
	if len(errList) > 0 {
		loggermdl.LogError(errList)
	}
	index, ok := fdbObj.GetFDBIndex(indexID)
	if !ok {
		log.Fatal("index not found")
	}
	loggermdl.LogDebug("deletd cnt", recordsDeleted)
	err = index.WriteIndexEntriesInFile()
	if err != nil {
		log.Fatal(err)
	}
	now := time.Now()

	errList = ReorganizeFiles("myfdb")
	if len(errList) > 0 {
		log.Fatal(errList)
	}
	elapsed := time.Since(now)

	loggermdl.LogError("elapsed", elapsed)

}

func TestReorgBucket(t *testing.T) {
	indexID := "studentPack"

	now := time.Now()
	requeries := []string{}
	errList := ReorganizeFDBBucketData(dbName, indexID, requeries)
	if len(errList) > 0 {
		log.Fatal(errList)
	}
	elapsed := time.Since(now)

	loggermdl.LogError("elapsed", elapsed)

}

func TestSaveRefDataInPackBucket(t *testing.T) {
	indexID := "presentation"

	// disable security on fdb
	{
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		// securityProvider := securityprovider.New(securityprovider.SecurityConfig{})
		err = fdbObj.SetSecurityProvider(nil)
		if err != nil {
			log.Fatal(err)
		}

	}
	data, _ := sjson.Set("", "url", "https://superschoolstg.mkcl.org/tp_cdn/Marketplace/5fec13a76fc724eb533ec7d8_925166362/Presentations/6033975c421cff1beb2ad2de")
	data, _ = sjson.Set(data, "presentationId", "6033975c421cff1beb2ad2de")
	data, _ = sjson.Set(data, "cachedPath", "Marketplace/5fec13a76fc724eb533ec7d8_925166362/Presentations/6033975c421cff1beb2ad2de")
	data, _ = sjson.Set(data, "lookupType", LookupTypeReference)
	data, _ = sjson.Set(data, "instituteId", "geca")
	data, _ = sjson.Set(data, "fileType", "PresentationDetails")

	studentObj := gjson.Parse(data)

	saveErr := SaveDataInFDB(dbName, indexID, &studentObj)
	if saveErr != nil {
		log.Fatal(saveErr)
	}

	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestGetRefDataFromPackBucket(t *testing.T) {

	// disable security on fdb
	{

		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		// securityProvider := securityprovider.New(securityprovider.SecurityConfig{})
		err = fdbObj.SetSecurityProvider(nil)
		if err != nil {
			log.Fatal(err)
		}
	}
	indexID := "presentation"
	data, _ := sjson.Set("", "presentationId", "6033975c421cff1beb2ad2de")
	data, _ = sjson.Set(data, "fileType", "PresentationDetails")
	data, _ = sjson.Set(data, "instituteId", "geca")

	studentObj := gjson.Parse(data)

	queries := []string{`#[presentationId=6033975c421cff1beb2ad2de]`}
	infileIndexQueries := []string{`#[presentationId=6033975c421cff1beb2ad2de]`}
	// infileIndexQueries := []string{`#[slideId=1nkK4VnLMGfxI9FmemThMbGHaG9]`}

	result, err := ReadDataFromFDB(dbName, indexID, &studentObj, queries, infileIndexQueries)
	if err != nil {
		log.Fatal(err)
	}
	loggermdl.LogError("result", result)

	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}

func TestGetRefMediaFromPackBucket(t *testing.T) {
	indexID := "presentation"

	// data, _ := sjson.Set("", "presentationId", "6033975c421cff1beb2ad2de")
	// data, _ = sjson.Set(data, "fileType", "Media")

	// studentObj := gjson.Parse(data)

	// queries := []string{`#[presentationId=6033975c421cff1beb2ad2de]`}
	// infileIndexQueries := []string{`#[recordID=1nkK7c9wjpcao3coFVOdfMPDO5e]`}
	// infileIndexQueries := []string{`#[slideId=1nkK4VnLMGfxI9FmemThMbGHaG9]`}
	rowID := "1740004634532177"
	recordID := "1nkK7c9wjpcao3coFVOdfMPDO5e"
	_, _, err := GetMediaFromFDB(dbName, indexID, rowID, recordID)
	if err != nil {
		log.Fatal(err)
	}

	if CheckLazyIndexWrite {
		time.Sleep(time.Second * time.Duration(5+lazyWriteInterval))
	} else {
		fdbObj, err := GetFDBInstance(dbName)
		if err != nil {
			log.Fatal(err)
		}
		index, ok := fdbObj.GetFDBIndex(indexID)
		if !ok {
			log.Fatal("index not found")
		}
		err = index.WriteIndexEntriesInFile()
		if err != nil {
			log.Fatal(err)
		}
	}
}
