package filetype

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/securityprovider"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/utiliymdl/guidmdl"
	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"
)

const (
	fileStatusReady                     = 0
	fileStatusUpdatingData              = 1
	fileStatusUpdatingIndex             = 2
	fileStatusOffsetInFile              = 0
	isReadyForUploadOffsetInFile        = 1
	isUpdatedAndNotCommitedOffsetInFile = 2
	isReorgRequiredOffsetInFile         = 3
	isReindexRequiredOffsetInFile       = 4
	footerOffsetInFile                  = 5
	footerOffsetReservedSize            = 15
	footerSizeOffset                    = 20
	filehashOffest                      = 35
	lastUpdatedOffset                   = 43
	dataStartOffset                     = 53
	sizeReservedForHeaders              = 53

	// IndexKeyValSeperator -
	IndexKeyValSeperator = "="
	// FileType - represents key for type of file. Used whenever we need to set file type field in json
	FileType = "fileType"
)

// ErrNoDataFound - ErrNoDataFound
var ErrNoDataFound = errors.New("No data found")

type PackFile struct {
	FilePath             string
	InfileIndexRows      *gjson.Result
	Fp                   *os.File
	SecurityProvider     securityprovider.SecurityProvider
	infileIndexSchemaMap map[string]InFileIndex
}

type InFileIndex struct {
	FileType    string             `json:"fileType"`
	IndexFields []InFileIndexField `json:"indexFields"`
}

type InFileIndexField struct {
	FieldName string `json:"fieldName"`
	Query     string `json:"query"`
}

func NewPackFile(filePath string, infileIndexSchemaMap map[string]InFileIndex, securityProvider securityprovider.SecurityProvider) (*PackFile, error) {
	if filePath == "" {
		return nil, errormdl.Wrap("please provide valid filepath")
	}

	path, err := filepath.Abs(filePath)
	if err != nil {
		return nil, err
	}
	dir, _ := filepath.Split(path)
	if dir != "" {
		err := filemdl.CreateDirectoryRecursive(dir)
		if err != nil {
			return nil, err
		}
	}
	f, err := os.OpenFile(path, os.O_CREATE|os.O_RDWR, os.ModePerm)
	if err != nil {
		return nil, errormdl.Wrap("fail to open file: " + err.Error())
	}
	packFile := PackFile{
		FilePath:         filePath,
		Fp:               f,
		SecurityProvider: securityProvider,
	}
	if infileIndexSchemaMap == nil {
		packFile.infileIndexSchemaMap = make(map[string]InFileIndex)
	} else {
		packFile.infileIndexSchemaMap = infileIndexSchemaMap
	}
	err = initializeFile(f)
	if err != nil {
		return nil, errormdl.Wrap("fail to create pack file: " + err.Error())
	}
	return &packFile, nil
}

func (p *PackFile) Close() error {
	return p.Fp.Close()
}
func (p *PackFile) Write(rs *gjson.Result) (err error) {
	// filePath := p.Fp.Name()
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return errormdl.Wrap("please specify fileType")
	}

	f := p.Fp
	infileIndexSchema, ok := p.infileIndexSchemaMap[fileType]
	if !ok {
		return errormdl.Wrap("infileIndex schema for specified fileType not found: " + fileType)
	}

	indexRowJSON, err := CreateIndexJSON(infileIndexSchema.IndexFields, rs)
	if err != nil {
		return err
	}
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", fileType)
	fileHash, err := securitymdl.GetHash(rs.String())
	if err != nil {
		return err
	}
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
	previousIndexData, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", f.Name(), err)
		return err
	}
	// loggermdl.LogDebug("previous index data", previousIndexData)
	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		return errormdl.Wrap("fail to fetch infile index data")
	}
	dataString := rs.String()
	err = setFileStatusFlag(f, fileStatusUpdatingData)
	if err != nil {
		return err
	}
	// write data
	dataSize, err := addFileDataInFile(f, footerStartOffset, dataString, true, rs, p.SecurityProvider)
	if err != nil {
		return err
	}
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	// append new entry in infile index
	parsedindexRowJSON := gjson.Parse(indexRowJSON)
	// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
	updatedIndexData, _ := sjson.Set(previousIndexData.String(), "-1", parsedindexRowJSON.Value())
	// loggermdl.LogDebug("updatedIndexData", updatedIndexData)
	// updating infile index
	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return err
	}

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(f, footerNewOffset)
	if err != nil {
		return err
	}
	err = setFooterSize(f, int64(len(updatedIndexData)))
	if err != nil {
		return err
	}
	err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
	if err != nil {
		return err
	}
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return err
	}
	err = f.Sync()
	if err != nil {
		return err
	}
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	p.InfileIndexRows = &updatedIndexDataObj
	return nil
}

func (p *PackFile) Read(queries []string, data *gjson.Result) (string, error) {

	filePath := p.Fp.Name()

	indexRows, err := getInFileIndexData(p.Fp)
	if err != nil {
		loggermdl.LogError("index data not found: ", filePath, err)
		return "", err
	}
	for i := 0; i < len(queries); i++ {
		indexRows = indexRows.Get(queries[i] + "#")
	}
	sb := strings.Builder{}
	sb.WriteString("[")
	indexRows.ForEach(func(key, indexRow gjson.Result) bool {
		// read files
		startOffSet := indexRow.Get("startOffset").Int()
		dataSize := indexRow.Get("dataSize").Int()
		if startOffSet == 0 || dataSize == 0 {
			return true
		}
		dataByte := []byte{'{', '}'}
		var err error
		// dataByte, err = filemdl.ReadFileFromOffset(f, startOffSet, dataSize)
		dataByte, err = getFileDataFromPack(p.Fp, startOffSet, dataSize, data, p.SecurityProvider)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		_, err = sb.Write(dataByte)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		sb.WriteString(",")

		return true // keep iterating
	})
	sb.WriteString("]")
	finalResult := strings.Replace(sb.String(), ",]", "]", 1)
	return finalResult, nil
}

func (p *PackFile) Update(queries []string, rs *gjson.Result) (gjson.Result, error) {
	// check fileType index availability
	// check is data present
	// if data present
	// then calculate size of updated data
	// if size is less than or equal to previuos data size
	// then write at the same location
	// else if size of updated data is more than existing data then append it to end of data
	// update startOffset and data size of file in index row
	// update footer offset and footer size
	updatedData := gjson.Result{}
	indexRows, err := getInFileIndexData(p.Fp)
	if err != nil {
		loggermdl.LogError("index data not found: ", p.Fp.Name(), err)
		return updatedData, err
	}

	indexRecordsToUpdate := indexRows
	for _, query := range queries {
		indexRecordsToUpdate = indexRecordsToUpdate.Get(query + "#")
	}

	indexRecordsToUpdateObjs := indexRecordsToUpdate.Array()
	if len(indexRecordsToUpdateObjs) == 0 {
		return updatedData, errormdl.Wrap("no data found")
	}
	resultArrayStr := "[]"
	var updatedInfileIndex *gjson.Result
	var result *gjson.Result
	for _, recordToUpdateIndexRow := range indexRecordsToUpdateObjs {
		result, updatedInfileIndex, err = updateSingleRecordInPackFileUsingFp(p.Fp, recordToUpdateIndexRow, updatedInfileIndex, rs, p.SecurityProvider)
		if err != nil {
			return updatedData, errormdl.Wrap("fail to update data" + err.Error())
		}
		resultArrayStr, _ = sjson.Set(resultArrayStr, "-1", result.Value())
	}
	resultData := gjson.Parse(resultArrayStr)
	return resultData, nil
}

func (p *PackFile) Remove(queries []string) (recordsDeletedCnt int, err error) {
	indexData, err := getInFileIndexData(p.Fp)
	if err != nil {
		loggermdl.LogError("index data not found: ", p.Fp.Name(), err)
		return recordsDeletedCnt, err
	}
	indexRecordsToDelete := indexData
	// loggermdl.LogDebug("indexRecordsToDelete file type", indexRecordsToDelete)
	for _, query := range queries {
		indexRecordsToDelete = indexRecordsToDelete.Get(query + "#")
	}
	indexRowsToDelete := indexRecordsToDelete.Array()
	if len(indexRowsToDelete) == 0 {
		loggermdl.LogError("ErrNoDataFound")
		return recordsDeletedCnt, errormdl.Wrap("not found")
	}
	updatedIndexRecords := indexData
	for _, indexRowToRemove := range indexRowsToDelete {
		updatedIndexRecords, err = removeIndexRow(updatedIndexRecords, indexRowToRemove.String())
		if err != nil {
			loggermdl.LogError("fail to delete record:", err)
			return recordsDeletedCnt, errormdl.Wrap("fail to delete record:" + err.Error())
		}
		recordsDeletedCnt++
	}

	footerOffset := getFooterOffset(p.Fp)
	if footerOffset == -1 {
		return recordsDeletedCnt, errormdl.Wrap("fail to fetch infile index offset")
	}
	newIndexDataSize := len(updatedIndexRecords.String())
	err = setFileStatusFlag(p.Fp, fileStatusUpdatingIndex)
	if err != nil {
		return recordsDeletedCnt, err
	}
	err = setIndexDataInFile(p.Fp, footerOffset, updatedIndexRecords.String())
	if err != nil {
		loggermdl.LogError("fail to update infile index data :", err)
		return recordsDeletedCnt, err
	}
	err = setFileStatusFlag(p.Fp, fileStatusReady)
	if err != nil {
		return recordsDeletedCnt, err
	}
	p.InfileIndexRows = &updatedIndexRecords
	return recordsDeletedCnt, setFooterSize(p.Fp, int64(newIndexDataSize))
}

func (p *PackFile) WriteMedia(mediaData []byte, rs *gjson.Result) (recordID string, err error) {

	fileHash, err := securitymdl.GetHash(rs.String())
	if err != nil {
		loggermdl.LogError("error writing to bucket: ", err)
		return recordID, err
	}
	previousIndexData, err := getInFileIndexData(p.Fp)
	if err != nil {
		loggermdl.LogError("index data not found: ", p.Fp.Name(), err)
		return recordID, err
	}
	// loggermdl.LogDebug("previous index data", previousIndexData)
	footerStartOffset := getFooterOffset(p.Fp)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return recordID, errormdl.Wrap("fail to fetch infile index data")
	}
	// write data
	dataSize, err := addByteDataInFile(p.Fp, footerStartOffset, mediaData, true)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, err
	}

	recordID = guidmdl.GetGUID()
	indexRowJSON := ""
	indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", "Media")
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

	// append new entry in infile index
	parsedindexRowJSON := gjson.Parse(indexRowJSON)
	// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
	updatedIndexData, _ := sjson.Set(previousIndexData.String(), "-1", parsedindexRowJSON.Value())

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(p.Fp, footerNewOffset)
	if err != nil {
		return recordID, err
	}
	err = setFooterSize(p.Fp, int64(len(updatedIndexData)))
	if err != nil {
		return recordID, err
	}
	err = setIndexDataInFile(p.Fp, footerNewOffset, updatedIndexData)
	if err != nil {
		return recordID, err
	}
	err = setFileStatusFlag(p.Fp, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, err
	}
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	err = p.Fp.Sync()
	if err != nil {
		loggermdl.LogError(err)
		return recordID, err
	}
	p.InfileIndexRows = &updatedIndexDataObj
	return recordID, nil

}

func (p *PackFile) UpdateMedia(recordID string, mediaData []byte, rs *gjson.Result) (err error) {

	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}
	indexData, err := getInFileIndexData(p.Fp)
	if err != nil {
		loggermdl.LogError("index data not found: ", p.Fp.Name(), err)
		return err
	}

	indexRows := indexData
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	foundAtIndex := -1
	foundIndexRow := gjson.Result{}
	for index, indexRow := range indexData.Array() {
		r := indexRow.Get("recordID").String()
		if r != "" && indexRow.Get("recordID").String() == recordID {
			foundAtIndex = index
			foundIndexRow = indexRow
			break
		}
	}
	if foundAtIndex == -1 {
		loggermdl.LogError("no data found to update: ", recordID)
		return errormdl.Wrap("no data found to update: " + recordID)
	}

	footerStartOffset := getFooterOffset(p.Fp)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return errormdl.Wrap("fail to fetch infile index data")
	}

	// write data
	dataSize, err := addByteDataInFile(p.Fp, footerStartOffset, mediaData, true)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	indexRowJSON := foundIndexRow.String()
	indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

	updatedIndexRow := gjson.Parse(indexRowJSON)
	updatedIndexData, _ := sjson.Set(indexData.String(), strconv.Itoa(foundAtIndex), updatedIndexRow.Value())

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(p.Fp, footerNewOffset)
	if err != nil {
		return err
	}
	err = setFooterSize(p.Fp, int64(len(updatedIndexData)))
	if err != nil {
		return err
	}
	err = setIndexDataInFile(p.Fp, footerNewOffset, updatedIndexData)
	if err != nil {
		return err
	}
	err = setFileStatusFlag(p.Fp, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	err = p.Fp.Sync()
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	p.InfileIndexRows = &updatedIndexDataObj
	return err
}

func (p *PackFile) UpsertMedia(recordID string, mediaData []byte, rs *gjson.Result) (string, error) {
	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}

	f := p.Fp
	indexData, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", f.Name(), err)
		return recordID, err
	}

	indexRows := indexData
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	foundAtIndex := -1
	foundIndexRow := gjson.Result{}
	for index, indexRow := range indexData.Array() {
		if indexRow.Get("recordID").String() == recordID {
			foundAtIndex = index
			foundIndexRow = indexRow
			break
		}
	}

	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return recordID, errormdl.Wrap("fail to fetch infile index data")
	}
	// TODO: write at previous location
	// write data
	dataSize, err := addByteDataInFile(f, footerStartOffset, mediaData, true)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, err
	}
	indexRowJSON := foundIndexRow.String()

	indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", "Media")
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

	updatedIndexRow := gjson.Parse(indexRowJSON)
	updatedIndexData, _ := sjson.Set(indexData.String(), strconv.Itoa(foundAtIndex), updatedIndexRow.Value())

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(f, footerNewOffset)
	if err != nil {
		return recordID, err
	}
	err = setFooterSize(f, int64(len(updatedIndexData)))
	if err != nil {
		return recordID, err
	}
	err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
	if err != nil {
		return recordID, err
	}
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, err
	}
	err = f.Sync()
	if err != nil {
		return recordID, err
	}
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	p.InfileIndexRows = &updatedIndexDataObj
	return recordID, nil
}

func (p *PackFile) ReadMedia(recordID string) ([]byte, *gjson.Result, error) {
	dataByte := []byte{}
	var metaData *gjson.Result

	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}

	f := p.Fp
	indexData, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", f.Name(), err)
		return dataByte, metaData, err
	}
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	if indexRows.String() == "" {
		loggermdl.LogError("data not found for recordId: ", recordID)
		return dataByte, metaData, errormdl.Wrap("data not found")
	}
	sb := strings.Builder{}
	sb.WriteString("[")
	indexRow := indexRows.Get("0")
	startOffSet := indexRow.Get("startOffset").Int()
	dataSize := indexRow.Get("dataSize").Int()
	if startOffSet == 0 || dataSize == 0 {
		return dataByte, metaData, errormdl.Wrap("data not found")
	}
	dataByte, err = getFileDataFromPack(f, startOffSet, dataSize, nil, nil)
	if err != nil {
		loggermdl.LogError(err)
		return dataByte, metaData, err
	}

	data, _ := sjson.Set("", "requiredData", indexRow.Get("requiredData").String())
	data, _ = sjson.Set(data, "infileIndex", indexData.String())
	metaDataObj := gjson.Parse(data)
	return dataByte, &metaDataObj, nil
}

func (p *PackFile) ReadMediaByQuery(inFileIndexQueries []string) (map[string][]byte, map[string]gjson.Result, error) {
	f := p.Fp
	indexData, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", f.Name(), err)
		return nil, nil, err
	}
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	if indexRows.String() == "" {
		loggermdl.LogError("data not found")
		return nil, nil, errormdl.Wrap("data not found")
	}
	dataMap := make(map[string][]byte, 0)
	metaDataMap := make(map[string]gjson.Result, 0)
	for _, indexRow := range indexRows.Array() {
		startOffSet := indexRow.Get("startOffset").Int()
		dataSize := indexRow.Get("dataSize").Int()
		if startOffSet == 0 || dataSize == 0 {
			return nil, nil, errormdl.Wrap("data not found")
		}
		dataByte, err := getFileDataFromPack(f, startOffSet, dataSize, nil, nil)
		if err != nil {
			loggermdl.LogError(err)
			return nil, nil, err
		}
		recordID := indexRow.Get("recordID").String()
		if recordID == "" {
			return nil, nil, errormdl.Wrap("record id not found")
		}
		data, _ := sjson.Set("", "requiredData", indexRow.Get("requiredData").String())
		// data, _ = sjson.Set(data, "infileIndex", indexData.String())
		metaDataObj := gjson.Parse(data)
		dataMap[recordID] = dataByte
		metaDataMap[recordID] = metaDataObj
	}

	return dataMap, metaDataMap, nil
}

func (p *PackFile) RemoveMedia(recordID string) error {
	queries := []string{`#[recordID=` + recordID + `]`}
	_, err := p.Remove(queries)
	return err
}

func (p *PackFile) Reorg() error {
	f := p.Fp
	_, sourceFileName := filepath.Split(p.FilePath)
	desFileName := sourceFileName + "_" + strconv.FormatInt(time.Now().Unix(), 10)
	tempFilepath, err := filepath.Abs(filepath.Join(filemdl.TempDir, desFileName))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	dir, _ := filepath.Split(tempFilepath)
	if dir != "" {
		createError := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(createError) != nil {
			return errormdl.CheckErr(createError)
		}
	}

	fpTemp, err := os.OpenFile(tempFilepath, os.O_CREATE|os.O_RDWR, os.ModePerm)
	if err != nil {
		return err
	}
	defer func() {
		fpTemp.Close()
	}()

	err = InitializeWithHeaderUsingFp(fpTemp)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	infileIndexRows, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("Reorg error:", err, " file path: ", p.FilePath)
		return err
	}
	// if len(infileIndexRows.Array()) == 0 {
	// 	return nil
	// }
	tempFileFooterStartOffset := getFooterOffset(fpTemp)
	if tempFileFooterStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return errormdl.Wrap("fail to fetch infile index data")
	}
	updatedIndexRowStr := "[]"
	for _, infileIndex := range infileIndexRows.Array() {
		startOffset, err := strconv.Atoi(infileIndex.Get("startOffset").String())
		if err != nil {
			loggermdl.LogError("Error occured while fetching startOffset", err)
			return err
		}
		dataSize, err := strconv.Atoi(infileIndex.Get("dataSize").String())
		if err != nil {
			loggermdl.LogError("Error occured while fetching dataSize", err)
			return err
		}

		byteArr, err := getFileDataFromPack(f, int64(startOffset), int64(dataSize), nil, nil)
		if err != nil {
			loggermdl.LogError("Error occured while reading file data from offset", err)
			return err
		}
		byteCnt, err := addByteDataInFile(fpTemp, tempFileFooterStartOffset, byteArr, false)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		indexRowJSON, _ := sjson.Set(infileIndex.String(), "startOffset", tempFileFooterStartOffset)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", byteCnt)
		indexRowJSONObj := gjson.Parse(indexRowJSON)
		updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "-1", indexRowJSONObj.Value())
		tempFileFooterStartOffset = tempFileFooterStartOffset + byteCnt
	}

	err = setFooterOffset(fpTemp, tempFileFooterStartOffset)
	if err != nil {
		return err
	}
	err = setFooterSize(fpTemp, int64(len(updatedIndexRowStr)))
	if err != nil {
		return err
	}
	err = setIndexDataInFile(fpTemp, tempFileFooterStartOffset, updatedIndexRowStr)
	if err != nil {
		return err
	}
	err = fpTemp.Sync()
	if err != nil {
		return err
	}
	err = fpTemp.Close()
	if err != nil {
		return err
	}
	err = f.Close()
	if err != nil {
		return err
	}

	return filemdl.AtomicReplaceFile(tempFilepath, p.FilePath)
}

//
func CreateIndexJSON(indexFields []InFileIndexField, rs *gjson.Result) (string, error) {
	json := `{}`
	for _, indexField := range indexFields {
		val := rs.Get(indexField.Query).Value()
		// validation
		if val == nil {
			return "", errormdl.Wrap("please provide value for index field: " + indexField.Query)
		}
		json, _ = sjson.Set(json, indexField.FieldName, val)
	}
	return json, nil
}

func updateIndexRow(indexRows *gjson.Result, previousIndexRow gjson.Result, updatedRow gjson.Result) (string, error) {
	indexRowObjs := indexRows.Array()
	if len(indexRowObjs) == 0 {
		return "", errormdl.Wrap("no data found to update")
	}
	// loggermdl.LogDebug("indexRows", indexRows)

	prevIndexRowString := previousIndexRow.String()
	foundRowIndex := -1

	for index, indexRowObj := range indexRowObjs {
		if len(indexRowObj.String()) == len(prevIndexRowString) && indexRowObj.String() != "" && prevIndexRowString != "" && indexRowObj.String() == prevIndexRowString {
			foundRowIndex = index
			break
		}
	}

	if foundRowIndex == -1 {
		return "", errormdl.Wrap("no record found to update")
	}
	var err error
	updatedIndexDataString := indexRows.String()
	// for _, foundRowIndex := range foundRowIndexes {
	updatedIndexDataString, err = sjson.Set(updatedIndexDataString, strconv.Itoa(foundRowIndex), updatedRow.Value())
	if err != nil {
		loggermdl.LogError(err)
		return "", errormdl.Wrap("failed to update index rows")
	}
	// }
	// updatedIndexData := gjson.Parse(updatedIndexDataString)
	return updatedIndexDataString, nil
}

func removeIndexRow(indexRows gjson.Result, indexRowToDelete string) (gjson.Result, error) {

	indexRowObjs := indexRows.Array()
	if len(indexRowObjs) == 0 {
		return indexRows, errormdl.Wrap("no data found to update")
	}
	// loggermdl.LogDebug("indexRows", indexRows)

	foundIndexToDelete := -1
	for index, indexRowObj := range indexRowObjs {
		if indexRowObj.String() != "" && indexRowToDelete != "" && indexRowObj.String() == indexRowToDelete {
			foundIndexToDelete = index
			break
		}
	}

	if foundIndexToDelete == -1 {
		return indexRows, errormdl.Wrap("no record found to delete")
	}
	var err error
	updatedIndexDataString, err := sjson.Delete(indexRows.String(), strconv.Itoa(foundIndexToDelete))
	if err != nil {
		loggermdl.LogError(err)
	}
	return gjson.Parse(updatedIndexDataString), nil
}

func isValidPackFile(f *os.File) (bool, error) {
	if f == nil {
		return false, errormdl.Wrap("file pointer not valid")
	}
	isFilePresent := filemdl.FileAvailabilityCheck(f.Name())
	if !isFilePresent {
		loggermdl.LogDebug(isFilePresent)
		return false, nil
	}
	info, err := f.Stat()
	if err != nil {
		loggermdl.LogDebug(err)
		return false, err
	}
	if info.Size() == 0 {
		return false, nil
	}
	return true, nil
}

func updateSingleRecordInPackFileUsingFp(f *os.File, recordToUpdateIndexRow gjson.Result, infileIndex, rs *gjson.Result, securityProvider securityprovider.SecurityProvider) (*gjson.Result, *gjson.Result, error) {
	fileStartOffset := recordToUpdateIndexRow.Get("startOffset").Int()
	dataSize := recordToUpdateIndexRow.Get("dataSize").Int()
	if fileStartOffset == 0 || dataSize == 0 {
		loggermdl.LogError("index row details incorrect - start offset :", fileStartOffset, " data size :", dataSize)
		return nil, nil, errormdl.Wrap("index row details incorrect")
	}

	existingData, err := getFileDataFromPack(f, fileStartOffset, dataSize, rs, securityProvider)
	if err != nil {
		return nil, nil, err
	}

	updatedDataStr := string(existingData)
	// updating existing data
	rs.ForEach(func(key, val gjson.Result) bool {
		updatedDataStr, _ = sjson.Set(updatedDataStr, key.String(), val.Value())
		return true
	})
	newDataSize := int64(len(updatedDataStr))
	footerStartOffset := getFooterOffset(f)
	updatedFooterOffset := footerStartOffset
	err = setFileStatusFlag(f, fileStatusUpdatingData)
	if err != nil {
		return nil, nil, err
	}
	existingIndexRows := gjson.Parse("[]")
	if infileIndex == nil {
		existingIndexRows, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return nil, nil, err
		}
	} else {
		existingIndexRows = *infileIndex
	}
	updatedDataSize := len(updatedDataStr)
	updatedDataBytes := []byte(updatedDataStr)
	if securityProvider != nil {
		updatedDataBytes, err = securityProvider.Encrypt(updatedDataBytes, f.Name(), rs)
		if err != nil {
			return nil, nil, err
		}
		updatedDataSize = len(updatedDataBytes)
	}
	if int64(updatedDataSize) <= dataSize {
		newDataSize, err = addByteDataInFile(f, fileStartOffset, updatedDataBytes, false)
		if err != nil {
			return nil, nil, err
		}
	} else {
		newDataSize, err = addByteDataInFile(f, footerStartOffset, updatedDataBytes, true)
		if err != nil {
			return nil, nil, err
		}
		updatedFooterOffset = footerStartOffset + newDataSize
		fileStartOffset = footerStartOffset
	}

	updatedIndexRowStr := recordToUpdateIndexRow.String()

	recordToUpdateIndexRow.ForEach(func(key, value gjson.Result) bool {
		indexFieldKey := key.String()
		if rs.Get(indexFieldKey).Exists() {
			updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, indexFieldKey, rs.Get(indexFieldKey).Value())
		}
		return true
	})
	fileHash, err := securitymdl.GetHash(updatedDataStr)
	if err != nil {
		return nil, nil, err
	}
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "startOffset", fileStartOffset)
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "dataSize", newDataSize)
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "fileHash", fileHash)
	updatedIndexRowListStr, err := updateIndexRow(&existingIndexRows, recordToUpdateIndexRow, gjson.Parse(updatedIndexRowStr))

	if err != nil {
		return nil, nil, err
	}

	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return nil, nil, err
	}

	err = setIndexDataInFile(f, updatedFooterOffset, updatedIndexRowListStr)
	if err != nil {
		return nil, nil, err
	}

	err = setFooterOffset(f, updatedFooterOffset)
	if err != nil {
		return nil, nil, err
	}

	err = setFooterSize(f, int64(len(updatedIndexRowListStr)))
	if err != nil {
		return nil, nil, err
	}

	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return nil, nil, err
	}

	err = f.Sync()
	if err != nil {
		return nil, nil, err
	}
	updatedData := gjson.Parse(updatedDataStr)
	updatedIndexRows := gjson.Parse(updatedIndexRowListStr)
	return &updatedData, &updatedIndexRows, nil
}

func initializeFile(fp *os.File) (err error) {
	filePath := fp.Name()
	isFilePresent := filemdl.FileAvailabilityCheck(filePath)
	info, err := fp.Stat()
	if err != nil {
		return
	}
	if !isFilePresent || info.Size() == 0 {
		dir, _ := filepath.Split(filePath)
		err = filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return
		}

		err = InitializeWithHeaderUsingFp(fp)
		if err != nil {
			loggermdl.LogError(err)
			return
		}
	}
	return
}

func InitializeWithHeaderUsingFp(f *os.File) error {
	_, err := f.WriteAt([]byte(strconv.Itoa(fileStatusReady)), fileStatusOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// isFile ready for upload  =0
	_, err = f.WriteAt([]byte("0"), isReadyForUploadOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isUpdatedAndNotCommitedOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReorgRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReindexRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// _, err = f.WriteAt([]byte(appendPaddingToNumber(sizeReservedForHeaders, 15)), footerOffsetInFile)
	err = setFooterOffset(f, sizeReservedForHeaders+int64(len(lineBreak)))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	err = setFooterSize(f, 0)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("filehash"), filehashOffest)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	timestamp := strconv.FormatInt(time.Now().Unix(), 10)
	_, err = f.WriteAt([]byte(timestamp), lastUpdatedOffset)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("\r\n"), sizeReservedForHeaders)
	return err
}

func getInFileIndexData(f *os.File) (gjson.Result, error) {
	infileIndexData := gjson.Parse("[]")
	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return infileIndexData, errormdl.Wrap("fail to fetch infile index data")
	}
	footerSize, err := getFooterSize(f)
	if err != nil {
		return infileIndexData, errormdl.Wrap("fail to fetch infile index size")
	}
	if footerSize == 0 {
		return infileIndexData, nil
	}
	dataByte, err := filemdl.ReadFileFromOffset(f, footerStartOffset, footerSize)
	if err != nil {
		if err.Error() == "EOF" {
			loggermdl.LogError("EOF")
			return infileIndexData, nil
		}
		loggermdl.LogError("error while fetching index data", err)
		return infileIndexData, err
	}
	infileIndexData = gjson.ParseBytes(dataByte)
	if !infileIndexData.IsArray() {
		loggermdl.LogError("invalid infile index data", infileIndexData.Value())
		return infileIndexData, errormdl.Wrap("invalid infile index data")
	}
	return infileIndexData, nil
}

func addFileDataInFile(f *os.File, offset int64, data string, breakLine bool, rs *gjson.Result, encrypter securityprovider.SecurityProvider) (int64, error) {
	dataBytes := []byte(data)
	var err error
	if encrypter != nil {
		dataBytes, err = encrypter.Encrypt(dataBytes, f.Name(), rs)
		if err != nil {
			return 0, err
		}
	}

	if breakLine {
		dataBytes = append(dataBytes, []byte(lineBreak)...)
	}

	return filemdl.WriteFileAtOffset(f, offset, dataBytes)
}

func addByteDataInFile(f *os.File, offset int64, dataBytes []byte, breakLine bool) (int64, error) {
	var err error

	if breakLine {
		dataBytes = append(dataBytes, []byte(lineBreak)...)
	}
	dataSize, err := filemdl.WriteFileAtOffset(f, offset, dataBytes)
	return dataSize, err
}

func getFileDataFromPack(f *os.File, startOffset, dataSize int64, rs *gjson.Result, decrypter securityprovider.SecurityProvider) ([]byte, error) {

	ba, err := filemdl.ReadFileFromOffset(f, startOffset, dataSize)

	if decrypter != nil {
		ba, err = decrypter.Decrypt(ba, f.Name(), rs)
		if err != nil {
			return nil, err
		}
	}
	return ba, err
}

func appendPaddingPadValue(value int64, padNumber int) string {
	no := strconv.Itoa(padNumber)
	return fmt.Sprintf("%0"+no+"d", value)
}

func getFileStatus(f *os.File) (int, error) {

	data, err := filemdl.ReadFileFromOffset(f, fileStatusOffsetInFile, 1)
	if err != nil {
		loggermdl.LogError(err)
		return -1, err
	}
	status, err := strconv.Atoi(string(data))
	return status, err
}

func getFooterOffset(f *os.File) int64 {
	data, err := filemdl.ReadFileFromOffset(f, footerOffsetInFile, footerOffsetReservedSize)
	if err != nil {
		loggermdl.LogError(err)
		return -1
	}
	footerOffset, err := strconv.Atoi(string(data))
	if err != nil {
		loggermdl.LogError("err", err)
		return -1
	}

	return int64(footerOffset)
}

func setFileStatusFlag(f *os.File, fileStatus int) error {
	status := strconv.Itoa(fileStatus)
	_, err := filemdl.WriteFileAtOffset(f, fileStatusOffsetInFile, []byte(status))
	return err
}

func setFileReadyForUploadFlag(f *os.File, isReadyToUpload bool) error {
	flagVal := strconv.FormatBool(isReadyToUpload)
	_, err := filemdl.WriteFileAtOffset(f, isReadyForUploadOffsetInFile, []byte(flagVal))
	return err
}

func setFileUpdatedAndNotCommitedFlag(f *os.File, isUpdatedAndNotCommited bool) error {
	flagVal := strconv.FormatBool(isUpdatedAndNotCommited)
	_, err := filemdl.WriteFileAtOffset(f, isUpdatedAndNotCommitedOffsetInFile, []byte(flagVal))
	return err
}

func setFileReorgRequiredFlag(f *os.File, isReorgRequired bool) error {
	flagVal := strconv.FormatBool(isReorgRequired)
	_, err := filemdl.WriteFileAtOffset(f, isReorgRequiredOffsetInFile, []byte(flagVal))
	return err
}

func setFileReindexRequiredFlag(f *os.File, isReindexRequired bool) error {
	flagVal := strconv.FormatBool(isReindexRequired)
	_, err := filemdl.WriteFileAtOffset(f, isReindexRequiredOffsetInFile, []byte(flagVal))
	return err
}

func setFooterOffset(f *os.File, footerOffset int64) error {
	footerOffestInString := appendPaddingPadValue(footerOffset, 15)
	_, err := filemdl.WriteFileAtOffset(f, footerOffsetInFile, []byte(footerOffestInString))
	return err
}

func setFooterSize(f *os.File, footerSize int64) error {
	footerSizeInString := appendPaddingPadValue(footerSize, 15)
	_, err := filemdl.WriteFileAtOffset(f, footerSizeOffset, []byte(footerSizeInString))
	return err
}

func getFooterSize(f *os.File) (int64, error) {
	data, err := filemdl.ReadFileFromOffset(f, footerSizeOffset, 15)
	if err != nil {
		return -1, err
	}
	footerSize, err := strconv.Atoi(string(data))
	if err != nil {
		loggermdl.LogError("err", err)
		return -1, err
	}

	return int64(footerSize), nil
}

func setIndexDataInFile(f *os.File, footerOffset int64, indexData string) error {
	_, err := filemdl.WriteFileAtOffset(f, footerOffset, []byte(indexData))
	return err
}
