package filetype

import (
	"errors"
	"io"
	"os"
	"path/filepath"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/securityprovider"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"
)

type SimpleFile struct {
	FilePath         string
	Fp               *os.File
	IsLazyEnable     bool
	securityProvider securityprovider.SecurityProvider
}

func NewSimpleFile(filePath string, securityProvider securityprovider.SecurityProvider) (*SimpleFile, error) {
	if filePath == "" {
		return nil, errormdl.Wrap("please provide valid filepath")
	}
	path, err := filepath.Abs(filePath)
	if err != nil {
		return nil, err
	}
	dir, _ := filepath.Split(path)
	if dir != "" {
		err := filemdl.CreateDirectoryRecursive(dir)
		if err != nil {
			return nil, err
		}
	}
	f, err := os.OpenFile(path, os.O_CREATE|os.O_RDWR, os.ModePerm)
	if err != nil {
		return nil, errormdl.Wrap("fail to open file: " + err.Error())
	}
	file := SimpleFile{
		FilePath:         filePath,
		Fp:               f,
		securityProvider: securityProvider,
	}
	return &file, nil
}

func (s *SimpleFile) Write(rs *gjson.Result) (err error) {
	dataBytes := []byte(rs.String())
	if s.securityProvider != nil {
		dataBytes, err = s.securityProvider.Encrypt(dataBytes, s.Fp.Name(), rs)
		if err != nil {
			loggermdl.LogError(err)
			return
		}
	}

	err = filemdl.WriteFileUsingFp(s.Fp, dataBytes, true, false)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	return nil
}

func (s *SimpleFile) Read(data *gjson.Result) ([]byte, error) {
	ba, err := filemdl.ReadFileUsingFp(s.Fp)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}

	if len(ba) == 0 {
		return ba, nil
	}

	if s.securityProvider != nil {
		ba, err = s.securityProvider.Decrypt(ba, s.Fp.Name(), data)
		if err != nil {
			loggermdl.LogError(err)
			return nil, err
		}
	}
	return ba, nil
}

func (s *SimpleFile) Update(rs *gjson.Result) (gjson.Result, error) {
	data, err := s.Read(rs)
	if err != nil {
		return gjson.Result{}, err
	}
	existingDataStr := string(data)
	rs.ForEach(func(key, val gjson.Result) bool {
		existingDataStr, _ = sjson.Set(existingDataStr, key.String(), val.Value())
		return true
	})

	updatedData := gjson.Parse(existingDataStr)
	err = s.Write(&updatedData)
	return updatedData, err
}

func (s *SimpleFile) Remove() error {
	if filemdl.FileAvailabilityCheck(s.Fp.Name()) {
		err := s.Fp.Truncate(0)
		if err != nil {
			return err
		}

		if _, err := s.Fp.Seek(0, io.SeekStart); err != nil {
			return err
		}
		return nil
	}
	return errors.New("not found")
}

func (s *SimpleFile) Close() error {
	return s.Fp.Close()
}
