package index

import (
	"path/filepath"
	"strings"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/securityprovider"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/corefdb/lazycache"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/lazywriter"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"github.com/tidwall/buntdb"
	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"
)

const (
	LineBreak            = "\r\n"
	IndexKeyValSeperator = "="
)

// Index - Index
type Index struct {
	indexStore       IndexStore
	IndexID          string       `json:"indexId"`
	IndexNameQuery   string       `json:"indexNameQuery"`
	BucketSequence   []string     `json:"bucketSequence"`
	IndexFields      []IndexField `json:"indexFields"`
	IsDynamicName    bool         `json:"isDynamicName"`
	IndexFilePath    string
	SecurityProvider securityprovider.SecurityProvider
}

// IndexField - IndexField
type IndexField struct {
	FieldName string `json:"fieldName"`
	Query     string `json:"query"`
}

func NewIndex(indexID, indexNameQuery string, IsDynamicName bool, indexFilePath string) (*Index, error) {
	idx := Index{
		IndexID:          indexID,
		IndexNameQuery:   indexNameQuery,
		IsDynamicName:    IsDynamicName,
		IndexFilePath:    indexFilePath,
		SecurityProvider: securityprovider.New(securityprovider.SecurityConfig{}),
	}

	var err error
	idx.indexStore, err = NewStore()
	if err != nil {
		return nil, err
	}
	err = idx.LoadIndexEntriesFromFile()
	if err != nil {
		return nil, err
	}
	lazyObj := lazywriter.LazyCacheObject{
		FileName:      indexFilePath,
		Identifier:    indexID,
		InterfaceData: idx,
		SaveFn:        lazyCallBackFnSaveIndex,
	}

	lazycache.IndexLazyObjHolder.SetNoExpiration(indexID, lazyObj)
	return &idx, nil
}

// SetFields - SetFields
func (i *Index) SetFields(indexField ...IndexField) *Index {
	i.IndexFields = indexField
	return i
}

func (i *Index) CreateIndex() error {
	var fns []func(a, b string) bool
	for _, idx := range i.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	err := i.indexStore.store.CreateIndex(i.IndexID, "*", fns...)
	if err != nil {
		return err
	}
	return nil
}

func (i *Index) ReplaceIndex() error {
	var fns []func(a, b string) bool
	for _, idx := range i.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	err := i.indexStore.store.ReplaceIndex(i.IndexID, "*", fns...)
	if err != nil {
		return err
	}
	return nil
}

func (i *Index) GetAllEntries() (map[string]string, error) {
	return i.indexStore.GetMany()
}

func (i *Index) GetEntryByQueries(queries []string) (Entry, bool, error) {
	return i.indexStore.GetOneByQuery(queries)
}

func (i *Index) GetEntriesByQueries(queries []string) (map[string]string, error) {
	return i.indexStore.GetManyByQuery(queries)
}

// DhawalD - Added GetEntriesByOrQueries - 31-July-2021
func (i *Index) GetEntriesByOrQueries(queries []string) (map[string]string, error) {
	return i.indexStore.GetManyByOrQuery(queries)
}
func (i *Index) GetEntryByPath(path string) (string, error) {
	return i.indexStore.Get(path)
}

func (i *Index) AddEntry(path string, rs *gjson.Result) error {
	json := rs.String()
	for _, indexField := range i.IndexFields {
		if rs.Get(indexField.Query).Value() == nil {
			return errormdl.Wrap("please provide value for index field: " + indexField.Query)
		}
		json, _ = sjson.Set(json, indexField.FieldName, rs.Get(indexField.Query).Value())
	}
	UpdateLazyCache(i)
	path = strings.Trim(path, string(filepath.Separator))
	return i.indexStore.Set(path, json)
}

func (i *Index) AddEntries(keyValMap map[string]string) error {
	UpdateLazyCache(i)
	return i.indexStore.AddMany(keyValMap)
}

func (i *Index) Delete(path string) error {
	UpdateLazyCache(i)
	return i.indexStore.Delete(path)
}

func (i *Index) DeleteMany(paths []string) error {
	for cnt := range paths {
		err := i.indexStore.Delete(paths[cnt])
		if err != nil {
			return err
		}
	}
	UpdateLazyCache(i)
	return nil
}

func (i *Index) CloseStore() error {
	return i.indexStore.Close()
}

func (index *Index) LoadIndexEntriesFromFile() error {

	if !filemdl.FileAvailabilityCheck(index.IndexFilePath) {
		return nil
	}
	fileData, err := filemdl.FastReadFile(index.IndexFilePath)
	if err != nil {
		loggermdl.LogError("failed to load FDB index from: ", index.IndexFilePath)
		return err
	}
	if len(fileData) == 0 {
		return nil
	}
	_, fileName := filepath.Split(index.IndexFilePath)
	fileData, err = index.SecurityProvider.Decrypt(fileData, fileName, nil)
	if err != nil {
		loggermdl.LogError("failed to decrypt FDB index data: ", err)
		return errormdl.Wrap("failed to decrypt FDB index data: " + err.Error())
	}
	data := string(fileData)
	indexRecords := strings.Split(data, LineBreak)
	indexDataMap := make(map[string]string)
	for _, indexRecord := range indexRecords {
		indexValues := strings.Split(indexRecord, IndexKeyValSeperator)
		if len(indexValues) == 2 {
			indexDataMap[indexValues[0]] = indexValues[1]
		}
	}
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	// update index file by reading all data and updating index file
	return index.AddEntries(indexDataMap)
}

// LogFDBIndexFile -LogFDBIndexFile
func (index *Index) WriteIndexEntriesInFile() error {
	dataToStore := ``
	indeKeyValMap, err := index.GetAllEntries()
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	for key, value := range indeKeyValMap {
		dataToStore = dataToStore + key + IndexKeyValSeperator + value + LineBreak
	}
	_, fileName := filepath.Split(index.IndexFilePath)
	var dataByteToWriteRes = []byte{}
	var hashError error
	if len(indeKeyValMap) > 0 {
		dataByteToWriteRes, hashError = index.SecurityProvider.Encrypt([]byte(dataToStore), fileName, nil)
		if errormdl.CheckErr1(hashError) != nil {
			return errormdl.CheckErr1(hashError)
		}
	}
	return filemdl.WriteFile(index.IndexFilePath, dataByteToWriteRes, true, false)
}

var lazyCallBackFnSaveIndex lazywriter.SaveDataFn = func(indexID string, data *lazywriter.LazyCacheObject) {
	index, ok := data.InterfaceData.(*Index)
	if !ok {
		return
	}

	err := index.WriteIndexEntriesInFile()
	if err != nil {
		loggermdl.LogError(err)
		return
	}
}

// UpdateLazyCache - updates index data in lay writer cache
func UpdateLazyCache(index *Index) error {
	// lazy cache must be present for provided indexID
	lazyObj, ok := lazycache.IndexLazyObjHolder.Get(index.IndexID)
	if !ok {
		loggermdl.LogError("index not found in lazy writer cache")
		return errormdl.Wrap("index not found in lazy writer cache")
	}

	idxLazyData, ok := lazyObj.(lazywriter.LazyCacheObject)
	if !ok {
		loggermdl.LogError("interface type is not lazywriter.LazyCacheObject")
		return errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
	}

	// idxLazyData.GJSONData = index
	idxLazyData.InterfaceData = index
	if ok := lazycache.IndexMaster.SaveOrUpdateDataInCache(idxLazyData); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return errormdl.Wrap("failed to update index data in lazy cache")
	}

	lazycache.IndexLazyObjHolder.SetNoExpiration(index.IndexID, idxLazyData)
	return nil
}
