package index

import (
	"path/filepath"
	"strings"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"

	"github.com/tidwall/buntdb"
	"github.com/tidwall/gjson"
)

type IndexStore struct {
	store *buntdb.DB
}

type Entry struct {
	Key   string
	Value string
}

// NewStore - returns new store object
func NewStore() (IndexStore, error) {
	db, err := buntdb.Open(":memory:")
	if err != nil {
		return IndexStore{}, err
	}
	store := IndexStore{
		store: db,
	}
	return store, nil
}

func (i IndexStore) Close() error {
	return i.store.Close()
}
func (i *IndexStore) Set(key, value string) (err error) {
	err = i.store.Update(func(tx *buntdb.Tx) error {
		key = strings.ReplaceAll(key, "\\", "/")
		_, _, err := tx.Set(key, value, nil)
		if err != nil {
			return err
		}
		return nil
	})
	return
}

func (i *IndexStore) AddMany(keyValMap map[string]string) (err error) {
	err = i.store.Update(func(tx *buntdb.Tx) error {
		for key, val := range keyValMap {
			key = strings.ReplaceAll(key, "\\", "/")
			_, _, err := tx.Set(key, val, nil)
			if err != nil {
				return err
			}
		}
		return nil
	})
	return
}

func (i *IndexStore) Get(path string) (val string, err error) {
	path = strings.ReplaceAll(path, "\\", "/")

	found := false
	err = i.store.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			if key == path && !found {
				val = value
				found = true
				return true
			}
			return true
		})
	})
	return
}

func (i *IndexStore) GetOneByQuery(queries []string) (entry Entry, found bool, err error) {
	err = i.store.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(queries); i++ {
				rsJSON = rsJSON.Get(queries[i] + "#")
			}
			if rsJSON.Value() != nil && rsJSON.IsArray() && len(rsJSON.Array()) > 0 && !found {
				found = true
				entry = Entry{
					Key:   filepath.Join(key),
					Value: value,
				}

				return true
			}
			return true
		})
	})
	return
}

func (i *IndexStore) GetManyByQuery(queries []string) (map[string]string, error) {
	entryMap := make(map[string]string, 0)
	err := i.store.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(queries); i++ {
				rsJSON = rsJSON.Get(queries[i] + "#")
			}
			if rsJSON.Value() != nil && rsJSON.IsArray() && len(rsJSON.Array()) > 0 {
				key = filepath.Join(key)
				entryMap[key] = value
			}
			return true
		})
	})
	return entryMap, err
}

// GetManyByOrQuery - For getting many data by OR entry
// DhawalD - Added GetManyByOrQuery - 31-July-2021
func (i *IndexStore) GetManyByOrQuery(queries []string) (map[string]string, error) {
	entryMap := make(map[string]string, 0)
	err := i.store.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			var rsJSONres gjson.Result
			for i := 0; i < len(queries); i++ {
				rsJSONres = rsJSON.Get(queries[i] + "#")
				if rsJSONres.Value() != nil && rsJSONres.IsArray() && len(rsJSONres.Array()) > 0 {
					key = filepath.Join(key)
					entryMap[key] = value
				}
			}
			return true
		})
	})
	return entryMap, err
}

func (i *IndexStore) GetMany() (map[string]string, error) {
	entryMap := make(map[string]string, 0)
	err := i.store.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			key = filepath.Join(key)
			entryMap[key] = value
			return true
		})
	})

	return entryMap, err
}

func (i *IndexStore) Delete(key string) error {
	key = strings.ReplaceAll(key, "\\", "/")
	err := i.store.Update(func(tx *buntdb.Tx) error {
		_, err := tx.Delete(key)
		if err != nil {
			return err
		}
		return nil
	})
	return err
}

// GenRowID generates hash for the given filename. The length of hash is 16
func GenRowID(name string) (string, error) {
	name = strings.ReplaceAll(filepath.Clean(name), string(filepath.Separator), "")
	rowID, err := securitymdl.GetHash(name)
	if err != nil {
		return "", err
	}

	if len(rowID) > 16 {
		return rowID[:16], nil
	}

	return rowID, nil
}
