package securityprovider

import (
	"errors"
	"path/filepath"
	"strings"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/hashmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"
	"github.com/tidwall/gjson"
)

const (
	// SharedPrefix prefix represents the file is sharable i.e encryption key does not include (fdbSec.fieldQuery)
	SharedPrefix = "ss_"
	// EmptySTR represents empty string
	EmptySTR = ""
)

type SecurityProvider interface {
	Encrypter
	Decrypter
}

type Encrypter interface {
	Encrypt([]byte, string, *gjson.Result) ([]byte, error)
}

type Decrypter interface {
	Decrypt([]byte, string, *gjson.Result) ([]byte, error)
}

type FdbSecurityProvider struct {
	encKey         string // the global encryption key used in the project. This key will be applicable in all cases.
	userDefinedKey string // the user defined key in the project. This key will be applicable in all cases.
	fieldQuery     string // query to get dynamic field. Ex. Each student data can be encrypted with studentID. Applicable only for the shared bucket.
}

type SecurityConfig struct {
	EncKey         string // the global encryption key used in the project. This key will be applicable in all cases.
	UserDefinedKey string // the user defined key in the project. This key will be applicable in all cases.
	FieldQuery     string // query to get dynamic field. Ex. Each student data can be encrypted with studentID. Applicable only for the shared bucket.
}

func New(config SecurityConfig) FdbSecurityProvider {

	fdbSecurity := FdbSecurityProvider{
		encKey:         config.EncKey,
		userDefinedKey: config.UserDefinedKey,
		fieldQuery:     config.FieldQuery,
	}

	return fdbSecurity
}

func (fs FdbSecurityProvider) GenerateSecurityKey(fileName string, data *gjson.Result) (key []byte, err error) {
	_, fileName = filepath.Split(fileName)
	if fileName == EmptySTR {
		return key, errors.New("GenerateSecurityKey: fileName must not be empty")
	}

	// Warning: The order of string concatenation must be preserved as specified.
	skey := fs.encKey + fs.userDefinedKey + fileName
	if !strings.HasPrefix(fileName, SharedPrefix) && fs.fieldQuery != "" {
		// this is a shared file OR no query provided. No need to check for dynamic query result.
		if data == nil || data.Get(fs.fieldQuery).String() == EmptySTR {
			return key, errormdl.Wrap("please provide value of field: " + fs.fieldQuery)
		}
		skey = data.Get(fs.fieldQuery).String() + skey
	}

	// loggermdl.LogDebug("key", string(skey))
	hash, err := hashmdl.Get128BitHash([]byte(skey))
	if err != nil {
		return key, errors.New("GenerateSecurityKey: " + err.Error())
	}
	key = hash[:]
	return
}

// Encrypt - encrypts provide data
func (fs FdbSecurityProvider) Encrypt(dataByte []byte, fileName string, data *gjson.Result) (res []byte, err error) {
	res, err = filemdl.ZipBytes(dataByte)
	if err != nil {
		return
	}
	securityKey, gerr := fs.GenerateSecurityKey(fileName, data)
	if gerr != nil {
		err = gerr
		return
	}
	return securitymdl.AESEncrypt(res, securityKey)
}

// Decrypt - decrypts provide data
func (fs FdbSecurityProvider) Decrypt(dataByte []byte, fileName string, data *gjson.Result) (res []byte, err error) {
	res = dataByte
	securityKey, gerr := fs.GenerateSecurityKey(fileName, data)
	if gerr != nil {
		err = gerr
		return
	}
	res, err = securitymdl.AESDecrypt(res, securityKey)
	if err != nil {
		return
	}

	return filemdl.UnZipBytes(res)
}
