package coreimmudb

import (
	"context"
	"encoding/json"
	"os"
	"path/filepath"
	"strings"
	"sync"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"github.com/codenotary/immudb/pkg/api/schema"
	"github.com/codenotary/immudb/pkg/auth"
	immudb "github.com/codenotary/immudb/pkg/client"
)

const (
	AdminPermission     = auth.PermissionAdmin
	ReadPermission      = auth.PermissionR
	ReadWritePermission = auth.PermissionRW

	defaultLimit = 100
)

var (
	instances       map[string]immudb.ImmuClient // Immudb client connection with host
	DefaultHostName string                       // Holds default hostname
	mutex           sync.Mutex
)

type ImmuHost struct {
	HostName       string `json:"hostName"`       // Connection name
	Server         string `json:"server"`         // Database server address
	Port           int    `json:"port"`           // Database port
	Database       string `json:"database"`       // Database name
	Username       string `json:"username"`       // Database username
	Password       string `json:"password"`       // Database password
	Dir            string `json:"dir"`            // Storing .identity (database server identity) and .state file
	MaxRecvMsgSize int    `json:"maxRecvMsgSize"` // Recieve message maximum size in MB (Default: 4MB)
	IsDefault      bool   `json:"isDefault"`      // Default ImmuDB connection
	IsDisabled     bool   `json:"isDisabled"`     // If true connection then connection to database server is skipped
}

type ImmuDAO struct {
	HostName string
	MetaData MetaData
}
type MetaData struct {
	GetOpts  GetOptions
	ScanOpts schema.ScanRequest
}
type GetOptions struct {
	AtTx       uint64 `json:"atTx"`
	SinceTx    uint64 `json:"sinceTx"`
	NoWait     bool   `json:"noWait"`
	AtRevision int64  `json:"atRevision"`
}

// For setting multiple keys - to know more see SetAllKeysData
type KeyValue struct {
	Key   []byte `json:"key"`
	Value []byte `json:"value"`
}

func init() {
	instances = make(map[string]immudb.ImmuClient)
}

// InitUsingJsonFile - Initializes connection to ImmuDB server using json config file
func InitUsingJson(filePath string) error {

	// Reading ImmuDB config file
	byteData, err := os.ReadFile(filePath)
	if err != nil {
		return err
	}
	// Binding byte data into []ImmuHost{}
	var hosts []ImmuHost
	err = json.Unmarshal(byteData, &hosts)
	if err != nil {
		return err
	}

	instances = make(map[string]immudb.ImmuClient)

	for _, host := range hosts {
		if host.IsDisabled { // Not connecting to database if isDisabled == true
			continue
		}
		err := AddDBConnection(host)
		if err != nil {
			instances = make(map[string]immudb.ImmuClient)
			return err
		}
	}
	return nil
}

// InitNewImmuDbSession - Initializes connection to ImmuDB server using []ImmuHost data
func InitImmuDBConnections(hosts []ImmuHost) error {
	instances = make(map[string]immudb.ImmuClient)

	for _, host := range hosts {
		if host.IsDisabled { // Not connecting to database if isDisabled == true
			continue
		}

		err := AddDBConnection(host)
		if err != nil {
			instances = make(map[string]immudb.ImmuClient)
			return err
		}
	}
	return nil
}

// AddDBConnection - Initializes single connection of ImmuDB server using ImmuHost data
func AddDBConnection(host ImmuHost) error {
	mutex.Lock()
	defer mutex.Unlock()

	var client immudb.ImmuClient
	opts := immudb.DefaultOptions()
	opts.Address = host.Server
	opts.Port = host.Port
	if host.Dir != "" {
		// Checking if folder name contains extension
		if filepath.Ext(host.Dir) != "" {
			return errormdl.Wrap("DIRECTORY NAME CONTAINS EXTENSION " + filepath.Ext(host.Dir))
		}
		path := strings.Split(host.Dir, "/")

		// Hiding folder by adding prefix '.'
		if !strings.HasPrefix(path[len(path)-1], ".") {
			path[len(path)-1] = "." + path[len(path)-1]
		}

		host.Dir = strings.Join(path, "/")
		if _, err := os.Stat(host.Dir); os.IsNotExist(err) { // Checking directory is present and accessible
			if err = os.Mkdir(host.Dir, 0740); err != nil {
				return err
			}
		}
		opts.Dir = host.Dir
	}

	if host.MaxRecvMsgSize > 0 { // Setting maximum receive message size if provided
		opts.MaxRecvMsgSize = host.MaxRecvMsgSize * 1024 * 1024
	}

	client = immudb.NewClient().WithOptions(opts)
	err := client.OpenSession(context.TODO(), []byte(host.Username), []byte(host.Password), host.Database)
	if err != nil {
		return err
	}
	instances[host.HostName] = client
	if host.IsDefault {
		DefaultHostName = host.HostName
	}

	return nil
}

// DeleteImmuDbSession - Disconnecting Immu Database connection
func DeleteImmuDbSession(hostName string) error {
	mutex.Lock()
	defer mutex.Unlock()
	inst, ok := instances[hostName]
	if !ok {
		return errormdl.Wrap("CONNECTION NOT FOUND FOR " + hostName)
	}
	delete(instances, hostName)

	return inst.CloseSession(context.TODO()) // Closing immudb connection if there are any errors then returning.
}

// TestImmuDbConnection - Connects to Immudb with provided Database host details, if failed then error is returned or else nil.
func TestDbConnection(host ImmuHost) error {
	var client immudb.ImmuClient

	opts := immudb.DefaultOptions()
	opts.Address = host.Server
	opts.Port = host.Port
	if host.Dir != "" {
		// Checking if folder name contains extension
		if filepath.Ext(host.Dir) != "" {
			return errormdl.Wrap("DIRECTORY NAME CONTAINS EXTENSION: " + filepath.Ext(host.Dir))
		}
	}
	client = immudb.NewClient().WithOptions(opts)
	err := client.OpenSession(context.TODO(), []byte(host.Username), []byte(host.Password), host.Database)
	if err != nil {
		return err
	}
	defer client.CloseSession(context.TODO())
	if ok := client.IsConnected(); !ok {
		return errormdl.Wrap("ImmuDB is not connected")
	}
	return nil // Sending true if connection successfully established to ImmuDB
}

// Get ImmuDB connection - immudb.ImmuClient
func GetImmuDbConnection(hostName string) (immudb.ImmuClient, error) {
	mutex.Lock()
	defer mutex.Unlock()
	// Returning if there are no immudb connections present
	if instances == nil {
		return nil, errormdl.Wrap("IMMUDB CONNECTION NOT FOUND FOR " + hostName)
	}

	// If hostname is empty
	if hostName == "" {
		hostName = DefaultHostName
	}

	client, ok := instances[hostName]
	if !ok {
		return nil, errormdl.Wrap("CONNECTION NOT FOUND FOR " + hostName)
	}

	// Checking Database connection
	if ok := client.IsConnected(); !ok {
		return nil, errormdl.Wrap(hostName + "CLIENT NOT CONNECTED TO IMMUDB SERVER")
	}

	return client, nil
}

/*
GetImmuDAOWithHost - return Immu DAO instance for provided host

	For default ImmuDB DAO:
		Pass empty string - coreimmudb.GetImmuDAOWithHost("")
									OR
		Use DefaultHostName - coreimmudb.DefaultHostName
*/
func GetImmuDAOWithHost(hostname string) *ImmuDAO {
	return &ImmuDAO{
		HostName: hostname,
	}
}

/* READ OPERATIONS */

/*
GetKeyData - Get single value for provided key

Use GetOptions for advance filtering as show below;

immuDAO.MetaData.GetOpts = coreimmudb.GetOptions{AtTx: 0, AtRevision: 0, SinceTx: 0, NoWait: true}

NOTE: 'immuDAO' is Data access object; To get Data access object use coreimmudb.GetImmuDAOWithHost("hostname")
*/
func (i *ImmuDAO) GetKeyData(key []byte) (*schema.Entry, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}
	var opts []immudb.GetOption
	if i.MetaData.GetOpts != (GetOptions{}) {
		opts = prepareOptions(i.MetaData.GetOpts)
	}
	return client.Get(context.TODO(), key, opts...)
}

/*
GetVerifiedKeyData - Get single value for provided key with additional server-provided proof validation.

Use GetOptions for advance filtering as show below;

immuDAO.MetaData.GetOpts = coreimmudb.GetOptions{AtTx: 0, AtRevision: 0, SinceTx: 0, NoWait: true}

NOTE: 'immuDAO' is Data access object; To get Data access object use coreimmudb.GetImmuDAOWithHost("hostname")
*/
func (i *ImmuDAO) GetVerifiedKeyData(key []byte) (*schema.Entry, error) {

	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}
	var opts []immudb.GetOption
	if i.MetaData.GetOpts != (GetOptions{}) {
		opts = prepareOptions(i.MetaData.GetOpts)
	}
	return client.VerifiedGet(context.TODO(), key, opts...)
}

// GetAllKeysData - Getting all keys data
/*
 GetAllKeysData retrives value for provided keys,
 values will be nil, for those keys which are not present in immudb
*/
func (i *ImmuDAO) GetAllKeysData(keys [][]byte) (*schema.Entries, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}
	result, err := client.GetAll(context.TODO(), keys)
	if err != nil {
		return nil, err
	}

	mapEntries := make(map[string]interface{})
	for _, val := range result.Entries {
		mapEntries[string(val.Key)] = val.Value
	}

	for _, key := range keys {
		if _, ok := mapEntries[string(key)]; !ok {
			result.Entries = append(result.Entries, &schema.Entry{Key: key, Value: nil})
		}
	}
	return result, nil
}

// GetKeyHistory - Get history of single key
/*
 Offset - Number of entries to skip
 Limit - Maximum numbers of entries to return
 Desc - Sorting in descending order
*/
func (i *ImmuDAO) GetKeyHistory(key []byte, offset uint64, limit int32, desc bool) (*schema.Entries, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}

	if limit <= 0 {
		limit = defaultLimit // Setting to defaultLimit
	}

	req := &schema.HistoryRequest{
		Key:    key,
		Offset: offset,
		Limit:  limit,
		Desc:   desc,
	}
	return client.History(context.TODO(), req)
}

/*
GetActiveKeys - List all avaliable keys from active database, this function does not retrive deleted keys

- Use scan options for advance filtering as shown below

	immuDAO.MetaData.ScanOpts = schema.ScanRequest{Desc: true, Limit: 10}
	NOTE: 'immuDAO' is Data access object; To get Data access object use coreimmudb.GetImmuDAOWithHost("hostname")

Scan options field description:

	SeekKey - List all keys which were inserted/saved after the specified key, include specified key in result if InclusiveSeek = True
	EndKey - List all keys which were inserted/saved before the specified key, include specified key in result if InclusiveEnd = True
	Prefix - List all keys which are perfixed with given character(s)
	Desc - Sorting in descending order
	Limit - Maximum numbers of entries to return
	SinceTx - Used to avoid waiting for the indexer to be up-to-date with the most recent transaction
	NoWait (True/False) - If set to true, then the server will not wait for the indexer to	be up-to-date with the most recent transaction
	InclusiveSeek (True/False)
	InclusiveEnd (True/False)
	Offset - Number of entries to skip
*/
func (i *ImmuDAO) GetActiveKeys() (*schema.Entries, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection
	if err != nil {
		return nil, err
	}

	// Scan iterates over the set of keys in a topological order
	return client.Scan(context.TODO(), &i.MetaData.ScanOpts)
}

/* END OF READ OPERATIONS */

/* WRITE OPERATIONS */

// SetKeyData - Commits provided value for provided key
func (i *ImmuDAO) SetKeyData(key, value []byte) (*schema.TxHeader, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}
	return client.Set(context.TODO(), key, value)
}

// SetVerifiedKeyData - Commits provided value for provided key and also requests a server-generated proof
func (i *ImmuDAO) SetVerifiedKeyData(key, value []byte) (*schema.TxHeader, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}

	return client.VerifiedSet(context.TODO(), key, value)
}

// SetAllKeysData - Set multiple keys with their values in a single transaction.
func (i *ImmuDAO) SetAllKeysData(kv []KeyValue) (*schema.TxHeader, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}

	var kvs []*schema.KeyValue
	for _, v := range kv {
		kvs = append(kvs, &schema.KeyValue{Key: v.Key, Value: v.Value})
	}
	return client.SetAll(context.TODO(), &schema.SetRequest{KVs: kvs})
}

// DeleteKey - Key is deleted logically from Database, physically it is present in database
func (i *ImmuDAO) DeleteKey(keys [][]byte) (*schema.TxHeader, error) {
	client, err := GetImmuDbConnection(i.HostName) // ImmuDB connection for the provided host
	if err != nil {
		return nil, err
	}

	req := &schema.DeleteKeysRequest{
		Keys: keys,
	}
	return client.Delete(context.TODO(), req)
}

// ExpirableKeySet - Sets value for the key with expiration time
func (i *ImmuDAO) ExpirableKeySet(key, value []byte, expiresAt time.Time) (*schema.TxHeader, error) {
	client, err := GetImmuDbConnection(i.HostName)
	if err != nil {
		return nil, err
	}
	return client.ExpirableSet(context.TODO(), key, value, expiresAt)
}

/* END OF WRITE OPERATIONS */

// CreateDatabase - Creates new database in ImmuDB, current database user requires SysAdmin permission level.
func (i *ImmuDAO) CreateDatabase(dbName string, dbSettings *schema.DatabaseNullableSettings) (*schema.CreateDatabaseResponse, error) {
	client, err := GetImmuDbConnection(i.HostName)
	if err != nil {
		return nil, err
	}
	return client.CreateDatabaseV2(context.TODO(), dbName, dbSettings)
}

// DeleteDatabase - Removes existing database from ImmuDB
func (i *ImmuDAO) DeleteDatabase(dbName string) (*schema.DeleteDatabaseResponse, error) {
	client, err := GetImmuDbConnection(i.HostName)
	if err != nil {
		return nil, err
	}
	// Unloading database before deletion
	_, err = client.UnloadDatabase(context.TODO(), &schema.UnloadDatabaseRequest{
		Database: dbName,
	})
	if err != nil && !strings.ContainsAny(err.Error(), "already closed") {
		return nil, err
	}

	return client.DeleteDatabase(context.TODO(), &schema.DeleteDatabaseRequest{
		Database: dbName,
	})
}

// ListDatabases - returns a list of databases the user has access.
func (i *ImmuDAO) ListDatabases() (*schema.DatabaseListResponseV2, error) {
	client, err := GetImmuDbConnection(i.HostName)
	if err != nil {
		return nil, err
	}

	return client.DatabaseListV2(context.TODO())
}

/*
CreateUser - Creates new user

username - Username of the new user
password - Password of the new user
(Password must have between 8 and 32 letters, digits and special characters of which at least 1 uppercase letter, 1 digit and 1 special character.)
dbName - Database name

permission:
1 (coreimmudb.ReadPermission) - read-only access
2 (coreimmudb.ReadWritePermission) - read-write access
254 (coreimmudb.AdminPermission) - read-write with admin rights
*/
func (i *ImmuDAO) CreateUser(username, password, dbName string, permission int) error {
	client, err := GetImmuDbConnection(i.HostName)
	if err != nil {
		return err
	}
	return client.CreateUser(context.TODO(), []byte(username), []byte(password), uint32(permission), dbName)
}

/*
ListUsers - List all database of user

This call requires Admin or SysAdmin permission level.

When called as a SysAdmin user, all users in the database are returned. When called as an Admin user, users for currently selected database are returned.
*/
func (i *ImmuDAO) ListUsers() (*schema.UserList, error) {
	client, err := GetImmuDbConnection(i.HostName)
	if err != nil {
		return nil, err
	}
	return client.ListUsers(context.TODO())
}

func prepareOptions(getOpts GetOptions) []immudb.GetOption {
	var opts []immudb.GetOption
	if getOpts.AtTx != 0 {
		opts = append(opts, immudb.AtTx(getOpts.AtTx))
	}
	if getOpts.SinceTx != 0 {
		opts = append(opts, immudb.SinceTx(getOpts.SinceTx))
	}
	if getOpts.NoWait {
		opts = append(opts, immudb.NoWait(getOpts.NoWait))
	}
	if getOpts.AtRevision != 0 {
		opts = append(opts, immudb.AtRevision(getOpts.AtRevision))
	}

	return opts
}
