package dgraph

import (
	"context"
	"testing"
)

var dgraphHost = &Host{
	Name:   "DGraphHost",
	Server: "localhost",
	Port:   9080,
}

func Test_NewClient(t *testing.T) {
	type args struct {
		h Host
	}
	tests := []struct {
		name string
		args args
		// want    *dgo.Dgraph
		wantErr bool
	}{
		{
			name:    "success on valid connection",
			args:    args{h: Host{Name: "graphDBHost", Server: dgraphHost.Server, Port: 9080}},
			wantErr: false,
		},
		{
			name:    "fail on connection fail",
			args:    args{h: Host{Name: "graphDBHost", Server: dgraphHost.Server, Port: 8080}},
			wantErr: true,
		},
		{
			name:    "success on default port used",
			args:    args{h: Host{Name: "graphDBHost", Server: dgraphHost.Server}},
			wantErr: false,
		},
		{
			name:    "fail on blank address",
			args:    args{h: Host{Name: "graphDBHost", Server: ""}},
			wantErr: true,
		},
		{
			name:    "fail on invalid address",
			args:    args{h: Host{Name: "graphDBHost", Server: "10.1.0"}},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			_, err := NewClient(tt.args.h)
			if (err != nil) != tt.wantErr {
				t.Errorf("NewClient() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
		})
	}
}

func Test_CreateSchema(t *testing.T) {
	type args struct {
		c      context.Context
		schema string
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "create schema successfully",
			args: args{c: context.Background(), schema: `
			 name: string @index(exact) .
			 age: int .
			`},
			wantErr: false,
		},
		{
			name: "pass invalid schema",
			args: args{c: context.Background(), schema: `
			name string @index(exact) .
			age int .
			 `},
			wantErr: true,
		},
		{
			name:    "pass blank schema",
			args:    args{c: context.Background(), schema: ``},
			wantErr: true,
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			err := InitInstances([]Host{*dgraphHost})
			if err != nil {
				t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
			}

			dg, err := GetDAO(dgraphHost.Name)
			if (err != nil) != tt.wantErr {
				t.Errorf("CreateSchema() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			err = dg.CreateSchema(tt.args.c, tt.args.schema)
			if (err != nil) != tt.wantErr {
				t.Errorf("CreateSchema() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
		})
	}
}

func Test_SetData(t *testing.T) {
	type args struct {
		c    context.Context
		data []byte
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "success on correct data",
			args: args{c: context.Background(), data: []byte(`{
				"name": "Person 1",
				"age": 29,
				"follows": {
					"name": "Person 2",
					"age": 18,
					"follows": {
						"name": "Person 3",
						"age": 37
					}
				}
			}`)},
			wantErr: false,
		},
		{
			name:    "failure on incorrect data",
			args:    args{c: context.Background(), data: []byte(``)},
			wantErr: true,
		},
	}

	for _, tt := range tests {
		err := InitInstances([]Host{*dgraphHost})
		if err != nil {
			t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
		}

		dg, err := GetDAO(dgraphHost.Name)
		if err != nil && !tt.wantErr {
			t.Errorf("SetData() error = %v", err)
		}

		err = dg.SetData(tt.args.c, tt.args.data)
		if err != nil && !tt.wantErr {
			t.Errorf("SetData() error = %v", err)
		}
	}
}

func Test_DropAttr(t *testing.T) {
	type args struct {
		c    context.Context
		attr string
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name:    "success case for delete known attribute",
			args:    args{c: context.Background(), attr: "age"},
			wantErr: false,
		},
		{
			name:    "fail case for deleting absent attribute",
			args:    args{c: context.Background(), attr: "height"},
			wantErr: true,
		},
		{
			name:    "fail case for blank attribute",
			args:    args{c: context.Background(), attr: ""},
			wantErr: true,
		},
	}

	for _, tt := range tests {
		err := InitInstances([]Host{*dgraphHost})
		if err != nil {
			t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
		}

		dg, err := GetDAO(dgraphHost.Name)
		if err != nil && !tt.wantErr {
			t.Errorf("DropAttr() error = %v", err)
		}

		err = dg.DropAttr(tt.args.c, tt.args.attr)
		if err != nil && !tt.wantErr {
			t.Errorf("DropAttr() error = %v", err)
		}
	}
}

func Test_GetData(t *testing.T) {
	type args struct {
		c     context.Context
		query string
		vars  map[string]string
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "success query to fetch data without params",
			args: args{c: context.Background(), query: `
			{
				people(func: has(name)) {
				  name
				  age,
				  follows
				}
			  }
			`},
			wantErr: false,
		},
		{
			name:    "failure on blank query",
			args:    args{c: context.Background(), query: ""},
			wantErr: true,
		},
	}

	for _, tt := range tests {
		err := InitInstances([]Host{*dgraphHost})
		if err != nil {
			t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
		}

		dg, err := GetDAO(dgraphHost.Name)
		if err != nil && !tt.wantErr {
			t.Errorf("GetData() error = %v", err)
		}

		_, err = dg.GetData(tt.args.c, tt.args.query, tt.args.vars)
		if err != nil && !tt.wantErr {
			t.Errorf("GetData() error = %v", err)
		}
	}
}

func Test_DeleteEdge(t *testing.T) {
	type args struct {
		c          context.Context
		uid        string
		predicates []string
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name:    "success case to delete an edge",
			args:    args{c: context.Background(), uid: "0x754f", predicates: []string{"follows"}},
			wantErr: false,
		},
		{
			name:    "fail case to delete blank edge",
			args:    args{c: context.Background(), uid: "0x7551", predicates: []string{""}},
			wantErr: true,
		},
		{
			name:    "fail case to delete blank UID",
			args:    args{c: context.Background(), uid: "", predicates: []string{""}},
			wantErr: true,
		},
	}

	for _, tt := range tests {
		err := InitInstances([]Host{*dgraphHost})
		if err != nil {
			t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
		}

		dg, err := GetDAO(dgraphHost.Name)
		if err != nil && !tt.wantErr {
			t.Errorf("DeleteEdge() error = %v", err)
		}

		err = dg.DeleteEdge(tt.args.c, tt.args.uid, tt.args.predicates...)
		if err != nil && !tt.wantErr {
			t.Errorf("DeleteEdge() error = %v", err)
		}
	}
}

func Test_DeleteData(t *testing.T) {
	type args struct {
		c    context.Context
		data []byte
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name: "success on delete correct data",
			args: args{c: context.Background(), data: []byte(`{
				"uid": "0x754f"
			}`)},
			wantErr: false,
		},
		{
			name:    "failure on incorrect delete data",
			args:    args{c: context.Background(), data: []byte(``)},
			wantErr: true,
		},
	}

	for _, tt := range tests {
		err := InitInstances([]Host{*dgraphHost})
		if err != nil {
			t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
		}

		dg, err := GetDAO(dgraphHost.Name)
		if err != nil && !tt.wantErr {
			t.Errorf("DeleteData() error = %v", err)
		}

		err = dg.DeleteData(tt.args.c, tt.args.data)
		if err != nil && !tt.wantErr {
			t.Errorf("DeleteData() error = %v", err)
		}
	}
}
func Test_DropData(t *testing.T) {
	type args struct {
		c context.Context
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name:    "Drop data case",
			args:    args{c: context.Background()},
			wantErr: false,
		},
	}

	for _, tt := range tests {
		err := InitInstances([]Host{*dgraphHost})
		if err != nil {
			t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
		}

		dg, err := GetDAO(dgraphHost.Name)
		if err != nil {
			t.Errorf("DropData() error = %v, wantErr %v", err, tt.wantErr)
		}

		err = dg.DropData(tt.args.c)
		if err != nil {
			t.Errorf("DropData() error = %v, wantErr %v", err, tt.wantErr)
		}
	}
}

func Test_DropSchema(t *testing.T) {
	type args struct {
		c context.Context
	}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{
			name:    "Drop schema case",
			args:    args{c: context.Background()},
			wantErr: false,
		},
	}

	for _, tt := range tests {
		err := InitInstances([]Host{*dgraphHost})
		if err != nil {
			t.Errorf("InitInstances() error = %v, wantErr %v", err, tt.wantErr)
		}

		dg, err := GetDAO(dgraphHost.Name)
		if err != nil && !tt.wantErr {
			t.Errorf("DropSchema() error = %v", err)
		}

		err = dg.DropSchema(tt.args.c)
		if err != nil && !tt.wantErr {
			t.Errorf("DropSchema() error = %v", err)
		}
	}
}
