// Author: Deepak Prakash [17-July-2018]
// Description: Package downloadhelpermdl will help download files to destination location.

package downloadhelpermdl

// All required imports
import (
	"io"
	"net/http"
	"os"
	"path"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/hashmdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
)

// Constants Required
const (
	EMPTY_STR                       = ""
	HTTP_PROTOCOL                   = "http"
	FILE_PRESENT_CHECK_ERROR        = 0
	FILE_PRESENT_AT_DESTINATION     = 1
	FILE_NOT_PRESENT_AT_DESTINATION = 2
	WAIT_TIME_DURATION              = 2000
	MIN_FILE_SIZE                   = 1
)

//TODO: 1. Add Skip List
//TODO: 2. Lenghth checking
//TODO: 3. HTTP Pooling
//TODO: 4. HTTP2 Support - Certificate
//TODO: 5. Resume Download

// DownloadHelper - Struct
type DownloadHelper struct {
	sourceURL       string
	destinationPath string
	retryCount      int
	DownloadParam   DownloadParameter
}

// Structure DownloadParameter: with different parameters which can be received for downloading the file
type DownloadParameter struct {
	jwtToken        string
	isCacheBurst    bool
	fileHashedValue string
	requestTimeOut  int64
	DownloadError   error
}

// GetDownloadHelper Gives you download helper from where you can run steps
func GetDownloadHelper(sourceURL, destinationPath string, retryCount int) *DownloadHelper {
	newDownloadHelper := DownloadHelper{}
	newDownloadHelper.sourceURL = sourceURL
	newDownloadHelper.destinationPath = destinationPath
	newDownloadHelper.retryCount = retryCount
	return &newDownloadHelper
}

// AddParamsJWT - method to add jwtToken at runtime
func (dh *DownloadHelper) AddParamsJWT(jwtToken string) *DownloadHelper {
	dh.DownloadParam.jwtToken = jwtToken
	return dh
}

// AddParamsFileHashed - method to add fileHashedValue at runtime
func (dh *DownloadHelper) AddParamsFileHashed(fileHashedValue string) *DownloadHelper {
	dh.DownloadParam.fileHashedValue = fileHashedValue
	return dh
}

// AddParamsFileHashed - method to add fileHashedValue at runtime
func (dh *DownloadHelper) AddParamsCacheBurst(isCacheBurst bool) *DownloadHelper {
	dh.DownloadParam.isCacheBurst = isCacheBurst
	return dh
}

func (dh *DownloadHelper) AddRequestTimeout(requestTimeOutValue int64) *DownloadHelper {
	dh.DownloadParam.requestTimeOut = requestTimeOutValue
	return dh
}

// Run all Steps one by one
func (dh *DownloadHelper) Run() *DownloadHelper {
	_, dh.DownloadParam.DownloadError = DownloadFile(dh)
	return dh
}

// Download File From Cloud With / Without Hash Checking
// params : DownloadParameter, consists of different aspects of downloading
func DownloadFile(params *DownloadHelper) (bool, error) {

	// Clean destination path
	params.destinationPath = filemdl.CleanPath(params.destinationPath)

	// Source URL and Destination path validations
	if strings.Trim(params.sourceURL, EMPTY_STR) == EMPTY_STR || strings.Trim(params.destinationPath, EMPTY_STR) == EMPTY_STR || !strings.HasPrefix(params.sourceURL, HTTP_PROTOCOL) {
		loggermdl.LogDebug("DownloadFile : sourceURL/destinationPath is empty or not proper")
		return false, errormdl.Wrap("Either Source/Destination url is empty or source url does not start with 'http'")
	}

	//Check file available on destination location
	isValid, err := destinationValidator(params.destinationPath)
	if err != nil {
		loggermdl.LogError("File destination validator error : ", err)
		return false, err
	}

	// Cache Burst Implementation
	if params.DownloadParam.isCacheBurst {
		mkclCacheBurst := strconv.FormatInt(time.Now().UnixNano(), 10)
		if strings.Contains(params.sourceURL, "?") {
			params.sourceURL = params.sourceURL + "&mkclCacheBurst=" + mkclCacheBurst
		} else {
			params.sourceURL = params.sourceURL + "?mkclCacheBurst=" + mkclCacheBurst
		}
	}
	loggermdl.LogDebug("DownloadFile : Downloading ", params.sourceURL, " at ", params.destinationPath)

	// Take backup in case required
	backupTaken := false
	downloadSuccessfull := false
	if isValid == FILE_PRESENT_AT_DESTINATION {
		// File present -> Backup required
		err := backupFile(params.destinationPath)
		if errormdl.CheckErr1(err) != nil {
			return false, errormdl.CheckErr1(err)
		}
		backupTaken = true
	}

	var errDownload error
	// Downloading with retry attempts
	for index := 0; index <= params.retryCount; index++ {
		loggermdl.LogDebug("Attempt number: ", index+1)

		// Call method to perform actual downloading
		errDownload = downloadFileFromCloud(params)

		if errormdl.CheckErr1(errDownload) != nil {
			loggermdl.LogError("Download failed: ", err)
			continue
		}
		downloadSuccessfull = true
		loggermdl.LogDebug("DownloadFile : Download Successful")

		// Delete backup
		DeleteBackup(params.destinationPath)
		// All operations done successfully
		return true, nil
	}

	// Restore backup in case downloading failed
	if backupTaken && errormdl.CheckBool(!downloadSuccessfull) {
		restoreBackup(params.destinationPath)
		time.Sleep(WAIT_TIME_DURATION * time.Millisecond)
	}
	return false, errDownload
}

// downloadFileFromCloud: To use http client and download a url to a local file.
// params : DownloadParameter, consists of different aspects of downloading
func downloadFileFromCloud(params *DownloadHelper) error {

	// Check hash check enabled or not
	destinationPath := ""
	if strings.Trim(params.DownloadParam.fileHashedValue, EMPTY_STR) != EMPTY_STR {
		// With hash check implementation
		destinationPath = os.TempDir() + string(os.PathSeparator) + filepath.Base(params.destinationPath)
	} else {
		// Without hash check implementation
		destinationPath = params.destinationPath
	}

	// Create file at destination
	out, err := os.Create(destinationPath)
	if errormdl.CheckErr(err) != nil {
		return errormdl.CheckErr(err)
	}
	defer out.Close()

	//TODO: Code might be refactor (in case provision is given in httpmdl)
	if params.DownloadParam.requestTimeOut <= 0 {
		params.DownloadParam.requestTimeOut = 0
	}

	// Fetching data from httpClient
	client := &http.Client{
		Timeout: time.Duration(params.DownloadParam.requestTimeOut) * time.Second,
	}

	// Call to fetch content
	req, err := http.NewRequest("GET", params.sourceURL, nil)
	if errormdl.CheckErr1(err) != nil {
		return errormdl.CheckErr1(err)
	}

	// Set JWT in request header in case required
	if strings.Trim(params.DownloadParam.jwtToken, EMPTY_STR) != EMPTY_STR {
		req.Header.Add("Authorization", params.DownloadParam.jwtToken)
	}

	// Call http client
	resp, err := client.Do(req)
	if errormdl.CheckErr2(err) != nil {
		return errormdl.CheckErr2(err)
	}
	defer resp.Body.Close()

	if errormdl.CheckBool(resp == nil) {
		return errormdl.Wrap("Empty response.")
	} else if resp.StatusCode != 200 && resp.StatusCode != 206 {
		loggermdl.LogError("Download failed.")
		return errormdl.Wrap("Response code NOT 200 or 206.")
	}
	//resp.StatusCode == 200 || resp.StatusCode == 206
	// Write the response body to file
	_, err = io.Copy(out, resp.Body)
	if errormdl.CheckErr3(err) != nil {
		return errormdl.CheckErr3(err)
	}
	out.Close()

	// File downloaded successfully
	loggermdl.LogDebug("DownloadFile : Download Successful")
	timeHeader := resp.Header.Get("Last-Modified")

	//Last-Modified is not available in case of GRIDFS CDN
	if strings.Trim(timeHeader, EMPTY_STR) != EMPTY_STR {
		err = updateModifiedDateAndValidate(timeHeader, destinationPath, params)
		if err != nil {
			return err
		}
	}

	// All operations done successfully
	return nil
}

// updateModifiedDate - Method will update modifed date and Validate downloaded file
func updateModifiedDateAndValidate(timeHeader, destinationPath string, params *DownloadHelper) error {
	// Get last modified date of file from header file
	lastModifiedTime, err := time.Parse("Mon, 02 Jan 2006 15:04:05 GMT", timeHeader)
	if errormdl.CheckErr(err) != nil {
		// Error while getting last modified date of the file.
		return errormdl.CheckErr(err)
	}

	// Set last modified date and last access date from server
	err = os.Chtimes(destinationPath, lastModifiedTime, lastModifiedTime)
	if errormdl.CheckErr1(err) != nil {
		// Error while setting last modified date of the file.
		return errormdl.CheckErr1(err)
	}

	// Validation of downloaded file
	err = validateDownloadedFile(destinationPath, params)
	if errormdl.CheckErr3(err) != nil {
		return errormdl.CheckErr3(err)
	}
	return nil
}

// ValidateDownloadedFile : Method to validate the file is proper and move the file from TEMP to destination in case hashing is required
func validateDownloadedFile(tempDestinationPath string, params *DownloadHelper) error {
	// Get file info -> to compare with response length
	fileInfo, err := filemdl.FileInfo(tempDestinationPath)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError("Error in getting file info: ", err)
		return errormdl.Wrap("Error in getting file info.")
	}

	// Compare downloaded file with minimum file size
	if errormdl.CheckBool(fileInfo.Size() < MIN_FILE_SIZE) {
		// Content is less than minimum file size
		loggermdl.LogError("Content length less than minimuma specified file size: ", err)
		return errormdl.Wrap("Error, file size is less than expected.")
	}

	// Check if hash checking requied and then calculate and compare hash of downloaded file
	if strings.Trim(params.DownloadParam.fileHashedValue, EMPTY_STR) != EMPTY_STR {
		// Calculate hash value of downloaded file
		hashValue, err := hashmdl.GetAtributeBasedHash(tempDestinationPath)
		if errormdl.CheckErr1(err) != nil {
			loggermdl.LogError("Error in getting hash value of file: ", err)
			return errormdl.Wrap("Error in getting hash value of file.")
		}

		// Compare hash value with provided value
		if errormdl.CheckBool1(hashValue == params.DownloadParam.fileHashedValue) {
			// Move file from Temp to actual destination
			err := filemdl.MoveFile(tempDestinationPath, params.destinationPath)
			if errormdl.CheckErr2(err) != nil {
				loggermdl.LogError("Error in moving downloaded file: ", err)
				return errormdl.Wrap("Error in moving downloaded file.")
			}
			loggermdl.LogDebug("File moved successfully to destination.")
			return nil
		}
		loggermdl.LogError("File and its hashed value is not matching: ", err)
		return errormdl.Wrap("File and its hashed value is not matching.")
	}
	// Validation of downloaded file successfully done
	return nil
}

// destinationValidator
func destinationValidator(destinationPath string) (int, error) {
	// Check file available or not
	if filemdl.FileAvailabilityCheck(destinationPath) {
		// File available at destination
		destInfo, infoError := filemdl.FileInfo(destinationPath)
		if errormdl.CheckErr(infoError) != nil {
			return FILE_PRESENT_CHECK_ERROR, errormdl.Wrap("error occured while getting information of destination directory : " + destinationPath)
		}

		if destInfo.IsDir() {
			loggermdl.LogDebug("DownloadFile : Destination Path is Directory")
			return FILE_PRESENT_CHECK_ERROR, errormdl.Wrap("destination path must be a file path, got directory : " + destinationPath)
		}
		return FILE_PRESENT_AT_DESTINATION, nil
	}
	// File not available at destination
	return FILE_NOT_PRESENT_AT_DESTINATION, nil
}

// backupFile Backup File
func backupFile(filePath string) error {
	fileExists := filemdl.FileAvailabilityCheck(filePath)
	if fileExists {
		loggermdl.LogDebug("backupFile : File Exists. Creating Backup")
		parentDirectory, fileName := path.Split(filePath)
		ext := path.Ext(fileName)
		backupFileName := strings.Replace(fileName, ext, "", -1) + "_backup" + ext

		backupFilePath := filepath.Join(parentDirectory, backupFileName)
		return filemdl.RenameFile(filePath, backupFilePath)
	}
	loggermdl.LogDebug("backupFile : File Not Exists.")
	return nil
}

// restoreBackup Restores Backup
func restoreBackup(filePath string) error {
	fileExists := filemdl.FileAvailabilityCheck(filePath)
	if fileExists {
		loggermdl.LogDebug("restoreBackup : File Exists. Deleting File")
		err := filemdl.DeleteFile(filePath)
		if errormdl.CheckErr(err) != nil {
			return errormdl.CheckErr(err)
		}
	}
	parentDirectory, fileName := path.Split(filePath)
	ext := path.Ext(fileName)
	backupFileName := strings.Replace(fileName, ext, "", -1) + "_backup" + ext

	backupFilePath := filepath.Join(parentDirectory, backupFileName)
	fileExists = filemdl.FileAvailabilityCheck(backupFilePath)
	if fileExists {
		loggermdl.LogDebug("restoreBackup : File Exists. Restoring Backup")
		return filemdl.RenameFile(backupFilePath, filePath)
	}
	loggermdl.LogDebug("restoreBackup : Backup File Not Exists.")
	return nil
}

// DeleteBackup Deletes Backup
func DeleteBackup(filePath string) error {
	parentDirectory, fileName := path.Split(filePath)
	ext := path.Ext(fileName)
	backupFileName := strings.Replace(fileName, ext, "", -1) + "_backup" + ext

	backupFilePath := filepath.Join(parentDirectory, backupFileName)
	fileExists := filemdl.FileAvailabilityCheck(backupFilePath)
	if fileExists {
		loggermdl.LogDebug("DeleteBackup : File Exists. Deleting Backup")
		return filemdl.DeleteFile(backupFilePath)
	}
	loggermdl.LogDebug("DeleteBackup : Backup File Not Exists.")
	return nil
}
