// Author: Deepak Prakash [17-July-2018]

// Description: Package downloadhelpermdl will help download files to destination location.

package downloadhelpermdl

import (
	"os"
	"path/filepath"
	"testing"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"github.com/stretchr/testify/assert"
)

const (
	RETRY_COUNT   = 2
	MODIFIED_DATE = "Sun, 06 Aug 2017 10:43:25 GMT"
	SOURCE_URL    = "https://golangcode.com/images/avatar.jpg?name=Deepak"
)

// TestCase 1 - Valid Case - Pass
func TestGetDownloadHelper_1(t *testing.T) {
	dh := GetDownloadHelper(SOURCE_URL, getDestinationFilePath(), RETRY_COUNT).
		AddParamsFileHashed("15020162054018").
		AddParamsJWT("abcdef").
		AddParamsCacheBurst(true)
	assert.Empty(t, dh.Run().downloadParam.DownloadError)
}

// TestCase 2 - Empty Source URL - Fail
func TestGetDownloadHelper_2(t *testing.T) {
	dh := GetDownloadHelper("", getDestinationFilePath(), RETRY_COUNT)
	assert.NotNil(t, dh.Run().downloadParam.DownloadError)
}

// TestCase 3 - Destination As Folder - Fail
func TestGetDownloadHelper_3(t *testing.T) {
	dh := GetDownloadHelper("https://golangcode.com/images/avatar.jpg", getWorkingDirectoryPath(), RETRY_COUNT)
	assert.NotNil(t, dh.Run().downloadParam.DownloadError)
}

// TestCase 4 - Wrong URL - Fail
func TestGetDownloadHelper_4(t *testing.T) {
	dh := GetDownloadHelper("http://g10.mkcl.org/test", getDestinationFilePath(), RETRY_COUNT).AddParamsCacheBurst(true)
	assert.NotNil(t, dh.Run().downloadParam.DownloadError)
}

// TestCase 5 - File Backup Error - Fail
func TestGetDownloadHelper_5(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn = true
	assert.NotNil(t, dh.Run().downloadParam.DownloadError)
	errormdl.IsTestingNegetiveCaseOn = false
}

// TestCase 6 - Cache Burst URL - Pass
func TestGetDownloadHelper_6(t *testing.T) {
	dh := createFile().AddParamsCacheBurst(true)
	assert.Nil(t, dh.Run().downloadParam.DownloadError)
}

// TestCase 7 - Wrong Hashed Value - Fail
func TestGetDownloadHelper_8(t *testing.T) {
	dh := createFile().AddParamsFileHashed("WrongValue")
	assert.NotNil(t, dh.Run().downloadParam.DownloadError)
}

// TestCase 8 - Delete File Failed Case
func Test_restoreBackup_1(t *testing.T) {
	filePath := getDestinationFilePath()
	errormdl.IsTestingNegetiveCaseOn = true
	assert.NotNil(t, restoreBackup(filePath))
	errormdl.IsTestingNegetiveCaseOn = false
}

// TestCase 9 - Backup File Does Not Exist
func Test_restoreBackup_2(t *testing.T) {
	filePath := getDestinationFilePath()
	errormdl.IsTestingNegetiveCaseOnCheckBool = true
	assert.Nil(t, restoreBackup(filePath))
	errormdl.IsTestingNegetiveCaseOnCheckBool = false
}

// TestCase 10 - Backup creation of a non-existing file
func TestBackupFile_1(t *testing.T) {
	filePath := getWorkingDirectoryPath() + "\\testCase22.jpg"
	assert.Nil(t, backupFile(filePath))
}

// TestCase 11 - Destination Validator
func Test_destinationValidator(t *testing.T) {
	destinationPath := getWorkingDirectoryPath() + "\\testCase22.jpg"
	errormdl.IsTestingNegetiveCaseOn = true
	_, err := destinationValidator(destinationPath)
	assert.Nil(t, err)
	errormdl.IsTestingNegetiveCaseOn = false
}

// TestCase 12 - validate Downloaded File
func Test_validateDownloadedFile_1(t *testing.T) {
	tempDestinationPath := getWorkingDirectoryPath() + "\\testCase22.jpg"
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOnCheckBool = true
	assert.NotNil(t, validateDownloadedFile(tempDestinationPath, dh))
	errormdl.IsTestingNegetiveCaseOnCheckBool = false
}

// TestCase 13 - validate Downloaded File
func Test_validateDownloadedFile_2(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOnCheckBool = true
	assert.NotNil(t, validateDownloadedFile(dh.destinationPath, dh))
	errormdl.IsTestingNegetiveCaseOnCheckBool = false
}

// TestCase 14 - validate Downloaded File
func Test_validateDownloadedFile_3(t *testing.T) {
	dh := createFile().AddParamsFileHashed("15020162054018")
	errormdl.IsTestingNegetiveCaseOn1 = true
	assert.NotNil(t, validateDownloadedFile(dh.destinationPath, dh))
	errormdl.IsTestingNegetiveCaseOn1 = false
}

// TestCase 15 - validate Downloaded File
func Test_validateDownloadedFile_4(t *testing.T) {
	dh := createFile().AddParamsFileHashed("15020162054018")
	errormdl.IsTestingNegetiveCaseOnCheckBool1 = true
	assert.NotNil(t, validateDownloadedFile(dh.destinationPath, dh))
	errormdl.IsTestingNegetiveCaseOnCheckBool1 = false
}

// TestCase 16 - validate Downloaded File
func Test_validateDownloadedFile_5(t *testing.T) {
	dh := createFile().AddParamsFileHashed("15020162054018")
	errormdl.IsTestingNegetiveCaseOn2 = true
	assert.NotNil(t, validateDownloadedFile(dh.destinationPath, dh))
	errormdl.IsTestingNegetiveCaseOn2 = false
}

// TestCase 17 - Download File From Cloud
func Test_downloadFileFromCloud_1(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn = true
	assert.NotNil(t, downloadFileFromCloud(dh))
	errormdl.IsTestingNegetiveCaseOn = false
}

// TestCase 18 - Download File From Cloud
func Test_downloadFileFromCloud_2(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn1 = true
	assert.NotNil(t, downloadFileFromCloud(dh))
	errormdl.IsTestingNegetiveCaseOn1 = false
}

// TestCase 19 - Download File From Cloud
func Test_downloadFileFromCloud_3(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn2 = true
	assert.NotNil(t, downloadFileFromCloud(dh))
	errormdl.IsTestingNegetiveCaseOn2 = false
}

// TestCase 20 - Download File From Cloud
func Test_downloadFileFromCloud_4(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn3 = true
	assert.NotNil(t, downloadFileFromCloud(dh))
	errormdl.IsTestingNegetiveCaseOn3 = false
}

// TestCase 21 - Update Modified Date and Validate
func Test_updateModifiedDateAndValidate_1(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn = true
	assert.NotNil(t, updateModifiedDateAndValidate(MODIFIED_DATE, getTempDestinationFilePath(), dh))
	errormdl.IsTestingNegetiveCaseOn = false
}

// TestCase 22 - Update Modified Date and Validate
func Test_updateModifiedDateAndValidate_2(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn1 = true
	assert.NotNil(t, updateModifiedDateAndValidate(MODIFIED_DATE, getTempDestinationFilePath(), dh))
	errormdl.IsTestingNegetiveCaseOn1 = false
}

// TestCase 23 - Update Modified Date and Validate
func Test_updateModifiedDateAndValidate_3(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn3 = true
	assert.NotNil(t, updateModifiedDateAndValidate(MODIFIED_DATE, getTempDestinationFilePath(), dh))
	errormdl.IsTestingNegetiveCaseOn3 = false
}

// TestCase 24 - Download File From Cloud
func Test_downloadFileFromCloud_5(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOnCheckBool = true
	assert.NotNil(t, downloadFileFromCloud(dh))
	errormdl.IsTestingNegetiveCaseOnCheckBool = false
}

// TestCase 25 - Download File From Cloud
func Test_downloadFileFromCloud_6(t *testing.T) {
	dh := createFile()
	assert.Nil(t, downloadFileFromCloud(dh))
}

// TestCase 26 - Destination File Creation Failed - Fail
func TestGetDownloadHelper_7(t *testing.T) {
	dh := createFile()
	errormdl.IsTestingNegetiveCaseOn1 = true
	_, err := DownloadFile(dh)
	assert.NotNil(t, err)
	errormdl.IsTestingNegetiveCaseOn1 = false
}

// Helper Method - to create file at application running level
func createFile() *DownloadHelper {
	destinationPath := getWorkingDirectoryPath() + "\\testCase.jpg"
	dh := GetDownloadHelper("https://golangcode.com/images/avatar.jpg?name=Deepak", destinationPath, RETRY_COUNT)
	fileExists := filemdl.FileAvailabilityCheck(destinationPath)
	if !fileExists {
		dh.Run()
	}
	return dh
}

// Helper Method - to fetch application running file path
func getWorkingDirectoryPath() string {
	dir, err := filepath.Abs(filepath.Dir(os.Args[0]))
	if err != nil {
		loggermdl.LogError(err)
	}
	return dir
}

// Helper Method - to provide destination file path
func getDestinationFilePath() string {
	return getWorkingDirectoryPath() + "\\testCase.jpg"
}

// Helper Method - to provide destination file path
func getTempDestinationFilePath() string {
	return os.TempDir() + string(os.PathSeparator) + "\\testCase.jpg"
}
