package filepack

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/hashmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/utiliymdl/guidmdl"
	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"
)

const (
	fileStatusReady                     = 0
	fileStatusUpdatingData              = 1
	fileStatusUpdatingIndex             = 2
	fileStatusOffsetInFile              = 0
	isReadyForUploadOffsetInFile        = 1
	isUpdatedAndNotCommitedOffsetInFile = 2
	isReorgRequiredOffsetInFile         = 3
	isReindexRequiredOffsetInFile       = 4
	footerOffsetInFile                  = 5
	footerOffsetReservedSize            = 15
	footerSizeOffset                    = 20
	filehashOffest                      = 35
	lastUpdatedOffset                   = 43
	dataStartOffset                     = 53
	sizeReservedForHeaders              = 53
	lineBreak                           = "\r\n"

	// IndexKeyValSeperator -
	IndexKeyValSeperator = "="
	// FileType - represents key for type of file. Used whenever we need to set file type field in json
	FileType = "fileType"
)

// ErrNoDataFound - ErrNoDataFound
var ErrNoDataFound = errors.New("No data found")

var isSecurityEnabled bool
var isCompressionEnabled bool
var defaultSecurityKey = []byte{}

var lineBreakBytes = []byte(lineBreak)

// Init - initializes pack config
func Init(isSecurityRequired, isCompressionRequired bool, securityKey string) {
	defaultSecurityKey = []byte(securityKey)
	isSecurityEnabled = isSecurityRequired
	isCompressionEnabled = isCompressionRequired
}

// InFileIndex -
type InFileIndex struct {
	FileType    string             `json:"fileType"`
	IndexFields []InFileIndexField `json:"indexFields"`
}

// InFileIndexField - InFileIndexField
type InFileIndexField struct {
	FieldName string `json:"fieldName"`
	Query     string `json:"query"`
}

var mutexMap = map[string]*sync.Mutex{}
var getMapSyncMutex = &sync.Mutex{}

func getLock(fileLocation string) *sync.Mutex {
	getMapSyncMutex.Lock()
	defer getMapSyncMutex.Unlock()
	m, found := mutexMap[fileLocation]
	if !found {
		m = &sync.Mutex{}
		mutexMap[fileLocation] = m
	}
	return m
}

func appendPaddingPadValue(value int64, padNumber int) string {
	no := strconv.Itoa(padNumber)
	return fmt.Sprintf("%0"+no+"d", value)
}

func getFileStatus(f *os.File) (int, error) {

	data, err := filemdl.ReadFileFromOffset(f, fileStatusOffsetInFile, 1)
	if err != nil {
		loggermdl.LogError(err)
		return -1, err
	}
	status, err := strconv.Atoi(string(data))
	return status, err
}

func getFooterOffset(f *os.File) int64 {
	data, err := filemdl.ReadFileFromOffset(f, footerOffsetInFile, footerOffsetReservedSize)
	if err != nil {
		loggermdl.LogError(err)
		return -1
	}
	footerOffset, err := strconv.Atoi(string(data))
	if err != nil {
		loggermdl.LogError("err", err)
		return -1
	}

	return int64(footerOffset)
}

func setFileStatusFlag(f *os.File, fileStatus int) error {
	status := strconv.Itoa(fileStatus)
	_, err := filemdl.WriteFileAtOffset(f, fileStatusOffsetInFile, []byte(status))
	return err
}

func setFileReadyForUploadFlag(f *os.File, isReadyToUpload bool) error {
	flagVal := strconv.FormatBool(isReadyToUpload)
	_, err := filemdl.WriteFileAtOffset(f, isReadyForUploadOffsetInFile, []byte(flagVal))
	return err
}

func setFileUpdatedAndNotCommitedFlag(f *os.File, isUpdatedAndNotCommited bool) error {
	flagVal := strconv.FormatBool(isUpdatedAndNotCommited)
	_, err := filemdl.WriteFileAtOffset(f, isUpdatedAndNotCommitedOffsetInFile, []byte(flagVal))
	return err
}

func setFileReorgRequiredFlag(f *os.File, isReorgRequired bool) error {
	flagVal := strconv.FormatBool(isReorgRequired)
	_, err := filemdl.WriteFileAtOffset(f, isReorgRequiredOffsetInFile, []byte(flagVal))
	return err
}

func setFileReindexRequiredFlag(f *os.File, isReindexRequired bool) error {
	flagVal := strconv.FormatBool(isReindexRequired)
	_, err := filemdl.WriteFileAtOffset(f, isReindexRequiredOffsetInFile, []byte(flagVal))
	return err
}

func setFooterOffset(f *os.File, footerOffset int64) error {
	footerOffestInString := appendPaddingPadValue(footerOffset, 15)
	_, err := filemdl.WriteFileAtOffset(f, footerOffsetInFile, []byte(footerOffestInString))
	return err
}

func setFooterSize(f *os.File, footerSize int64) error {
	footerSizeInString := appendPaddingPadValue(footerSize, 15)
	_, err := filemdl.WriteFileAtOffset(f, footerSizeOffset, []byte(footerSizeInString))
	return err
}

func getFooterSize(f *os.File) (int64, error) {
	data, err := filemdl.ReadFileFromOffset(f, footerSizeOffset, 15)
	if err != nil {
		return -1, err
	}
	footerSize, err := strconv.Atoi(string(data))
	if err != nil {
		loggermdl.LogError("err", err)
		return -1, err
	}

	return int64(footerSize), nil
}

func setIndexDataInFile(f *os.File, footerOffset int64, indexData string) error {
	_, err := filemdl.WriteFileAtOffset(f, footerOffset, []byte(indexData))
	return err
}

func compressData(data []byte) ([]byte, error) {
	var hashError error
	dataByteToWriteRes, hashError := filemdl.ZipBytes(data)
	if errormdl.CheckErr2(hashError) != nil {
		return data, errormdl.CheckErr2(hashError)
	}
	return dataByteToWriteRes, nil
}

func decompressData(data []byte) ([]byte, error) {
	var hashError error
	dataByteToWriteRes, hashError := filemdl.UnZipBytes(data)
	if errormdl.CheckErr2(hashError) != nil {
		return data, errormdl.CheckErr2(hashError)
	}
	return dataByteToWriteRes, nil
}

func encryptData(data []byte, fileName string) (dataOut []byte, err error) {
	key, err := getSecurityKey(fileName)
	if err != nil {
		return dataOut, err
	}
	dataOut, err = compressData(data)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	dataOut, err = securitymdl.AESEncrypt(dataOut, key)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	return
}

func decryptData(data []byte, fileName string) (dataOut []byte, err error) {
	key, err := getSecurityKey(fileName)
	if err != nil {
		return dataOut, err
	}

	dataOut, err = securitymdl.AESDecrypt(data, key)
	if errormdl.CheckErr1(err) != nil {
		return
	}

	dataOut, err = decompressData(dataOut)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	return
}

func getSecurityKey(fileName string) ([]byte, error) {
	key := []byte{}
	var err error
	securityKeyGenFunc := securitymdl.GetSecurityKeyGeneratorFunc()
	if securityKeyGenFunc != nil {
		key, err = securityKeyGenFunc(fileName)
	} else {
		key, err = GetKeyWithFileNameAndDefaultKey(fileName)
	}
	return key, err
}

func addFileDataInFile(f *os.File, offset int64, data string, breakLine bool, rs *gjson.Result, secParams securitymdl.FDBSecParams) (int64, error) {
	dataBytes := []byte(data)

	// if isSecurityEnabled {
	// 	dataBytes, err = encryptData(dataBytes, f.Name())
	// 	if errormdl.CheckErr(err) != nil {
	// 		return 0, errormdl.CheckErr(err)
	// 	}
	// }
	// if !isSecurityEnabled && isCompressionEnabled {
	// 	dataBytes, err = filemdl.ZipBytes(dataBytes)
	// 	if err != nil {
	// 		return 0, err
	// 	}
	// }

	if secParams.EnableSecurity {
		// _, fileName := filepath.Split(f.Name())

		key, err := securitymdl.GenSecKeyBytes(f.Name(), rs)
		if err != nil {
			loggermdl.LogError("failed to generate security key: ", err)
			return 0, err
		}

		dataBytes, err = encryptWithCompression(dataBytes, secParams.EnableCompression, key)
		if err != nil {
			loggermdl.LogError(err)
			return 0, err
		}
	}

	if breakLine {
		dataBytes = append(dataBytes, lineBreakBytes...)
	}

	return filemdl.WriteFileAtOffset(f, offset, dataBytes)
}

func addByteDataInFile(f *os.File, offset int64, dataBytes []byte, breakLine bool) (int64, error) {
	var err error

	if breakLine {
		dataBytes = append(dataBytes, lineBreakBytes...)
	}
	dataSize, err := filemdl.WriteFileAtOffset(f, offset, dataBytes)
	return dataSize, err
}

func getInFileIndexData(f *os.File) (string, error) {
	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return "[]", errormdl.Wrap("fail to fetch infile index data")
	}
	footerSize, err := getFooterSize(f)
	if err != nil {
		return "[]", nil
	}
	if footerSize == 0 {
		return "[]", nil
	}
	dataByte, err := filemdl.ReadFileFromOffset(f, footerStartOffset, footerSize)
	if err != nil {
		if err.Error() == "EOF" {
			loggermdl.LogError("EOF")
			return "[]", nil
		}
		loggermdl.LogError("error while fetching index data", err)
		return "[]", err
	}
	return string(dataByte), nil
}

func getFileDataFromPack(f *os.File, startOffset, dataSize int64, rs *gjson.Result, secParams securitymdl.FDBSecParams) ([]byte, error) {

	ba, err := filemdl.ReadFileFromOffset(f, startOffset, dataSize)

	if secParams.EnableSecurity {
		// _, fileName := filepath.Split(fp.Name())
		key, err := securitymdl.GenSecKeyBytes(f.Name(), rs)
		if err != nil {
			loggermdl.LogError("failed to generate security key: ", err)
			return nil, err
		}
		ba, err = decryptwithDecompression(ba, secParams.EnableCompression, key)
		if err != nil {
			loggermdl.LogError(err)
			return nil, err
		}
	}
	return ba, err
}

// GetKeyWithFileNameAndDefaultKey generates key using file name + Default key
func GetKeyWithFileNameAndDefaultKey(filePath string) ([]byte, error) {
	fileName := filepath.Base(filePath)
	fileNameBytes := []byte(fileName)
	fileNameBytes = append(fileNameBytes, defaultSecurityKey...)
	keyBytes, getHashError := hashmdl.Get128BitHash(fileNameBytes)
	if errormdl.CheckErr(getHashError) != nil {
		return nil, errormdl.CheckErr(getHashError)
	}
	return keyBytes[:], nil
}

//
func CreateIndexJSON(indexFields []InFileIndexField, rs *gjson.Result) (string, error) {
	json := `{}`
	for _, indexField := range indexFields {
		val := rs.Get(indexField.Query).Value()
		// validation
		if val == nil {
			return "", errormdl.Wrap("please provide value for index field: " + indexField.Query)
		}
		json, _ = sjson.Set(json, indexField.FieldName, val)
	}
	return json, nil
}

func initializeWithHeader(filePath string) error {
	// fileLock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(fileLock)
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return errormdl.CheckErr(err)
	// }
	// initailize file with headers and fdb index entry
	f, err := filemdl.OpenFile(filePath, os.O_CREATE|os.O_RDWR|os.O_SYNC, 0777)
	defer f.Close()
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	_, err = f.WriteAt([]byte(strconv.Itoa(fileStatusReady)), fileStatusOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// isFile ready for upload  =0
	_, err = f.WriteAt([]byte("0"), isReadyForUploadOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isUpdatedAndNotCommitedOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReorgRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReindexRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// _, err = f.WriteAt([]byte(appendPaddingToNumber(sizeReservedForHeaders, 15)), footerOffsetInFile)
	err = setFooterOffset(f, sizeReservedForHeaders+int64(len(lineBreak)))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	err = setFooterSize(f, 0)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("filehash"), filehashOffest)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	timestamp := strconv.FormatInt(time.Now().Unix(), 10)
	_, err = f.WriteAt([]byte(timestamp), lastUpdatedOffset)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("\r\n"), sizeReservedForHeaders)
	return err
}

func initializeWithHeaderUsingFp(f *os.File) error {
	// fileLock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(fileLock)
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return errormdl.CheckErr(err)
	// }
	// initailize file with headers and fdb index entry
	// f, err := filemdl.OpenFile(filePath, os.O_CREATE|os.O_RDWR|os.O_SYNC, 0777)
	// defer f.Close()
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return errormdl.CheckErr(err)
	// }
	_, err := f.WriteAt([]byte(strconv.Itoa(fileStatusReady)), fileStatusOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// isFile ready for upload  =0
	_, err = f.WriteAt([]byte("0"), isReadyForUploadOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isUpdatedAndNotCommitedOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReorgRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("0"), isReindexRequiredOffsetInFile)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// _, err = f.WriteAt([]byte(appendPaddingToNumber(sizeReservedForHeaders, 15)), footerOffsetInFile)
	err = setFooterOffset(f, sizeReservedForHeaders+int64(len(lineBreak)))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	err = setFooterSize(f, 0)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("filehash"), filehashOffest)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	timestamp := strconv.FormatInt(time.Now().Unix(), 10)
	_, err = f.WriteAt([]byte(timestamp), lastUpdatedOffset)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	_, err = f.WriteAt([]byte("\r\n"), sizeReservedForHeaders)
	return err
}

// AddFileInPackFile -
func AddFileInPackFile(filePath string, infileIndexFields []InFileIndexField, rs *gjson.Result, secParams securitymdl.FDBSecParams) error {
	isFilePresent := filemdl.FileAvailabilityCheck(filePath)
	if !isFilePresent {
		dir, _ := filepath.Split(filePath)
		err := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return errormdl.CheckErr(err)
		}

		err = initializeWithHeader(filePath)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}

	}
	return addFileInPackFile(filePath, infileIndexFields, rs, secParams)
}

// AddMediaInPackFile -
func AddMediaInPackFile(filePath string, mediaData []byte, infileIndexFields []InFileIndexField, rs *gjson.Result) (string, error) {
	isFilePresent := filemdl.FileAvailabilityCheck(filePath)
	if !isFilePresent {
		dir, _ := filepath.Split(filePath)
		err := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return "", errormdl.CheckErr(err)
		}

		err = initializeWithHeader(filePath)
		if err != nil {
			loggermdl.LogError(err)
			return "", err
		}

	}
	return addMediaInPackFile(filePath, mediaData, infileIndexFields, rs)
}

// AddMediaInPackFileUsingFp - AddMediaInPackFileUsingFp
func AddMediaInPackFileUsingFp(f *os.File, infileIndex *gjson.Result, mediaData []byte, infileIndexFields []InFileIndexField, rs *gjson.Result) (string, *gjson.Result, error) {
	isFilePresent := filemdl.FileAvailabilityCheck(f.Name())

	info, err := f.Stat()
	if err != nil {
		return "", infileIndex, err
	}
	if !isFilePresent || info.Size() == 0 {
		dir, _ := filepath.Split(f.Name())
		err := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return "", infileIndex, errormdl.CheckErr(err)
		}

		err = initializeWithHeaderUsingFp(f)
		if err != nil {
			loggermdl.LogError(err)
			return "", infileIndex, err
		}

	}
	return addMediaInPackFileUsingFp(f, infileIndex, mediaData, infileIndexFields, rs)
}

func addMediaInPackFile(filePath string, mediaData []byte, infileIndexFields []InFileIndexField, rs *gjson.Result) (string, error) {
	recordID := ""
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return recordID, errormdl.Wrap("please specify fileType")
	}
	lock := getLock(filePath)
	lock.Lock()
	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		f.Close()
	}()
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return recordID, errormdl.CheckErr(err)
	}
	// if file is being created for first time
	fileStatus := fileStatusReady
	// if isFilePresent {
	fileStatus, err = getFileStatus(f)
	if err != nil {
		if err.Error() == "EOF" {
			fileStatus = fileStatusReady
		} else {
			loggermdl.LogError(err)
			return recordID, err
		}
	}
	// }
	if fileStatusReady == fileStatus {
		// if its first write in file then add fdb index data of fileType FDBIndex
		//prepare data
		indexRowJSON, err := CreateIndexJSON(infileIndexFields, rs)
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return recordID, err
		}

		fileHash, err := securitymdl.GetHash(rs.String())
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return recordID, err
		}
		// compress & encrypt data
		previousIndexData, err := getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return recordID, err
		}
		// loggermdl.LogDebug("previous index data", previousIndexData)
		footerStartOffset := getFooterOffset(f)
		if footerStartOffset == -1 {
			loggermdl.LogError("fail to fetch infile index offset")
			return recordID, errormdl.Wrap("fail to fetch infile index data")
		}

		err = setFileStatusFlag(f, fileStatusUpdatingData)
		if err != nil {
			loggermdl.LogError(err)
			return recordID, err
		}
		// write data
		dataSize, err := addByteDataInFile(f, footerStartOffset, mediaData, true)
		if err != nil {
			loggermdl.LogError(err)
			return recordID, err
		}
		recordID = rs.Get("recordID").String()
		if recordID == "" {
			recordID = guidmdl.GetGUID()
		}

		indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", fileType)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

		// append new entry in infile index
		parsedindexRowJSON := gjson.Parse(indexRowJSON)
		// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
		updatedIndexData, _ := sjson.Set(previousIndexData, "-1", parsedindexRowJSON.Value())
		// loggermdl.LogDebug("updatedIndexData", updatedIndexData)
		// updating infile index
		err = setFileStatusFlag(f, fileStatusUpdatingIndex)
		if err != nil {
			loggermdl.LogError(err)
			return recordID, err
		}

		footerNewOffset := footerStartOffset + dataSize
		err = setFooterOffset(f, footerNewOffset)
		err = setFooterSize(f, int64(len(updatedIndexData)))
		err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
		err = setFileStatusFlag(f, fileStatusReady)
		if err != nil {
			loggermdl.LogError(err)
			return recordID, err
		}
		f.Sync()
	} else {
		// retry after timeout
	}

	return recordID, nil

}
func addMediaInPackFileUsingFp(f *os.File, infileIndex *gjson.Result, mediaData []byte, infileIndexFields []InFileIndexField, rs *gjson.Result) (string, *gjson.Result, error) {
	recordID := ""
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return recordID, infileIndex, errormdl.Wrap("please specify fileType")
	}
	lock := getLock(f.Name())
	lock.Lock()
	// f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return recordID, infileIndex, errormdl.CheckErr(err)
	// }
	// if file is being created for first time
	// fileStatus := fileStatusReady
	// if isFilePresent {

	// if its first write in file then add fdb index data of fileType FDBIndex
	//prepare data
	indexRowJSON, err := CreateIndexJSON(infileIndexFields, rs)
	if err != nil {
		loggermdl.LogError("error writing to bucket: ", err)
		return recordID, infileIndex, err
	}

	fileHash, err := securitymdl.GetHash(rs.String())
	if err != nil {
		loggermdl.LogError("error writing to bucket: ", err)
		return recordID, infileIndex, err
	}
	previousIndexData := "[]"
	if infileIndex == nil {
		previousIndexData, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return recordID, infileIndex, err
		}
	} else {
		previousIndexData = infileIndex.String()
	}
	// loggermdl.LogDebug("previous index data", previousIndexData)
	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return recordID, infileIndex, errormdl.Wrap("fail to fetch infile index data")
	}

	err = setFileStatusFlag(f, fileStatusUpdatingData)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	// write data
	dataSize, err := addByteDataInFile(f, footerStartOffset, mediaData, true)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}

	recordID = guidmdl.GetGUID()

	indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", fileType)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

	// append new entry in infile index
	parsedindexRowJSON := gjson.Parse(indexRowJSON)
	// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
	updatedIndexData, _ := sjson.Set(previousIndexData, "-1", parsedindexRowJSON.Value())
	// loggermdl.LogDebug("updatedIndexData", updatedIndexData)
	// updating infile index
	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(f, footerNewOffset)
	err = setFooterSize(f, int64(len(updatedIndexData)))
	err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	err = f.Sync()
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	return recordID, &updatedIndexDataObj, nil

}

func addFileInPackFile(filePath string, infileIndexFields []InFileIndexField, rs *gjson.Result, secParams securitymdl.FDBSecParams) error {

	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return errormdl.Wrap("please specify fileType")
	}
	lock := getLock(filePath)
	lock.Lock()
	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		f.Close()
	}()
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	// if file is being created for first time
	fileStatus := fileStatusReady
	// if isFilePresent {
	fileStatus, err = getFileStatus(f)
	if err != nil {
		if err.Error() == "EOF" {
			fileStatus = fileStatusReady
		} else {
			loggermdl.LogError(err)
			return err
		}
	}
	// }
	if fileStatusReady == fileStatus {

		// if its first write in file then add fdb index data of fileType FDBIndex
		//prepare data
		indexRowJSON, err := CreateIndexJSON(infileIndexFields, rs)
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", fileType)
		fileHash, err := securitymdl.GetHash(rs.String())
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
		// compress & encrypt data
		previousIndexData, err := getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		// loggermdl.LogDebug("previous index data", previousIndexData)
		footerStartOffset := getFooterOffset(f)
		if footerStartOffset == -1 {
			loggermdl.LogError("fail to fetch infile index offset")
			return errormdl.Wrap("fail to fetch infile index data")
		}
		dataString := rs.String()
		err = setFileStatusFlag(f, fileStatusUpdatingData)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		// write data
		dataSize, err := addFileDataInFile(f, footerStartOffset, dataString, true, rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
		// append new entry in infile index
		parsedindexRowJSON := gjson.Parse(indexRowJSON)
		// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
		updatedIndexData, _ := sjson.Set(previousIndexData, "-1", parsedindexRowJSON.Value())
		// loggermdl.LogDebug("updatedIndexData", updatedIndexData)
		// updating infile index
		err = setFileStatusFlag(f, fileStatusUpdatingIndex)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}

		footerNewOffset := footerStartOffset + dataSize
		err = setFooterOffset(f, footerNewOffset)
		err = setFooterSize(f, int64(len(updatedIndexData)))
		err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
		err = setFileStatusFlag(f, fileStatusReady)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		f.Sync()
	} else {
		// retry after timeout
	}

	return nil

}

// GetDataFromPackFile - GetDataFromPackFile
func GetDataFromPackFile(filePath string, inFileIndexQueries []string, rs *gjson.Result, secParams securitymdl.FDBSecParams) (string, error) {
	if !filemdl.FileAvailabilityCheck(filePath) {
		return "", errormdl.Wrap("file not found at:" + filePath)
	}
	lock := getLock(filePath)
	lock.Lock()
	f, err := filemdl.Open(filePath)
	defer func() {
		lock.Unlock()
		f.Close()
	}()

	if err != nil {
		loggermdl.LogError("err while opening file: ", filePath, err)
		return "", err
	}
	indexDataString, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", filePath, err)
		return "", err
	}
	indexData := gjson.Parse(indexDataString)
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	sb := strings.Builder{}
	sb.WriteString("[")
	indexRows.ForEach(func(key, indexRow gjson.Result) bool {
		// read files
		startOffSet := indexRow.Get("startOffset").Int()
		dataSize := indexRow.Get("dataSize").Int()
		if startOffSet == 0 || dataSize == 0 {
			return true
		}
		dataByte := []byte{'{', '}'}
		var err error
		// dataByte, err = filemdl.ReadFileFromOffset(f, startOffSet, dataSize)
		dataByte, err = getFileDataFromPack(f, startOffSet, dataSize, rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		_, err = sb.Write(dataByte)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		sb.WriteString(",")

		return true // keep iterating
	})
	sb.WriteString("]")
	finalResult := strings.Replace(sb.String(), ",]", "]", 1)
	return finalResult, nil
}

// GetMediaFromPackFile - GetMediaFromPackFile
func GetMediaFromPackFile(filePath string, recordID string) ([]byte, gjson.Result, error) {
	dataByte := []byte{}
	var indexRow gjson.Result
	if !filemdl.FileAvailabilityCheck(filePath) {
		return dataByte, indexRow, errormdl.Wrap("file not found at:" + filePath)
	}

	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}

	lock := getLock(filePath)
	lock.Lock()
	f, err := filemdl.Open(filePath)
	defer func() {
		lock.Unlock()
		f.Close()
	}()

	if err != nil {
		loggermdl.LogError("err while opening file: ", filePath, err)
		return dataByte, indexRow, err
	}
	indexDataString, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", filePath, err)
		return dataByte, indexRow, err
	}
	indexData := gjson.Parse(indexDataString)
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	if indexRows.String() == "" {
		loggermdl.LogError("data not found for record id: ", recordID)
		return dataByte, indexRow, errormdl.Wrap("data not found for record id: " + recordID)
	}
	sb := strings.Builder{}
	sb.WriteString("[")
	indexRow = indexRows.Get("0")
	// indexRows.ForEach(func(key, indexRow gjson.Result) bool {
	// read files
	startOffSet := indexRow.Get("startOffset").Int()
	dataSize := indexRow.Get("dataSize").Int()
	if startOffSet == 0 || dataSize == 0 {
		return dataByte, indexRow, errormdl.Wrap("data not found")
	}
	// dataByte := []byte{'{', '}'}
	// var err error
	// dataByte, err = filemdl.ReadFileFromOffset(f, startOffSet, dataSize)
	dataByte, err = getFileDataFromPack(f, startOffSet, dataSize, nil, securitymdl.FDBSecParams{EnableSecurity: false})
	if err != nil {
		loggermdl.LogError(err)
		return dataByte, indexRow, err
	}

	return dataByte, indexRow.Get("requiredData"), nil
}

// GetMediaFromPackFileUsingFp - GetMediaFromPackFileUsingFp
func GetMediaFromPackFileUsingFp(f *os.File, infileIndex *gjson.Result, recordID string) ([]byte, *gjson.Result, error) {
	dataByte := []byte{}
	var metaData *gjson.Result
	if !filemdl.FileAvailabilityCheck(f.Name()) {
		return dataByte, metaData, errormdl.Wrap("file not found at:" + f.Name())
	}

	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}

	lock := getLock(f.Name())
	lock.Lock()
	// f, err := filemdl.Open(filePath)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()

	// if err != nil {
	// 	loggermdl.LogError("err while opening file: ", filePath, err)
	// 	return dataByte, metaData, err
	// }
	indexDataString := "[]"
	var err error
	if infileIndex == nil {
		indexDataString, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError("index data not found: ", f.Name(), err)
			return dataByte, metaData, err
		}
	} else {
		indexDataString = infileIndex.String()
	}
	indexData := gjson.Parse(indexDataString)
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	if indexRows.String() == "" {
		loggermdl.LogError("data not found for recordId: ", recordID)
		return dataByte, metaData, errormdl.Wrap("data not found")
	}
	sb := strings.Builder{}
	sb.WriteString("[")
	indexRow := indexRows.Get("0")
	// indexRows.ForEach(func(key, indexRow gjson.Result) bool {
	// read files
	startOffSet := indexRow.Get("startOffset").Int()
	dataSize := indexRow.Get("dataSize").Int()
	if startOffSet == 0 || dataSize == 0 {
		return dataByte, metaData, errormdl.Wrap("data not found")
	}
	// dataByte := []byte{'{', '}'}
	// var err error
	// dataByte, err = filemdl.ReadFileFromOffset(f, startOffSet, dataSize)
	dataByte, err = getFileDataFromPack(f, startOffSet, dataSize, nil, securitymdl.FDBSecParams{EnableSecurity: false})
	if err != nil {
		loggermdl.LogError(err)
		return dataByte, metaData, err
	}

	data, _ := sjson.Set("", "requiredData", indexRow.Get("requiredData").String())
	data, _ = sjson.Set(data, "infileIndex", indexData.String())
	metaDataObj := gjson.Parse(data)
	return dataByte, &metaDataObj, nil
}

// UpdateMediaInPackFile - UpdateMediaInPackFile
func UpdateMediaInPackFile(filePath string, recordID string, mediaData []byte, infileIndexFields []InFileIndexField, rs *gjson.Result) (string, error) {

	if !filemdl.FileAvailabilityCheck(filePath) {
		return recordID, errormdl.Wrap("file not found at:" + filePath)
	}

	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}

	lock := getLock(filePath)
	lock.Lock()

	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		f.Close()
	}()

	if err != nil {
		loggermdl.LogError("err while opening file: ", filePath, err)
		return recordID, err
	}
	indexDataString, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError("index data not found: ", filePath, err)
		return recordID, err
	}

	indexData := gjson.Parse(indexDataString)
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	foundAtIndex := -1
	foundIndexRow := gjson.Result{}
	for index, indexRow := range indexRows.Array() {
		if indexRow.Get("recordID").String() == recordID {
			foundAtIndex = index
			foundIndexRow = indexRow
			break
		}
	}
	// if indexRows.String() == "" {
	// 	loggermdl.LogError("data not found for record id: ", recordID)
	// 	return recordID, errormdl.Wrap("data not found for record id: " + recordID)
	// }

	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return recordID, errormdl.Wrap("fail to fetch infile index data")
	}

	// write data
	dataSize, err := addByteDataInFile(f, footerStartOffset, mediaData, true)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, err
	}
	indexRowJSON := foundIndexRow.String()
	if foundAtIndex == -1 {
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", rs.Get("fileType").String())
	}
	indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

	// append new entry in infile index
	updatedIndexRow := gjson.Parse(indexRowJSON)
	// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
	updatedIndexData, _ := sjson.Set(indexDataString, strconv.Itoa(foundAtIndex), updatedIndexRow.Value())
	// updatedIndexData := gjson.Parse(indexDataString)
	// loggermdl.LogDebug("updatedIndexData", updatedIndexData)

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(f, footerNewOffset)
	err = setFooterSize(f, int64(len(updatedIndexData)))
	err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, err
	}
	err = f.Sync()
	return recordID, err
}

// UpdateMediaInPackFileUsingFp - UpdateMediaInPackFileUsingFp
func UpdateMediaInPackFileUsingFp(f *os.File, infileIndex *gjson.Result, recordID string, mediaData []byte, infileIndexFields []InFileIndexField, rs *gjson.Result) (string, *gjson.Result, error) {

	isFilePresent := filemdl.FileAvailabilityCheck(f.Name())

	info, err := f.Stat()
	if err != nil {
		return "", infileIndex, err
	}
	if !isFilePresent || info.Size() == 0 {
		return recordID, infileIndex, errormdl.Wrap("file not found at:" + f.Name())
	}
	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}

	lock := getLock(f.Name())
	lock.Lock()

	// f, err := filemdl.OpenFile(f.Name(), os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()

	// if err != nil {
	// 	loggermdl.LogError("err while opening file: ", filePath, err)
	// 	return recordID, infileIndex, err
	// }
	indexDataString := "[]"

	if infileIndex == nil {

		indexDataString, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError("index data not found: ", f.Name(), err)
			return recordID, infileIndex, err
		}
	} else {
		indexDataString = infileIndex.String()
	}

	indexData := gjson.Parse(indexDataString)
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	foundAtIndex := -1
	foundIndexRow := gjson.Result{}
	for index, indexRow := range indexData.Array() {
		r := indexRow.Get("recordID").String()
		if r != "" && indexRow.Get("recordID").String() == recordID {
			foundAtIndex = index
			foundIndexRow = indexRow
			break
		}
	}
	if foundAtIndex == -1 {
		loggermdl.LogError("no data found to update: ", recordID)
		return recordID, infileIndex, errormdl.Wrap("no data found to update ")
	}

	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return recordID, infileIndex, errormdl.Wrap("fail to fetch infile index data")
	}

	// write data
	dataSize, err := addByteDataInFile(f, footerStartOffset, mediaData, true)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	indexRowJSON := foundIndexRow.String()
	// if foundAtIndex == -1 {
	// 	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", rs.Get("fileType").String())
	// }
	indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

	// append new entry in infile index
	updatedIndexRow := gjson.Parse(indexRowJSON)
	// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
	updatedIndexData, _ := sjson.Set(indexDataString, strconv.Itoa(foundAtIndex), updatedIndexRow.Value())
	// updatedIndexData := gjson.Parse(indexDataString)
	// loggermdl.LogDebug("updatedIndexData", updatedIndexData)

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(f, footerNewOffset)
	err = setFooterSize(f, int64(len(updatedIndexData)))
	err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	err = f.Sync()
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	return recordID, &updatedIndexDataObj, err
}

// UpsertMediaInPackFileUsingFp - UpsertMediaInPackFileUsingFp
func UpsertMediaInPackFileUsingFp(f *os.File, infileIndex *gjson.Result, recordID string, mediaData []byte, infileIndexFields []InFileIndexField, rs *gjson.Result) (string, *gjson.Result, error) {

	isFilePresent := filemdl.FileAvailabilityCheck(f.Name())

	info, err := f.Stat()
	if err != nil {
		return "", infileIndex, err
	}
	if !isFilePresent || info.Size() == 0 {
		dir, _ := filepath.Split(f.Name())
		err := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return "", infileIndex, errormdl.CheckErr(err)
		}

		err = initializeWithHeaderUsingFp(f)
		if err != nil {
			loggermdl.LogError(err)
			return "", infileIndex, err
		}

	}
	inFileIndexQueries := []string{`#[recordID=` + recordID + `]`}

	lock := getLock(f.Name())
	lock.Lock()

	// f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()

	// if err != nil {
	// 	loggermdl.LogError("err while opening file: ", filePath, err)
	// 	return recordID, err
	// }
	indexDataString := "[]"

	if infileIndex == nil {
		indexDataString, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError("index data not found: ", f.Name(), err)
			return recordID, infileIndex, err
		}
	} else {
		indexDataString = infileIndex.String()
	}

	indexData := gjson.Parse(indexDataString)
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	foundAtIndex := -1
	foundIndexRow := gjson.Result{}
	for index, indexRow := range indexData.Array() {
		if indexRow.Get("recordID").String() == recordID {
			foundAtIndex = index
			foundIndexRow = indexRow
			break
		}
	}
	// if foundAtIndex == -1 {
	// 	loggermdl.LogError("no data found to update: ", recordID)
	// 	return recordID, infileIndex, errormdl.Wrap("no data found to update")
	// }
	// if indexRows.String() == "" {
	// 	loggermdl.LogError("data not found for record id: ", recordID)
	// 	return recordID,infileIndex, errormdl.Wrap("data not found for record id: " + recordID)
	// }

	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return recordID, infileIndex, errormdl.Wrap("fail to fetch infile index data")
	}

	// write data
	dataSize, err := addByteDataInFile(f, footerStartOffset, mediaData, true)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	indexRowJSON := foundIndexRow.String()

	indexRowJSON, _ = sjson.Set(indexRowJSON, "requiredData", rs.String())
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "recordID", recordID)

	// append new entry in infile index
	updatedIndexRow := gjson.Parse(indexRowJSON)
	// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
	updatedIndexData, _ := sjson.Set(indexDataString, strconv.Itoa(foundAtIndex), updatedIndexRow.Value())
	// updatedIndexData := gjson.Parse(indexDataString)
	// loggermdl.LogDebug("updatedIndexData", updatedIndexData)

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(f, footerNewOffset)
	err = setFooterSize(f, int64(len(updatedIndexData)))
	err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return recordID, infileIndex, err
	}
	err = f.Sync()
	if err != nil {
		return recordID, infileIndex, err
	}
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	return recordID, &updatedIndexDataObj, err
}

// UpdateFilesInPackFile - UpdateFilesInPackFile
func UpdateFilesInPackFile(filePath string, infileIndexQueries []string, rs *gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, error) {
	// check fileType index availability
	// check is data present
	// if data present
	// then calculate size of updated data
	// if size is less than or equal to previuos data size
	// then write at the same location
	// else if size of updated data is more than existing data then append it to end of data
	// update startOffset and data size of file in index row
	// update footer offset and footer size
	requestedFileType := rs.Get("fileType").String()
	if len(requestedFileType) == 0 {
		return nil, errormdl.Wrap("please specify fileType")
	}

	if !filemdl.FileAvailabilityCheck(filePath) {
		return nil, errormdl.Wrap("file not found: " + filePath)
	}
	// lock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(lock)
	// if err != nil {
	// 	return nil, errormdl.Wrap("fail to update data" + err.Error())
	// }
	lock := getLock(filePath)
	lock.Lock()
	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		f.Close()
	}()
	if err != nil {
		return nil, errormdl.Wrap("fail to update data" + err.Error())
	}
	indexDataString, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	// infileIndexQueries = append(infileIndexQueries, `#[fileType==`+requestedFileType+`]`)
	indexRows := gjson.Parse(indexDataString)
	indexRecordsToUpdate := indexRows
	for _, query := range infileIndexQueries {
		indexRecordsToUpdate = indexRecordsToUpdate.Get(query + "#")
	}

	indexRecordsToUpdateObjs := indexRecordsToUpdate.Array()
	if len(indexRecordsToUpdateObjs) == 0 {
		return nil, ErrNoDataFound
	}
	// updating first record
	resultArrayStr := "[]"
	for _, recordToUpdateIndexRow := range indexRecordsToUpdateObjs {
		// loggermdl.LogDebug("recordToUpdateIndexRow", recordToUpdateIndexRow)
		result, err := updateSingleRecordInPackFile(f, recordToUpdateIndexRow, *rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			return nil, errormdl.Wrap("fail to update data" + err.Error())
		}
		resultArrayStr, _ = sjson.Set(resultArrayStr, "-1", result.Value())
	}
	resultData := gjson.Parse(resultArrayStr)
	return &resultData, nil
}

func updateSingleRecordInPackFile(f *os.File, recordToUpdateIndexRow gjson.Result, rs gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, error) {
	fileStartOffset := recordToUpdateIndexRow.Get("startOffset").Int()
	dataSize := recordToUpdateIndexRow.Get("dataSize").Int()
	if fileStartOffset == 0 || dataSize == 0 {
		loggermdl.LogError("index row details incorrect - start offset :", fileStartOffset, " data size :", dataSize)
		return nil, errormdl.Wrap("index row details incorrect")
	}
	// compare data size
	// TODO: compare with encrypted data

	// update fields data
	// existingData, err := filemdl.ReadFileFromOffset(f, fileStartOffset, dataSize)
	existingData, err := getFileDataFromPack(f, fileStartOffset, dataSize, &rs, secParams)
	if err != nil {
		return nil, err
	}

	updatedDataStr := strings.TrimSpace(string(existingData))
	rs.ForEach(func(key, val gjson.Result) bool {
		// updating existing data
		updatedDataStr, _ = sjson.Set(updatedDataStr, key.String(), val.Value())
		return true
	})
	newDataSize := int64(len(updatedDataStr))
	footerStartOffset := getFooterOffset(f)
	updatedFooterOffset := footerStartOffset
	err = setFileStatusFlag(f, fileStatusUpdatingData)
	if err != nil {
		return nil, err
	}
	indexDataString, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	existingIndexRows := gjson.Parse(indexDataString)
	if len(strings.TrimSpace(updatedDataStr)) <= len(strings.TrimSpace(string(existingData))) {
		newDataSize, err = addFileDataInFile(f, fileStartOffset, updatedDataStr, false, &rs, secParams)
		if err != nil {
			return nil, err
		}
		// newDataSize = newDataSize + 2
	} else {

		newDataSize, err = addFileDataInFile(f, footerStartOffset, updatedDataStr, true, &rs, secParams)
		updatedFooterOffset = footerStartOffset + newDataSize
		fileStartOffset = footerStartOffset
	}
	// indexRows, err := getInFileIndexData(f)
	// if err != nil {
	// 	return nil, err
	// }
	// update startofffset and datasize in infile index
	// updateIndexRow
	updatedIndexRowStr := recordToUpdateIndexRow.String()

	recordToUpdateIndexRow.ForEach(func(key, value gjson.Result) bool {
		indexFieldKey := key.String()
		if rs.Get(indexFieldKey).Exists() {
			updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, indexFieldKey, rs.Get(indexFieldKey).Value())
		}
		return true
	})
	fileHash, err := securitymdl.GetHash(updatedDataStr)
	if err != nil {
		return nil, err
	}
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "startOffset", fileStartOffset)
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "dataSize", newDataSize)
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "fileHash", fileHash)
	updatedIndexRows, err := updateIndexRow(&existingIndexRows, recordToUpdateIndexRow, gjson.Parse(updatedIndexRowStr))

	if err != nil {
		return nil, err
	}

	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return nil, err
	}
	// loggermdl.LogDebug("updated index data", updatedIndexRows)
	err = setIndexDataInFile(f, updatedFooterOffset, updatedIndexRows.String())
	if err != nil {
		return nil, err
	}
	err = setFooterOffset(f, updatedFooterOffset)
	if err != nil {
		return nil, err
	}

	err = setFooterSize(f, int64(len(updatedIndexRows.String())))
	if err != nil {
		return nil, err
	}

	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return nil, err
	}
	f.Sync()
	// TODO: discussion
	// err = setFileReadyForUploadFlag(f, true)
	// if err != nil {
	// 	return err
	// }
	updatedData := gjson.Parse(updatedDataStr)
	return &updatedData, nil
}

func updateIndexRow(indexRows *gjson.Result, previousIndexRow gjson.Result, updatedRow gjson.Result) (*gjson.Result, error) {
	indexRowObjs := indexRows.Array()
	if len(indexRowObjs) == 0 {
		return nil, errormdl.Wrap("no data found to update")
	}
	// loggermdl.LogDebug("indexRows", indexRows)

	prevIndexRowString := previousIndexRow.String()
	foundRowIndex := -1

	for index, indexRowObj := range indexRowObjs {
		if indexRowObj.String() != "" && prevIndexRowString != "" && indexRowObj.String() == prevIndexRowString {
			foundRowIndex = index
			break
		}
	}

	if foundRowIndex == -1 {
		return nil, errormdl.Wrap("no record found to update")
	}
	var err error
	updatedIndexDataString := indexRows.String()
	// for _, foundRowIndex := range foundRowIndexes {
	updatedIndexDataString, err = sjson.Set(updatedIndexDataString, strconv.Itoa(foundRowIndex), updatedRow.Value())
	if err != nil {
		loggermdl.LogError(err)
		return nil, errormdl.Wrap("failed to update index rows")
	}
	// }
	updatedIndexData := gjson.Parse(updatedIndexDataString)
	return &updatedIndexData, nil
}

// DeletDataFromPackFile -DeletDataFromPackFile
func DeletDataFromPackFile(filePath string, infileIndexQueries []string) (int, error) {
	// get indexdata
	// get update index records
	// save updated index
	// update index size header
	// lock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(lock)
	// if err != nil {
	// 	return err
	// }
	recordsDeletedCnt := 0
	lock := getLock(filePath)
	lock.Lock()
	f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		f.Close()
	}()
	if err != nil {
		return recordsDeletedCnt, err
	}

	indexDataStr, err := getInFileIndexData(f)
	if err != nil {
		return recordsDeletedCnt, err
	}

	indexData := gjson.Parse(indexDataStr)
	indexRecordsToDelete := indexData
	// loggermdl.LogDebug("indexRecordsToDelete file type", indexRecordsToDelete)
	for _, query := range infileIndexQueries {
		indexRecordsToDelete = indexRecordsToDelete.Get(query + "#")
	}
	indexRowsToDelete := indexRecordsToDelete.Array()
	if len(indexRowsToDelete) == 0 {
		loggermdl.LogError("ErrNoDataFound")
		return recordsDeletedCnt, ErrNoDataFound
	}
	updatedIndexRecords := indexData
	for _, indexRowToRemove := range indexRowsToDelete {
		updatedIndexRecords, err = removeIndexRow(updatedIndexRecords, indexRowToRemove.String())
		if err != nil {
			loggermdl.LogError("fail to delete record:", err)
			return recordsDeletedCnt, errormdl.Wrap("fail to delete record:" + err.Error())
		}
		recordsDeletedCnt++
	}

	footerOffset := getFooterOffset(f)
	if footerOffset == -1 {
		return recordsDeletedCnt, errormdl.Wrap("fail to fetch infile index offset")
	}
	newIndexDataSize := len(updatedIndexRecords.String())
	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return recordsDeletedCnt, err
	}
	err = setIndexDataInFile(f, footerOffset, updatedIndexRecords.String())
	if err != nil {
		loggermdl.LogError("fail to update infile index data :", err)
		return recordsDeletedCnt, err
	}
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return recordsDeletedCnt, err
	}
	return recordsDeletedCnt, setFooterSize(f, int64(newIndexDataSize))
}

func removeIndexRow(indexRows gjson.Result, indexRowToDelete string) (gjson.Result, error) {

	indexRowObjs := indexRows.Array()
	if len(indexRowObjs) == 0 {
		return indexRows, errormdl.Wrap("no data found to update")
	}
	// loggermdl.LogDebug("indexRows", indexRows)

	foundIndexToDelete := -1
	for index, indexRowObj := range indexRowObjs {
		if indexRowObj.String() != "" && indexRowToDelete != "" && indexRowObj.String() == indexRowToDelete {
			foundIndexToDelete = index
			break
		}
	}

	if foundIndexToDelete == -1 {
		return indexRows, errormdl.Wrap("no record found to delete")
	}
	var err error
	updatedIndexDataString, err := sjson.Delete(indexRows.String(), strconv.Itoa(foundIndexToDelete))
	if err != nil {
		loggermdl.LogError(err)
	}
	return gjson.Parse(updatedIndexDataString), nil
}

// AddFileInPackFileUsingFp -
func AddFileInPackFileUsingFp(f *os.File, infileIndexFields []InFileIndexField, rs *gjson.Result, secParams securitymdl.FDBSecParams) error {
	filePath := f.Name()
	isFilePresent := filemdl.FileAvailabilityCheck(filePath)
	info, err := f.Stat()
	if err != nil {
		return err
	}
	if !isFilePresent || info.Size() == 0 {
		dir, _ := filepath.Split(filePath)
		err := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return errormdl.CheckErr(err)
		}

		err = initializeWithHeaderUsingFp(f)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
	}
	return addFileInPackFileUsingFp(f, infileIndexFields, rs, secParams)
}

func addFileInPackFileUsingFp(f *os.File, infileIndexFields []InFileIndexField, rs *gjson.Result, secParams securitymdl.FDBSecParams) error {
	filePath := f.Name()
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return errormdl.Wrap("please specify fileType")
	}
	lock := getLock(filePath)
	lock.Lock()
	// f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()
	// if errormdl.CheckErr(err) != nil {
	// 	loggermdl.LogError(err)
	// 	return errormdl.CheckErr(err)
	// }
	// if file is being created for first time
	fileStatus := fileStatusReady
	// if isFilePresent {
	fileStatus, err := getFileStatus(f)
	if err != nil {
		if err.Error() == "EOF" {
			fileStatus = fileStatusReady
		} else {
			loggermdl.LogError(err)
			return err
		}
	}
	// }
	if fileStatusReady == fileStatus {

		// if its first write in file then add fdb index data of fileType FDBIndex
		//prepare data
		indexRowJSON, err := CreateIndexJSON(infileIndexFields, rs)
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", fileType)
		fileHash, err := securitymdl.GetHash(rs.String())
		if err != nil {
			loggermdl.LogError("error writing to bucket: ", err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
		// compress & encrypt data
		previousIndexData, err := getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		// loggermdl.LogDebug("previous index data", previousIndexData)
		footerStartOffset := getFooterOffset(f)
		if footerStartOffset == -1 {
			loggermdl.LogError("fail to fetch infile index offset")
			return errormdl.Wrap("fail to fetch infile index data")
		}
		dataString := rs.String()
		err = setFileStatusFlag(f, fileStatusUpdatingData)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		// write data
		dataSize, err := addFileDataInFile(f, footerStartOffset, dataString, true, rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
		// append new entry in infile index
		parsedindexRowJSON := gjson.Parse(indexRowJSON)
		// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
		updatedIndexData, _ := sjson.Set(previousIndexData, "-1", parsedindexRowJSON.Value())
		// loggermdl.LogDebug("updatedIndexData", updatedIndexData)
		// updating infile index
		err = setFileStatusFlag(f, fileStatusUpdatingIndex)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}

		footerNewOffset := footerStartOffset + dataSize
		err = setFooterOffset(f, footerNewOffset)
		err = setFooterSize(f, int64(len(updatedIndexData)))
		err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
		err = setFileStatusFlag(f, fileStatusReady)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		f.Sync()
	} else {
		// retry after timeout
	}
	return nil
}

// GetDataFromPackFileUsingFp - GetDataFromPackFileUsingFp
func GetDataFromPackFileUsingFp(f *os.File, infileIndexData *gjson.Result, inFileIndexQueries []string, rs *gjson.Result, secParams securitymdl.FDBSecParams) (string, *gjson.Result, error) {
	filePath := f.Name()
	if !filemdl.FileAvailabilityCheck(filePath) {
		return "", infileIndexData, errormdl.Wrap("file not found at:" + filePath)
	}

	lock := getLock(filePath)
	lock.Lock()
	// f, err := filemdl.Open(filePath)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()

	// if err != nil {
	// 	loggermdl.LogError("err while opening file: ", filePath, err)
	// 	return "", err
	// }
	indexDataString := ""
	var err error
	if infileIndexData == nil {
		indexDataString, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError("index data not found: ", filePath, err)
			return "", infileIndexData, err
		}
	} else {
		indexDataString = infileIndexData.String()
	}
	indexData := gjson.Parse(indexDataString)
	indexRows := indexData
	// indexRows := indexData.Get(`#[fileType==` + requestedFileType + `]#`)
	for i := 0; i < len(inFileIndexQueries); i++ {
		indexRows = indexRows.Get(inFileIndexQueries[i] + "#")
	}
	sb := strings.Builder{}
	sb.WriteString("[")
	indexRows.ForEach(func(key, indexRow gjson.Result) bool {
		// read files
		startOffSet := indexRow.Get("startOffset").Int()
		dataSize := indexRow.Get("dataSize").Int()
		if startOffSet == 0 || dataSize == 0 {
			return true
		}
		dataByte := []byte{'{', '}'}
		var err error
		// dataByte, err = filemdl.ReadFileFromOffset(f, startOffSet, dataSize)
		dataByte, err = getFileDataFromPack(f, startOffSet, dataSize, rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		_, err = sb.Write(dataByte)
		if err != nil {
			loggermdl.LogError(err)
			return true
		}
		sb.WriteString(",")

		return true // keep iterating
	})
	sb.WriteString("]")
	finalResult := strings.Replace(sb.String(), ",]", "]", 1)
	return finalResult, &indexData, nil
}

// AddPackFileUsingFp -
func AddPackFileUsingFp(f *os.File, infileIndexFields []InFileIndexField, infileIndex, rs *gjson.Result, secParams securitymdl.FDBSecParams) (infileIndexData *gjson.Result, err error) {
	filePath := f.Name()
	isFilePresent := filemdl.FileAvailabilityCheck(filePath)
	info, err := f.Stat()
	if err != nil {
		return
	}
	if !isFilePresent || info.Size() == 0 {
		dir, _ := filepath.Split(filePath)
		err = filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			err = err
			return
		}

		err = initializeWithHeaderUsingFp(f)
		if err != nil {
			loggermdl.LogError(err)
			return
		}
	}
	return addPackFileUsingFp(f, infileIndex, infileIndexFields, rs, secParams)
}

func addPackFileUsingFp(f *os.File, infileIndex *gjson.Result, infileIndexFields []InFileIndexField, rs *gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, error) {
	filePath := f.Name()
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return nil, errormdl.Wrap("please specify fileType")
	}
	lock := getLock(filePath)
	lock.Lock()
	// f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
	}()

	//prepare data
	indexRowJSON, err := CreateIndexJSON(infileIndexFields, rs)
	if err != nil {
		loggermdl.LogError("error writing to bucket: ", err)
		return nil, err
	}
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileType", fileType)
	fileHash, err := securitymdl.GetHash(rs.String())
	if err != nil {
		loggermdl.LogError("error writing to bucket: ", err)
		return nil, err
	}
	indexRowJSON, _ = sjson.Set(indexRowJSON, "fileHash", fileHash)
	// compress & encrypt data
	previousIndexData := "[]"
	if infileIndex == nil {
		previousIndexData, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return nil, err
		}
	} else {
		previousIndexData = infileIndex.String()
	}
	// loggermdl.LogDebug("previous index data", previousIndexData)
	footerStartOffset := getFooterOffset(f)
	if footerStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return nil, errormdl.Wrap("fail to fetch infile index data")
	}
	dataString := rs.String()
	err = setFileStatusFlag(f, fileStatusUpdatingData)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	// write data
	dataSize, err := addFileDataInFile(f, footerStartOffset, dataString, true, rs, secParams)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	indexRowJSON, _ = sjson.Set(indexRowJSON, "startOffset", footerStartOffset)
	indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", dataSize)
	// append new entry in infile index
	parsedindexRowJSON := gjson.Parse(indexRowJSON)
	// indexTableRecords, _ = sjson.Set(indexTableRecords, "-1", parsedJsonzObj.Value())
	updatedIndexData, _ := sjson.Set(previousIndexData, "-1", parsedindexRowJSON.Value())
	// loggermdl.LogDebug("updatedIndexData", updatedIndexData)
	// updating infile index
	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}

	footerNewOffset := footerStartOffset + dataSize
	err = setFooterOffset(f, footerNewOffset)
	err = setFooterSize(f, int64(len(updatedIndexData)))
	err = setIndexDataInFile(f, footerNewOffset, updatedIndexData)
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	f.Sync()
	// } else {
	// 	// retry after timeout
	// }
	updatedIndexDataObj := gjson.Parse(updatedIndexData)
	return &updatedIndexDataObj, nil
}

// UpdateFilesInPackFileUsingFp - UpdateFilesInPackFileUsingFp
func UpdateFileInPackFileUsingFp(f *os.File, infileIndexQueries []string, infileIndex, rs *gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, *gjson.Result, error) {
	// check fileType index availability
	// check is data present
	// if data present
	// then calculate size of updated data
	// if size is less than or equal to previuos data size
	// then write at the same location
	// else if size of updated data is more than existing data then append it to end of data
	// update startOffset and data size of file in index row
	// update footer offset and footer size
	requestedFileType := rs.Get("fileType").String()
	if len(requestedFileType) == 0 {
		return nil, infileIndex, errormdl.Wrap("please specify fileType")
	}

	if !filemdl.FileAvailabilityCheck(f.Name()) {
		return nil, infileIndex, errormdl.Wrap("file not found: " + f.Name())
	}
	// lock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(lock)
	// if err != nil {
	// 	return nil, infileIndex, errormdl.Wrap("fail to update data" + err.Error())
	// }
	lock := getLock(f.Name())
	lock.Lock()
	// f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()
	// if err != nil {
	// 	return nil, infileIndex, errormdl.Wrap("fail to update data" + err.Error())
	// }
	indexDataString := "[]"
	var err error
	if infileIndex == nil {
		indexDataString, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return nil, infileIndex, err
		}
	} else {
		indexDataString = infileIndex.String()
	}
	// infileIndexQueries = append(infileIndexQueries, `#[fileType==`+requestedFileType+`]`)
	indexRows := gjson.Parse(indexDataString)
	indexRecordsToUpdate := indexRows
	for _, query := range infileIndexQueries {
		indexRecordsToUpdate = indexRecordsToUpdate.Get(query + "#")
	}

	indexRecordsToUpdateObjs := indexRecordsToUpdate.Array()
	if len(indexRecordsToUpdateObjs) == 0 {
		return nil, infileIndex, ErrNoDataFound
	}
	// updating first record
	resultArrayStr := "[]"
	var updatedInfileIndex *gjson.Result
	var result *gjson.Result
	for _, recordToUpdateIndexRow := range indexRecordsToUpdateObjs {
		// loggermdl.LogDebug("recordToUpdateIndexRow", recordToUpdateIndexRow)
		result, updatedInfileIndex, err = updateSingleRecordInPackFileUsingFp(f, recordToUpdateIndexRow, updatedInfileIndex, rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			return nil, infileIndex, errormdl.Wrap("fail to update data" + err.Error())
		}
		resultArrayStr, _ = sjson.Set(resultArrayStr, "-1", result.Value())
	}
	resultData := gjson.Parse(resultArrayStr)
	return &resultData, updatedInfileIndex, nil
}

func updateSingleRecordInPackFileUsingFp(f *os.File, recordToUpdateIndexRow gjson.Result, infileIndex, rs *gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, *gjson.Result, error) {
	fileStartOffset := recordToUpdateIndexRow.Get("startOffset").Int()
	dataSize := recordToUpdateIndexRow.Get("dataSize").Int()
	if fileStartOffset == 0 || dataSize == 0 {
		loggermdl.LogError("index row details incorrect - start offset :", fileStartOffset, " data size :", dataSize)
		return nil, nil, errormdl.Wrap("index row details incorrect")
	}
	// compare data size
	// TODO: compare with encrypted data

	// update fields data
	// existingData, err := filemdl.ReadFileFromOffset(f, fileStartOffset, dataSize)
	existingData, err := getFileDataFromPack(f, fileStartOffset, dataSize, rs, secParams)
	if err != nil {
		return nil, nil, err
	}

	updatedDataStr := strings.TrimSpace(string(existingData))
	rs.ForEach(func(key, val gjson.Result) bool {
		// updating existing data
		updatedDataStr, _ = sjson.Set(updatedDataStr, key.String(), val.Value())
		return true
	})
	newDataSize := int64(len(updatedDataStr))
	footerStartOffset := getFooterOffset(f)
	updatedFooterOffset := footerStartOffset
	err = setFileStatusFlag(f, fileStatusUpdatingData)
	if err != nil {
		return nil, nil, err
	}
	indexDataString := "[]"
	if infileIndex == nil {
		indexDataString, err = getInFileIndexData(f)
		if err != nil {
			loggermdl.LogError(err)
			return nil, nil, err
		}
	} else {
		indexDataString = infileIndex.String()
	}
	existingIndexRows := gjson.Parse(indexDataString)
	if len(strings.TrimSpace(updatedDataStr)) <= len(strings.TrimSpace(string(existingData))) {
		newDataSize, err = addFileDataInFile(f, fileStartOffset, updatedDataStr, false, rs, secParams)
		if err != nil {
			return nil, nil, err
		}
		// newDataSize = newDataSize + 2
	} else {

		newDataSize, err = addFileDataInFile(f, footerStartOffset, updatedDataStr, true, rs, secParams)
		updatedFooterOffset = footerStartOffset + newDataSize
		fileStartOffset = footerStartOffset
	}
	// indexRows, err := getInFileIndexData(f)
	// if err != nil {
	// 	return nil, err
	// }
	// update startofffset and datasize in infile index
	// updateIndexRow
	updatedIndexRowStr := recordToUpdateIndexRow.String()

	recordToUpdateIndexRow.ForEach(func(key, value gjson.Result) bool {
		indexFieldKey := key.String()
		if rs.Get(indexFieldKey).Exists() {
			updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, indexFieldKey, rs.Get(indexFieldKey).Value())
		}
		return true
	})
	fileHash, err := securitymdl.GetHash(updatedDataStr)
	if err != nil {
		return nil, nil, err
	}
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "startOffset", fileStartOffset)
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "dataSize", newDataSize)
	updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "fileHash", fileHash)
	updatedIndexRows, err := updateIndexRow(&existingIndexRows, recordToUpdateIndexRow, gjson.Parse(updatedIndexRowStr))

	if err != nil {
		return nil, nil, err
	}

	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return nil, nil, err
	}
	// loggermdl.LogDebug("updated index data", updatedIndexRows)
	err = setIndexDataInFile(f, updatedFooterOffset, updatedIndexRows.String())
	if err != nil {
		return nil, nil, err
	}
	err = setFooterOffset(f, updatedFooterOffset)
	if err != nil {
		return nil, nil, err
	}

	err = setFooterSize(f, int64(len(updatedIndexRows.String())))
	if err != nil {
		return nil, nil, err
	}

	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return nil, nil, err
	}
	f.Sync()
	// TODO: discussion
	// err = setFileReadyForUploadFlag(f, true)
	// if err != nil {
	// 	return err
	// }
	updatedData := gjson.Parse(updatedDataStr)
	return &updatedData, updatedIndexRows, nil
}

// DeletDataFromPackFileUsingFp -DeletDataFromPackFileUsingFp
func DeletDataFromPackFileUsingFp(f *os.File, infileIndex *gjson.Result, infileIndexQueries []string) (int, *gjson.Result, error) {
	// get indexdata
	// get update index records
	// save updated index
	// update index size header
	// lock, err := filemdl.AcquireFileLock(filePath)
	// defer filemdl.ReleaseFileLock(lock)
	// if err != nil {
	// 	return err
	// }

	recordsDeletedCnt := 0
	lock := getLock(f.Name())
	lock.Lock()
	// f, err := filemdl.OpenFile(filePath, os.O_RDWR|os.O_SYNC, 0777)
	defer func() {
		lock.Unlock()
		// f.Close()
	}()
	// if err != nil {
	// 	return recordsDeletedCnt, err
	// }
	indexDataStr := "[]"
	var err error
	if infileIndex == nil {

		indexDataStr, err = getInFileIndexData(f)
		if err != nil {
			return recordsDeletedCnt, infileIndex, err
		}
		loggermdl.LogError("infile index in file", indexDataStr)
	} else {
		indexDataStr = infileIndex.String()
		loggermdl.LogError("infile index in cache", indexDataStr)
	}

	indexData := gjson.Parse(indexDataStr)
	indexRecordsToDelete := indexData
	// loggermdl.LogDebug("indexRecordsToDelete file type", indexRecordsToDelete)
	for _, query := range infileIndexQueries {
		indexRecordsToDelete = indexRecordsToDelete.Get(query + "#")
	}
	indexRowsToDelete := indexRecordsToDelete.Array()
	if len(indexRowsToDelete) == 0 {
		loggermdl.LogError("ErrNoDataFound")
		return recordsDeletedCnt, infileIndex, ErrNoDataFound
	}
	loggermdl.LogError("before delete ", indexData)
	updatedIndexRecords := indexData
	for _, indexRowToRemove := range indexRowsToDelete {
		updatedIndexRecords, err = removeIndexRow(updatedIndexRecords, indexRowToRemove.String())
		if err != nil {
			loggermdl.LogError("fail to delete record:", err)
			return recordsDeletedCnt, infileIndex, errormdl.Wrap("fail to delete record:" + err.Error())
		}
		recordsDeletedCnt++
	}
	loggermdl.LogError("updatedIndexRecords after delete ", updatedIndexRecords, f.Name())
	footerOffset := getFooterOffset(f)
	if footerOffset == -1 {
		return recordsDeletedCnt, infileIndex, errormdl.Wrap("fail to fetch infile index offset")
	}
	newIndexDataSize := len(updatedIndexRecords.String())
	err = setFileStatusFlag(f, fileStatusUpdatingIndex)
	if err != nil {
		return recordsDeletedCnt, infileIndex, err
	}
	err = setIndexDataInFile(f, footerOffset, updatedIndexRecords.String())
	if err != nil {
		loggermdl.LogError("fail to update infile index data :", err)
		return recordsDeletedCnt, infileIndex, err
	}
	err = setFileStatusFlag(f, fileStatusReady)
	if err != nil {
		return recordsDeletedCnt, infileIndex, err
	}
	return recordsDeletedCnt, &updatedIndexRecords, setFooterSize(f, int64(newIndexDataSize))
}

func encryptWithCompression(data []byte, compress bool, encKey []byte) (res []byte, err error) {
	if compress {
		res, err = filemdl.ZipBytes(data)
		if err != nil {
			return
		}
	}

	return securitymdl.AESEncrypt(res, encKey)
}

func decryptwithDecompression(data []byte, deCompress bool, encKey []byte) (res []byte, err error) {
	res, err = securitymdl.AESDecrypt(data, encKey)
	if err != nil {
		return
	}

	if deCompress {
		return filemdl.UnZipBytes(res)
	}

	return
}

func ReorgPackFile(filePath string) error {

	isFilePresent := filemdl.FileAvailabilityCheck(filePath)
	if !isFilePresent {
		return errormdl.Wrap("file not found")
	}
	lock := getLock(filePath)
	lock.Lock()
	f, err := os.OpenFile(filePath, os.O_CREATE|os.O_RDWR, os.ModePerm)
	defer func() {
		lock.Unlock()
		f.Close()
	}()
	if err != nil {
		loggermdl.LogError("Error occured during reOrg of file data", err)
		return errormdl.Wrap("Error occured during reOrg of file data")
	}
	_, sourceFileName := filepath.Split(filePath)
	desFileName := sourceFileName + "_" + strconv.FormatInt(time.Now().Unix(), 10)
	tempFilepath, err := filepath.Abs(filepath.Join(filemdl.TempDir, desFileName))
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	tempFilDir, _ := filepath.Split(tempFilepath)
	err = filemdl.CreateDirectoryRecursive(tempFilDir)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	err = initializeWithHeader(tempFilepath)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	dir, _ := filepath.Split(tempFilepath)
	if dir != "" {
		createError := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(createError) != nil {
			return errormdl.CheckErr(createError)
		}
	}
	fpTemp, err := os.OpenFile(tempFilepath, os.O_CREATE|os.O_RDWR, os.ModePerm)
	if err != nil {
		return err
	}
	defer func() {
		fpTemp.Close()
	}()
	infileIndexData, err := getInFileIndexData(f)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	infileIndexRows := gjson.Parse(infileIndexData)
	if len(infileIndexRows.Array()) == 0 {
		return nil
	}
	tempFileFooterStartOffset := getFooterOffset(fpTemp)
	if tempFileFooterStartOffset == -1 {
		loggermdl.LogError("fail to fetch infile index offset")
		return errormdl.Wrap("fail to fetch infile index data")
	}
	updatedIndexRowStr := "[]"
	for _, infileIndex := range infileIndexRows.Array() {
		startOffset, err := strconv.Atoi(infileIndex.Get("startOffset").String())
		if err != nil {
			loggermdl.LogError("Error occured while fetching startOffset", err)
			return err
		}
		dataSize, err := strconv.Atoi(infileIndex.Get("dataSize").String())
		if err != nil {
			loggermdl.LogError("Error occured while fetching dataSize", err)
			return err
		}

		byteArr, err := getFileDataFromPack(f, int64(startOffset), int64(dataSize), nil, securitymdl.FDBSecParams{EnableSecurity: false})
		if err != nil {
			loggermdl.LogError("Error occured while reading file data from offset", err)
			return err
		}
		byteCnt, err := addByteDataInFile(fpTemp, tempFileFooterStartOffset, byteArr, false)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		indexRowJSON, _ := sjson.Set(infileIndex.String(), "startOffset", tempFileFooterStartOffset)
		indexRowJSON, _ = sjson.Set(indexRowJSON, "dataSize", byteCnt)
		indexRowJSONObj := gjson.Parse(indexRowJSON)
		updatedIndexRowStr, _ = sjson.Set(updatedIndexRowStr, "-1", indexRowJSONObj.Value())
		tempFileFooterStartOffset = tempFileFooterStartOffset + byteCnt
	}

	err = setFooterOffset(fpTemp, tempFileFooterStartOffset)
	if err != nil {
		return err
	}
	err = setFooterSize(fpTemp, int64(len(updatedIndexRowStr)))
	if err != nil {
		return err
	}
	err = setIndexDataInFile(fpTemp, tempFileFooterStartOffset, updatedIndexRowStr)
	if err != nil {
		return err
	}
	err = fpTemp.Sync()
	if err != nil {
		return err
	}
	err = fpTemp.Close()
	if err != nil {
		return err
	}
	err = f.Close()
	if err != nil {
		return err
	}

	return filemdl.AtomicReplaceFile(tempFilepath, filePath)
}
