package routebuildermdl

import (
	"context"
	"strings"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/constantmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/servicebuildermdl"

	version "github.com/hashicorp/go-version"
	"github.com/tidwall/gjson"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/cachemdl"
)

var (
	restrictedServices cachemdl.FastCacheHelper
	roleBasedServices  cachemdl.FastCacheHelper
	openServices       cachemdl.FastCacheHelper
	loginService       func(*gjson.Result, servicebuildermdl.Principal) (interface{}, string, error)

	applicationVersion      *version.Version
	minimumSupportedVersion *version.Version
	isAppVersionEnabled     bool
	isStrictMode            bool
)

func init() {
	restrictedServices.Setup(10, time.Second*600, time.Second*600)
	openServices.Setup(10, time.Second*600, time.Second*600)
	roleBasedServices.Setup(10, time.Second*600, time.Second*600)
}

// Hook - Hook
type Hook struct {
	ActivityName string
	ActorType    string
	Branch       string
}

// ServiceCache ServiceCache object
type ServiceCache struct {
	Service            func(*gjson.Result, servicebuildermdl.Principal) (interface{}, *servicebuildermdl.AbstractBusinessLogicHolder, error)
	AsyncService       bool
	HeavyDataActivity  bool
	PreHooks           []Hook
	PostHooks          []Hook
	Groups             []string
	TransactionEnabled bool
	DatabaseType       string // database type for transaction begin(MYSQL,SQLSERVER etc.)
	TransactionHost    string // host to begin transaction. if it is empty then transaction will begin on default host
}

// ServiceType -ServiceType
type ServiceType = func(*gjson.Result, servicebuildermdl.Principal) (interface{}, *servicebuildermdl.AbstractBusinessLogicHolder, error)

// GetRestrictedServicePtr -GetRestrictedServicePtr
func GetRestrictedServicePtr(serviceName string) (ServiceType, error) {
	intService, ok := restrictedServices.Get(serviceName)
	if !ok {
		return nil, errormdl.Wrap("Service Not found")
	}
	serviceCache, ok := intService.(ServiceCache)
	if !ok {
		return nil, errormdl.Wrap("Service Not Casted")
	}
	return serviceCache.Service, nil
}

// GetOpenServicePtr -GetOpenServicePtr
func GetOpenServicePtr(serviceName string) (ServiceType, error) {
	intService, ok := openServices.Get(serviceName)
	if !ok {
		return nil, errormdl.Wrap("Service Not found")
	}
	serviceCache, ok := intService.(ServiceCache)
	if !ok {
		return nil, errormdl.Wrap("Service Not Casted")
	}
	return serviceCache.Service, nil
}

// GetRoleBasedServicePtr -GetRoleBasedServicePtr
func GetRoleBasedServicePtr(serviceName string) (ServiceType, error) {
	intService, ok := roleBasedServices.Get(serviceName)
	if !ok {
		return nil, errormdl.Wrap("Service Not found")
	}
	serviceCache, ok := intService.(ServiceCache)
	if !ok {
		return nil, errormdl.Wrap("Service Not Casted")
	}
	return serviceCache.Service, nil
}

// RegisterNormalService is for transactional services
func RegisterNormalService(serviceName string, servicePtr ServiceType, isRestricted, isRoleBased bool) {
	service := ServiceCache{
		Service: servicePtr,
	}
	commonServiceRegistration(serviceName, service, isRestricted, isRoleBased)
}

// RegisterServiceInSLS is for transactional services
func RegisterServiceInSLS(serviceName string, servicePtr ServiceCache, isRestricted, isRoleBased bool) {
	// service := ServiceCache{
	// 	Service: servicePtr,
	// }
	commonServiceRegistration(serviceName, servicePtr, isRestricted, isRoleBased)
}

// RegisterLoginSerivce is for login service
func RegisterLoginSerivce(service func(*gjson.Result, servicebuildermdl.Principal) (interface{}, string, error)) {
	loginService = service
}

func commonServiceRegistration(serviceName string, service ServiceCache, isRestricted, isRoleBased bool) {
	if isRestricted {
		if isRoleBased {
			roleBasedServices.SetNoExpiration(serviceName, service)
		} else {
			restrictedServices.SetNoExpiration(serviceName, service)
		}
	} else {
		openServices.SetNoExpiration(serviceName, service)
	}
}

// SetAppVersion -SetAppVersion
func SetAppVersion(appVersion, minSupportedVersion string) error {

	app, err := version.NewVersion(appVersion)
	if errormdl.CheckErr(err) != nil {
		return errormdl.CheckErr(err)
	}
	min, err := version.NewVersion(minSupportedVersion)
	if errormdl.CheckErr(err) != nil {
		return errormdl.CheckErr(err)
	}

	versionValidation := min.Compare(app)
	if versionValidation == 1 {
		return errormdl.Wrap("Minimum version is more than app version")
	}
	if versionValidation == 0 {
		isStrictMode = true
	}
	applicationVersion = app
	minimumSupportedVersion = min
	isAppVersionEnabled = true
	return nil
}

// CallService - calls service with provided configuration, returns result and error from executed service
func CallService(name string, rs *gjson.Result, isRestricted bool, isRoleBased bool, p servicebuildermdl.Principal) (interface{}, error) {
	return CallServiceWithBranch(name, constantmdl.Branch_Main, rs, isRestricted, isRoleBased, p)
}

// CallServiceWithBranch - calls service of given branch with provided configuration, returns result and error from executed service
func CallServiceWithBranch(name, branch string, rs *gjson.Result, isRestricted bool, isRoleBased bool, p servicebuildermdl.Principal) (interface{}, error) {
	var found bool
	var service interface{}
	if strings.TrimSpace(branch) == "" {
		loggermdl.LogError("Branch is empty")
		return nil, errormdl.Wrap("Branch is empty")
	}
	activityName := ConcatenateEntityWithBranch(name, branch)
	// if isRestricted {
	// 	if isRoleBased {
	// 		service, found = roleBasedServices.Get(activityName)
	// 	} else {
	// 		service, found = restrictedServices.Get(activityName)
	// 	}
	// } else {
	// 	service, found = openServices.Get(activityName)
	// }
	switch {
	case isRestricted == true && isRoleBased == true:
		service, found = roleBasedServices.Get(activityName)
	case isRestricted == true && isRoleBased == false:
		service, found = restrictedServices.Get(activityName)
	case isRestricted == false:
		service, found = openServices.Get(activityName)
	}

	if !found {
		loggermdl.LogError("Service Not Found: " + name)
		return nil, errormdl.Wrap("Service Not Found: " + name)
	}

	tmpSvcCache, ok := service.(ServiceCache)
	if !ok {
		loggermdl.LogError("Unable to cast service object for ", name)
		return nil, errormdl.Wrap("Service execution failed for " + name)
	}

	serviceCache := tmpSvcCache
	res, _, err := serviceCache.Service(rs, p)
	if err != nil {
		loggermdl.LogError("Service execution failed for ", name, " : ", err)
		return nil, err
	}
	return res, nil
}

// CallServiceTXN - calls service with provided configuration, returns result, abstractBusinessLogicHolder and error from executed service
func CallServiceReturnAB(name string, rs *gjson.Result, isRestricted bool, isRoleBased bool, p servicebuildermdl.Principal) (interface{}, *servicebuildermdl.AbstractBusinessLogicHolder, error) {
	return CallServiceReturnABWithBranch(name, constantmdl.Branch_Main, rs, isRestricted, isRoleBased, p)
}

// CallServiceWithBranchHandleTXN - calls service of given branch with provided configuration, returns result, abstractBusinessLogicHolder and error from executed service and handles transaction
func CallServiceReturnABWithBranch(name, branch string, rs *gjson.Result, isRestricted bool, isRoleBased bool, p servicebuildermdl.Principal) (interface{}, *servicebuildermdl.AbstractBusinessLogicHolder, error) {
	var found bool
	var service interface{}
	if strings.TrimSpace(branch) == "" {
		loggermdl.LogError("Branch is empty")
		return nil, nil, errormdl.Wrap("Branch is empty")
	}
	activityName := ConcatenateEntityWithBranch(name, branch)
	// if isRestricted {
	// 	if isRoleBased {
	// 		service, found = roleBasedServices.Get(activityName)
	// 	} else {
	// 		service, found = restrictedServices.Get(activityName)
	// 	}
	// } else {
	// 	service, found = openServices.Get(activityName)
	// }
	switch {
	case isRestricted == true && isRoleBased == true:
		service, found = roleBasedServices.Get(activityName)
	case isRestricted == true && isRoleBased == false:
		service, found = restrictedServices.Get(activityName)
	case isRestricted == false:
		service, found = openServices.Get(activityName)
	}

	if !found {
		loggermdl.LogError("Service Not Found: " + name)
		return nil, nil, errormdl.Wrap("Service Not Found: " + name)
	}

	tmpSvcCache, ok := service.(ServiceCache)
	if !ok {
		loggermdl.LogError("Unable to cast service object for ", name)
		return nil, nil, errormdl.Wrap("Service execution failed for " + name)
	}

	serviceCache := tmpSvcCache
	res, ab, err := serviceCache.Service(rs, p)
	loggermdl.LogError("error while calling activity:", err.Error())
	if errormdl.CheckErr1(err) != nil {
		// Rollback
		if ab.TransactionEnable {
			switch ab.DatabaseType {
			case dalmdl.MYSQL:
				if ab.TXN != nil {
					loggermdl.LogError("MYSQL Transaction Rollbacked")
					err := ab.TXN.Rollback()
					if err != nil {
						loggermdl.LogError("Error while rollback")
						return res, ab, errormdl.Wrap("Error while rollback")
					}
				}

			case dalmdl.SQLSERVER:
				if ab.SQLServerTXN != nil {
					loggermdl.LogError("SQLSERVER Transaction Rollbacked")
					err := ab.SQLServerTXN.Rollback()
					if err != nil {
						loggermdl.LogError("Error while rollback")
						return res, ab, errormdl.Wrap("Error while rollback")
					}
				}

			case dalmdl.GraphDB:
				if ab.GraphDbTXN != nil {
					loggermdl.LogError("GRAPHDB Transaction Rollbacked")
					err := ab.GraphDbTXN.Discard(context.TODO())
					if err != nil {
						loggermdl.LogError("Error while rollback")
						return res, ab, errormdl.Wrap("Error while rollback")
					}
				}

			default:
				loggermdl.LogError("Invalid database type while rollback transaction")
			}
		}
		return res, ab, err //returning error
	} else {
		if ab.TransactionEnable {
			// Commit
			switch ab.DatabaseType {
			case dalmdl.MYSQL:
				if ab.TXN != nil {
					loggermdl.LogError("MYSQL Transaction Commited")
					err := ab.TXN.Commit()
					if err != nil {
						loggermdl.LogError("Error while commit")
						return res, ab, errormdl.Wrap("Error while commit")
					}
				}

			case dalmdl.SQLSERVER:
				if ab.SQLServerTXN != nil {
					loggermdl.LogError("SQLSERVER Transaction Commited")
					err := ab.SQLServerTXN.Commit()
					if err != nil {
						loggermdl.LogError("Error while commit")
						return res, ab, errormdl.Wrap("Error while commit")
					}
				}

			case dalmdl.GraphDB:
				if ab.GraphDbTXN != nil {
					loggermdl.LogError("GRAPHDB Transaction Commited")
					err := ab.GraphDbTXN.Commit(context.TODO())
					if err != nil {
						loggermdl.LogError("Error while commit")
						return res, ab, errormdl.Wrap("Error while commit")
						// responseDataObj.Error = err.Error()
						// responseDataObj.ErrorCode = errormdl.GRAPHDBERROR
					}
				}
			default:
				loggermdl.LogError("Invalid database type while commiting transaction")
			}
		}
	}

	return res, ab, nil
}
