package securitymdl

/*

All FDB related security options are provided.

User needs to Set fdb options on application startup.
In case of CS, this is done by restricted and rolebased global cache.

*/

import (
	"errors"
	"path/filepath"
	"strings"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/hashmdl"

	"github.com/tidwall/gjson"
)

const (
	SEC_ENCKey     = "encKey"
	SEC_UserKey    = "userKey"
	SEC_FieldQuery = "fieldQuery"
	SEC_SECURITY   = "security"
)

const (
	// SharedPrefix prefix represents the file is sharable i.e encryption key does not include (fdbSec.fieldQuery)
	SharedPrefix = "ss_"
	EmptySTR     = ""
)

// FDBSecOptions provide options to be set for fdb security.
// All fields are compulsorry. On applying options, we will get error if any field is empty.
// Only these options are exposed as actual
type FDBSecOptions struct {
	EncKey         string // the global encryption key used in the project. This key will be applicable in all cases.
	UserDefinedKey string // the user defined key in the project. This key will be applicable in all cases.
	FieldQuery     string // query to get dynamic field. Ex. Each student data can be encrypted with studentID. Applicable only for the shared bucket.
}

// FDBSecParams provide options for the FDB security for the FDB CRUD operations.
type FDBSecParams struct {
	EnableSecurity    bool
	EnableCompression bool
}

func NewFDBOptions() *FDBSecOptions {
	return new(FDBSecOptions)
}

// Apply sets the internal security variable.
//
// Returns error if options fail while validations and if Apply() is already called.
func (fo *FDBSecOptions) Apply() error {
	if fdbSec.setSec {
		return errors.New("fdb security is already set")
	}

	if err := fo.Validate(); err != nil {
		return err
	}
	fdbSec.setSec = true
	fdbSec.encKey = fo.EncKey
	fdbSec.userDefinedKey = fo.UserDefinedKey
	fdbSec.fieldQuery = fo.FieldQuery

	fdbSec.enableSharedFiles = fdbSec.fieldQuery != EmptySTR // if true, trow error if result from query is empty string
	return nil
}

type fdbSecurity struct {
	setSec         bool   // set is in Apply() method.
	encKey         string // the global encryption key used in the project
	userDefinedKey string // the user defined key in the project
	fieldQuery     string // query to get dynamic field. Ex. Each student data can be encrypted with studentID. Allowed to be empty. If empty, use same encryption for all cases.

	enableSharedFiles bool // True if fieldQuery is not empty.
}

var fdbSec fdbSecurity

// Validate checks options for empty fields.
func (fs *FDBSecOptions) Validate() error {
	if fs.EncKey == "" {
		return errors.New("FDB security: encryption key can not be empty")
	}
	if fs.UserDefinedKey == "" {
		return errors.New("FDB security: user defined key can not be empty")
	}
	// if fs.FieldQuery == "" {
	// 	// this is allowed for now. If empty, same encryption is used for all buckets.
	// 	// return errors.New("FDB security: field query can not be empty")
	// }
	return nil
}

// GetFDBSecOptions returns the copy of fdb options set by Apply() method.
func GetFDBSecOptions() FDBSecOptions {
	return FDBSecOptions{
		EncKey:         fdbSec.encKey,
		UserDefinedKey: fdbSec.userDefinedKey,
		FieldQuery:     fdbSec.fieldQuery,
	}
}

// GenSecKey generates the security key for FDB.
//
// If fileName has prefix "ss_" then key is generated by
// 1. Global encryption key (fdbSec.encKey)
// 2. UserDefinedKey (fdbSec.userDefinedKey)
// 3. fileName
//
// else result by firing (fdbSec.fieldQuery) on data is also included in generation. If rusult is empty, expect an error.
//
// Order for key: data.Get(fdbSec.fieldQuery).String() + fdbSec.encKey + fdbSec.userDefinedKey + fileName(only)
func GenSecKeyBytes(fileName string, data *gjson.Result) ([]byte, error) {
	// here data is String not JSON
	rs := gjson.Parse(data.String())
	_, fileName = filepath.Split(fileName)
	if fileName == EmptySTR {
		return nil, errors.New("GenerateSecurityKey: fileName must not be empty")
	}

	// Warning: The order of string concatenation must be preserved as specified.
	key := fdbSec.encKey + fdbSec.userDefinedKey + fileName

	useFieldQuery, err := CheckDataForFieldQuery(fileName, &rs)
	if err != nil {
		return nil, err
	}

	if useFieldQuery {
		key = rs.Get(fdbSec.fieldQuery).String() + key
	}
	hash, err := hashmdl.Get128BitHash([]byte(key))
	if err != nil {
		return nil, errors.New("GenerateSecurityKey: " + err.Error())
	}

	hashedKey := hash[:]
	return hashedKey, nil
}

func GenSecKey(fileName string, data *gjson.Result) (string, error) {
	key, err := GenSecKeyBytes(fileName, data)
	if err != nil {
		return "", err
	}

	return string(key), nil
}

// CheckDataForFieldQuery checks for result of field query in data.
//
// If no query is provided OR Not shared file, the result will be false, nil
//
// Else will check for query result. If result is empty, error will be thrown.
//
// Check only if security is enabled for the database.
//
// fileName must not be empty and data must not be nil
func CheckDataForFieldQuery(fileName string, data *gjson.Result) (bool, error) {
	if strings.HasPrefix(fileName, SharedPrefix) || !fdbSec.enableSharedFiles {
		// this is a shared file OR no query provided. No need to check for dynamic query result.
		return false, nil
	}

	if data.Get(fdbSec.fieldQuery).String() == EmptySTR {
		return false, errors.New("CheckDataForFieldQuery: got empty data for defined field query " + fdbSec.fieldQuery)
	}

	return true, nil
}
